/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef DATACACHE_H
#define DATACACHE_H

#include "RCAttr.h"
#include "core/RCAttrPack.h"
#include "core/DPN.h"
#ifdef __GNUC__
	#include <ext/hash_map>
	#ifndef stdext
	#define stdext __gnu_cxx
	#endif
#else
	#include <hash_map>
#endif
#include <utility>

#include <boost/shared_ptr.hpp>

class RCAttr;
class IBMutex;

// NOTE: These classes are still undergoing significant changes
// TBD:
//	- more complete lifecycle, interface with memory manager as appropriate
// 	- need equivalent semantics of "lock in memory"
class DataCache
{
	protected:

		typedef struct {
			int table,column,dp;
		} PackCoordinate;
#ifdef __GNUC__
		class PCEqual {
		public:
			bool operator()(const PackCoordinate &a, const PackCoordinate &b)
			{
			return (a.table == b.table) &&
				(a.column == b.column) &&
				(a.dp == b.dp);
			}
		};

		class PCHash 
		{
		public:
			// try and get as much randomness as possible in the lowest order bits
			// because small hash tables will have most collisions there
			size_t
			operator()(const PackCoordinate& a) const
			{ return (a.table << 6) ^ (a.column << 4) ^ (a.dp); };
		};

		//
		// need size limits
		//

		typedef stdext::hash_map<PackCoordinate, AttrPack *, PCHash, PCEqual> PackContainer;
		// might have to store DPNs in a vector if the lookup is too slow...
		// are DPNs mostly accessed in a random pattern or traversed in order?
		typedef stdext::hash_map<PackCoordinate, DPN *, PCHash, PCEqual> DPNContainer;
#else
		class PCHashCompare
		{
		public:
			const static  size_t bucket_size = 4;
			const static size_t min_buckets = 8;
			bool operator()(const PackCoordinate& value1, const PackCoordinate& value2) const
			{
				return	(value1.table < value2.table) ||
						(value1.table == value2.table && value1.column < value2.column) ||
						(value1.table == value2.table && value1.column == value2.column && value1.dp < value2.dp);
			};

			size_t operator()(const PackCoordinate& value1) const
			{
				return (value1.table << 6) ^ (value1.column << 4) ^ (value1.dp); 
			};
		};

		typedef stdext::hash_map<PackCoordinate, AttrPack*, PCHashCompare> PackContainer;
#endif

	    PackContainer m_datapacks;
	    //DPNContainer m_dpns;
		//std::map<AttrId, boost::shared_ptr<AttrDataCache> > cache;
        _int64 m_cacheHits, m_cacheMisses, m_packsReleased;	
	public:

		DataCache() : m_cacheHits(0), m_cacheMisses(0), m_packsReleased(0) {}
		
		virtual AttrPack * GetAttrPackRO(int table, int column, int dp);
		virtual AttrPack * GetAttrPackRW(int table, int column, int dp) { return GetAttrPackRO(table, column, dp); }

		// what when the dp is already in the container?
		//Currently:	if this is the same one then do nothing
		//				if it is new one then replace the old one
		virtual bool PutAttrPackRO(int table, int column, int dp, AttrPack *p);

		// This is for new DPs
		virtual bool PutAttrPackRW(int table, int column, int dp, AttrPack *p) { return !true;}

		// what when dp isn't in the map?
		// TBD: does caller unlock pack from use?
		virtual void ReleasePack(int table, int column, int dp);
		virtual void ReleasePack(AttrPack* p);
		virtual void ReleasePacks(int table);
		virtual void ReleaseColumnPacks(int table, int column);
		virtual void ReleasePackRow(int table, int pack_row);

		virtual void MoveCache(DataCache& local);

};

// Threadsafe version of DataCache
class SynchronizedDataCache : public DataCache
{

	public:

		virtual AttrPack * GetAttrPackRO(int table, int column, int dp);

		// make a copy of an existing pack and put it in the local map
		virtual AttrPack * GetAttrPackRW(int table, int column, int dp) { return GetAttrPackRO(table, column, dp); }

		// what when the dp is already in the container?
		//Currently:	if this is the same one then do nothing
		//				if it is new one then replace the old one
		virtual bool PutAttrPackRO(int table, int column, int dp, AttrPack *p);

		// This is for new DPs
		virtual bool PutAttrPackRW(int table, int column, int dp, AttrPack *p) { return !true;}

		// what when dp isn't in the map?
		// TBD: does caller unlock pack from use?
		virtual void ReleasePack(int table, int column, int dp);
		virtual void ReleasePack(AttrPack* p);
		virtual void ReleasePacks(int table);
		virtual void ReleaseColumnPacks(int table, int column);
		virtual void ReleasePackRow(int table, int pack_row);

		virtual void MoveCache(DataCache& local);

	private:
	
		IBMutex m_cache_mutex;


	static SynchronizedDataCache global_data_cache;

public:
	static SynchronizedDataCache& GlobalDataCache() { return global_data_cache; }



#if 0
		// should be in a generic header 
		template<typename T>
		class Allocator {
			public:
			virtual T* alloc() = 0;	
		};

		// should be in the same header as AttrPack
		class AttrPackAllocator : public Allocator<AttrPack> 
		{
				AttrPackType pt;
				RCAttr *rc;
				int pack_no;
				
				public:

				AttrPackAllocator( AttrPackType t, RCAttr *r, int p ) :
					pt(t), rc(r), pack_no(p)
				{}

				void setPackNum(int p) {
					pack_no = p;
				}

				AttrPack* alloc() {
					if(pt == PackS)
						return new AttrPackS(*rc, pack_no);
					else	
						return new AttrPackN(*rc, pack_no);
				}
		};
#endif


};

/*

// delegates requests to global if it can't find a local copy
class LocalPackCache : public DataPackCache
{
	GlobalPackCache &m_global;
	PackMap m_modifiedPacks;

	public:
	LocalPackCache( GlobalPackCache &g ) : m_global(g) {}

	virtual AttrPack * GetPack(int table,int column,int dp)
	{
		PackCoordinate coord;
		coord.table = table;
		coord.column = column;
		coord.dp = dp;
		PackMap::iterator it = m_modifiedPacks.find( coord );
		if( it != m_modifiedPacks.end()) {
			return (*it).second;
		} else {
			AttrPack *res = DataPackCache::GetPack(table,column,dp);
			if( res != NULL ) {
				return res;
			} else {
				return m_global.GetPack(table,column,dp);
			}
		}
			
	}
		
	//virtual ReleasePack(RCAttrPack *)
	//virtual PersistPack(RCAttrPack *)
};
*/

#endif
