#include "Descriptor.h"
#include "vc/VirtualColumn.h"
#include "vc/ConstColumn.h"
#include "vc/ConstExpressionColumn.h"
#include "vc/ExpressionColumn.h"
#include "vc/TypeCastColumn.h"
#include "vc/InSetColumn.h"
#include "vc/SubSelectColumn.h"
#include "vc/SingleColumn.h"
#include "QueryOperator.h"
#include "ValueOrNull.h"
#include "Query.h"

using namespace std;

Descriptor::Descriptor(TempTable* t)
	:	op(O_TRUE), lop(O_AND), sharp(false), encoded(false), done(false),
		evaluation(0), outer_dim(-1), delayed(false),
		table(t), desc_t(DT_NOT_KNOWN_YET), collation(DTCollation())
{ assert(table);
}

Descriptor::~Descriptor()
{
}

Descriptor::Descriptor(const Descriptor& desc)
{
	attr = desc.attr;
	op = desc.op;
	lop = desc.lop;
	val1 = desc.val1;
	val2 = desc.val2;
	sharp = desc.sharp;
	encoded = desc.encoded;
	done = desc.done;
	evaluation = desc.evaluation;
	outer_dim = desc.outer_dim;
	desc_t = desc.desc_t;
	delayed = desc.delayed;
	table = desc.table;
	collation = desc.collation;
	buf_attr = desc.buf_attr;
	buf_val1 = desc.buf_val1;
	buf_len = desc.buf_len;
}

void Descriptor::swap(Descriptor& d)
{
	std::swap(op, d.op);
	std::swap(lop, d.lop);
	std::swap(attr, d.attr);
	std::swap(val1, d.val1);
	std::swap(val2, d.val2);
	std::swap(sharp, d.sharp);
	std::swap(encoded, d.encoded);
	std::swap(done, d.done);
	std::swap(evaluation, d.evaluation);
	std::swap(desc_t, d.desc_t);
	std::swap(delayed, d.delayed);
	std::swap(outer_dim, d.outer_dim);
	std::swap(table, d.table);
	std::swap(collation, d.collation);
	std::swap(buf_attr, d.buf_attr);
	std::swap(buf_val1, d.buf_val1);
	std::swap(buf_len, d.buf_len);
}

Descriptor& Descriptor::operator=(const Descriptor& desc)
{
	Descriptor tmp(desc);
	swap(tmp);
	return *this;
}

int Descriptor::operator==(const Descriptor& sec)
{
	return (
		attr == sec.attr	&&
		op == sec.op		&&
		lop == sec.lop		&&
		val1 == sec.val1	&&
		val2 == sec.val2	&&
		sharp == sec.sharp	&&
		encoded == sec.encoded &&
		delayed == sec.delayed &&
		outer_dim == sec.outer_dim &&
		table == sec.table &&
		collation.collation == sec.collation.collation &&
		collation.derivation == sec.collation.derivation
		);
}

bool Descriptor::EqualExceptOuter(const Descriptor& sec)
{
	return (
		attr == sec.attr	&&
		op == sec.op		&&
		lop == sec.lop		&&
		val1 == sec.val1	&&
		val2 == sec.val2	&&
		sharp == sec.sharp	&&
		table == sec.table  &&
		collation.collation == sec.collation.collation &&
		collation.derivation == sec.collation.derivation
		);
}

void Descriptor::SwitchSides()		// change "a<b" into "b>a" etc; throw error if not possible (e.g. between)
{
	if( op != O_EQ && op != O_NOT_EQ && op != O_LESS && op != O_MORE &&
		op != O_LESS_EQ && op != O_MORE_EQ )
		throw NotImplementedRCException("Cannot switch this operator");
	SwitchOperator(op);
	CQTerm p = attr;
	attr = val1;
	val1 = p;
}

bool Descriptor::IsOneDimensional()
{
	set<int> all_dims = GetDimensions();
	return (all_dims.size() <= 1);
}

set<int> Descriptor::GetDimensions()
{
	set<int> dims, all_dims;
	if(attr.vc) {
		dims = attr.vc->GetDimensions();
		all_dims.insert(dims.begin(), dims.end());
	}
	if(val1.vc) {
		dims = val1.vc->GetDimensions();
		all_dims.insert(dims.begin(), dims.end());
	}
	if(val2.vc) {
		dims = val2.vc->GetDimensions();
		all_dims.insert(dims.begin(), dims.end());
	}
	return all_dims;
}

bool Descriptor::IsType_AnyVal()						// true if "column op val" or "column IS NULL/NOT NULL
{
	return 	(val1.vc && val1.vc->IsConst() && !val2.vc)
			|| op == O_IS_NULL || op == O_NOT_NULL;
}

bool Descriptor::IsType_AnyValAny()						// true if "column op val"
{
	return 	val1.vc && val1.vc->IsConst() &&
			val2.vc;
}

bool Descriptor::IsType_AnyAnyVal()						// true if "column op val"
{
	return 	val2.vc && val2.vc->IsConst() &&
			val1.vc;
}

bool Descriptor::IsType_AttrValOrAttrValVal() const		// true if "phys column op val or column between val and val or "
{														// or "phys_column IS NULL/NOT NULL"
	if(dynamic_cast<SingleColumn*>(attr.vc) == NULL)
		return false;
	return ((val1.vc && val1.vc->IsConst()) ||
			(op == O_IS_NULL || op == O_NOT_NULL))
			&&
			(!val2.vc || (val2.vc && val2.vc->IsConst()));
}

bool Descriptor::IsType_AttrMultiVal() const
{
	return 	op != O_BETWEEN && op != O_NOT_BETWEEN &&
		dynamic_cast<SingleColumn*>(attr.vc) != NULL &&
		dynamic_cast<MultiValColumn*>(val1.vc) &&
		val1.vc->IsConst();
}

bool Descriptor::IsType_AttrVal()						// true if "phys_column op val" or "phys_column IS NULL/NOT NULL"
{
	if(dynamic_cast<SingleColumn*>(attr.vc) == NULL)
		return false;
	return ((val1.vc && val1.vc->IsConst()) ||
			(op == O_IS_NULL || op == O_NOT_NULL))
			&& !val2.vc;
}

bool Descriptor::IsType_AttrValVal()						// true if "phys column between val and val
{
	return 	dynamic_cast<SingleColumn*>(attr.vc) != NULL &&
			val1.vc && val1.vc->IsConst() &&
			val2.vc && val2.vc->IsConst() ;
}


bool Descriptor::IsType_JoinSimple()							// true if more than one table involved
{
	// Make sure to use CalculateJoinType() before
	assert(desc_t != DT_NOT_KNOWN_YET);		
	return GetJoinType() == DT_SIMPLE_JOIN ;
}

bool Descriptor::IsType_AttrAttr()						// true if "column op column" (including between) from one table
{
	if(dynamic_cast<SingleColumn*>(attr.vc) == NULL ||
			dynamic_cast<SingleColumn*>(val1.vc) == NULL ||	val2.vc)
		return false;

	return true;
}

void Descriptor::CalculateJoinType()
{
	set<int> tables_a, tables_v1, tables_v2, tables_all;

	if(attr.vc) {
		tables_a = attr.vc->GetDimensions();
		tables_all.insert(tables_a.begin(), tables_a.end());
	}
	if(val1.vc) {
		tables_v1 = val1.vc->GetDimensions();
		tables_all.insert(tables_v1.begin(), tables_v1.end());
	}
	if(val2.vc) {
		tables_v2 = val2.vc->GetDimensions();
		tables_all.insert(tables_v2.begin(), tables_v2.end());
	}
	if(tables_all.size() <= 1)
		desc_t = DT_NON_JOIN;
	else if(tables_a.size() > 1 || tables_v1.size() > 1 || tables_v2.size() > 1 ||
		// 1 BETWEEN a1 AND b1
		(tables_a.size() == 0 && (op == O_BETWEEN || op == O_NOT_BETWEEN)) )
		desc_t = DT_COMPLEX_JOIN;
	else
		desc_t = DT_SIMPLE_JOIN;
}

//! put table aliases existing both in vc variables and table_aliases into tables
bool Descriptor::ExtractInvolvedTables(std::set<int>& tables, VirtualColumn* vc, vector<int>& table_aliases)
{
	if(!vc)
		return false;
	for(vector<VirtualColumn::VarMap>::const_iterator it = vc->GetVarMap().begin(); it != vc->GetVarMap().end(); it++)
		if(find(table_aliases.begin(), table_aliases.end(), (*it).var.tab) != table_aliases.end())
			tables.insert((*it).var.tab);
	if(dynamic_cast<ExpressionColumn*>(attr.vc) != NULL)
		return true;
	else
		return false;
}

// true if contains complex expression but there are no columns from different tables
// from table_aliases list (if not on the list = parameter of correleated subquery)
bool Descriptor::IsType_NonJoinComplex(vector<int> table_aliases)
{ // TODO: implement mechanism not to compute this each time
	bool ret ;
	if(attr.vc && dynamic_cast<MultiValColumn*>(attr.vc) != NULL)
		return false;
	if(val1.vc && dynamic_cast<MultiValColumn*>(val1.vc) != NULL)
		return false;
	if(val2.vc && dynamic_cast<MultiValColumn*>(val2.vc) != NULL)
		return false;
	std::set<int> tables; // identifiers of involved tables
	ret =  ExtractInvolvedTables(tables, attr.vc, table_aliases);
	ret |= ExtractInvolvedTables(tables, val1.vc, table_aliases);
	ret |= ExtractInvolvedTables(tables, val2.vc, table_aliases);

	if(tables.size() > 1)
		ret = false; //it is a join condition
	return ret;
}

// true if contains complex expressions involving columns from different tables
// from table_aliases list (if not on the list = parameter of correleated subquery)
bool Descriptor::IsType_JoinComplex(vector<int> table_aliases)
{
	std::set<int> tables; // identifiers of tables
	bool is_complex;

	is_complex =  ExtractInvolvedTables(tables, attr.vc, table_aliases);
	is_complex |= ExtractInvolvedTables(tables, val1.vc, table_aliases);
	is_complex |= ExtractInvolvedTables(tables, val2.vc, table_aliases);

	if(tables.size() > 1 && is_complex)
		return true;
	else
		return false;
}

bool Descriptor::IsType_JoinComplex()
{
	// Make sure to use CalculateJoinType() before
	assert(desc_t != DT_NOT_KNOWN_YET);		
	return GetJoinType() == DT_COMPLEX_JOIN ;
}

void Descriptor::Simplify(bool in_having)
{
	static MIIterator const mit(NULL);
	if (op == O_FALSE || op == O_TRUE)
		return;

	bool res = false;

	if(attr.vc && attr.vc->IsConst() &&
	   val1.vc && !val1.vc->IsConst()  &&
	   op != O_BETWEEN && op != O_NOT_BETWEEN && !IsSetOperator(op)) {
		SwitchSides();
	}
	if( (attr.vc && (!attr.vc->IsConst() || in_having && attr.vc->IsParameterized())) ||
		(val1.vc && (!val1.vc->IsConst() || in_having && val1.vc->IsParameterized())) ||
		(val2.vc && (!val2.vc->IsConst() || in_having && val2.vc->IsParameterized())) ) {
		return;
	}

	// from here attr, val1 and val2 (if exists) are constants

	VirtualColumn* acc = attr.vc;
	VirtualColumn* v1cc = val1.vc;
	VirtualColumn* v2cc = val2.vc;

	if( op == O_BETWEEN && (v1cc->IsNull(mit) || v2cc->IsNull(mit)) ) {
		op = O_FALSE;
		return;
	} else if( op == O_NOT_BETWEEN && (v1cc->IsNull(mit) || v2cc->IsNull(mit)) ) {
		if(v1cc->IsNull(mit) && v2cc->IsNull(mit)) {
			op = O_FALSE;
			return;
		}
		if(v1cc->IsNull(mit)) {		// a not between null and x  ==>  a > x
			op = O_MORE;
			val1 = val2;
			v1cc = v2cc;
		}
		else 						// a not between x and null  ==>  a < x
			op = O_LESS;
	}
	if(IsSetOperator(op)) {
		res = CheckSetCondition(mit, op);
		op = res ? O_TRUE : O_FALSE;
		return;
	}

	switch(op) {
		case O_IS_NULL:
			res = acc->IsNull(mit) ? true : false;
			break;
		case O_NOT_EXISTS:
		case O_EXISTS: {
			MultiValColumn* mvc = static_cast<MultiValColumn*>(attr.vc);
			res = mvc->NoValues(mit) == 0 ? false : true;
			if(op == O_NOT_EXISTS)
				res = !res;
			break;
		}
		case O_NOT_NULL:
			res = acc->IsNull(mit) ? false : true;
			break;
		case O_BETWEEN:
		case O_NOT_BETWEEN: {
			assert(sharp == false);
			BHReturnCode rc;

			if(acc->IsNull(mit))
				res = false;
			else if(ATI::IsDateTimeType(acc->TypeName()) || ATI::IsDateTimeType(v1cc->TypeName()) || ATI::IsDateTimeType(v2cc->TypeName())) {
				RCDateTime x, y, z;

				AttributeType t = acc->TypeName();
				rc = acc->GetValueDateTime(mit, x, t);
				if(rc != BHRC_SUCCESS)
					throw(DataTypeConversionRCException("Conversion to numeric type failed"));
				rc = v1cc->GetValueDateTime(mit, y, t);
				if(rc != BHRC_SUCCESS)
					throw(DataTypeConversionRCException("Conversion to numeric type failed"));
				rc = v2cc->GetValueDateTime(mit, z, t);
				if(rc != BHRC_SUCCESS)
					throw(DataTypeConversionRCException("Conversion to numeric type failed"));
				res = RCDataType::compare(x, y, O_MORE_EQ) && RCDataType::compare(x, z, O_LESS_EQ);
			} else if(ATI::IsNumericType(acc->TypeName()) || ATI::IsNumericType(v1cc->TypeName()) || ATI::IsNumericType(v2cc->TypeName())) {
				RCNum x, y, z;

				rc = acc->GetValueRCNum(mit, x);
				if(rc != BHRC_SUCCESS)
					throw(DataTypeConversionRCException("Conversion to numeric type failed"));
				rc = v1cc->GetValueRCNum(mit, y);
				if(rc != BHRC_SUCCESS)
					throw(DataTypeConversionRCException("Conversion to numeric type failed"));
				rc = v2cc->GetValueRCNum(mit, z);
				if(rc != BHRC_SUCCESS)
					throw(DataTypeConversionRCException("Conversion to numeric type failed"));

				res = RCDataType::compare(x, y, O_MORE_EQ) && RCDataType::compare(x, z, O_LESS_EQ);
			} else if(ATI::IsTxtType(acc->TypeName()) || ATI::IsTxtType(v1cc->TypeName()) || ATI::IsTxtType(v2cc->TypeName()))
				res = (strcmp(acc->GetValueStringPtr(mit) , v1cc->GetValueStringPtr(mit)) >= 0 && strcmp(acc->GetValueStringPtr(mit), v2cc->GetValueStringPtr(mit)) <= 0);
			else
				BHASSERT(0, "Non comparable data types");

			if(op == O_NOT_BETWEEN && !acc->IsNull(mit))
				res = !res;
			break;
		}
		default: {
			RCValueObject rv1 = acc->GetValue(mit);
			RCValueObject rv2 = v1cc->GetValue(mit);
			res = RCValueObject::compare(rv1, rv2, op);
		}
	}
	op = res ? O_TRUE : O_FALSE;
	return;
}

int SortDescriptor::operator==(const SortDescriptor& sec)
{
	return (dir==sec.dir) && (vc == sec.vc);
}

/**
Appends as string representation of this descriptor's operator
and data values to a buffer.

\param The destination buffer.
\param The capacity of the destination buffer.
*/
char* Descriptor::ToString(char buffer[], size_t buffer_size)
{
	/*
	Only the first few of the several steps needed to completely
	refactor this code were needed to resolve the valgrind issue.  The
	next steps in refactoring this code would be done via 'Replace
	Conditional with Polymorphism'.
	i.e. Create the operator objects with a common appendToString()
	member overriden to function as one of the groups below and
	by moving append*ToString() calls into the QueryOperator
	base/derived classes (the operator_object parameter will
	not be needed, etc).
	One last note, the Operator classes would need to be the objects
	that are sentient of the relationships with the objects around them.
	*/
	auto_ptr<const QueryOperator> operator_object(CreateQueryOperator(op));

	switch(operator_object->GetType()) {
		case O_ESCAPE:
		case O_EXISTS:
		case O_NOT_EXISTS:
		default: {
			size_t offset = strlen(buffer);
			AppendString(buffer, buffer_size, "(complex cond.)", 15, offset);
			break;
		}
		case O_FALSE:
		case O_TRUE:
			AppendConstantToString(buffer, buffer_size, operator_object.get());
			break;

		case O_IS_NULL:
		case O_NOT_NULL:
			AppendUnaryOperatorToString(buffer, buffer_size, operator_object.get());
			break;

		case O_EQ:
		case O_EQ_ALL:
		case O_EQ_ANY:
		case O_LESS:
		case O_LESS_ALL:
		case O_LESS_ANY:
		case O_MORE:
		case O_MORE_ALL:
		case O_MORE_ANY:
		case O_LESS_EQ:
		case O_LESS_EQ_ALL:
		case O_LESS_EQ_ANY:
		case O_MORE_EQ:
		case O_MORE_EQ_ALL:
		case O_MORE_EQ_ANY:
		case O_NOT_EQ:
		case O_NOT_EQ_ALL:
		case O_NOT_EQ_ANY:
		case O_IN:
		case O_NOT_IN:
		case O_LIKE:
		case O_NOT_LIKE:
			AppendBinaryOperatorToString(buffer, buffer_size, operator_object.get());
			break;

		case O_BETWEEN:
		case O_NOT_BETWEEN:
			AppendTernaryOperatorToString(buffer, buffer_size, operator_object.get());
			break;
	}
	if( !IsInner() ) {
		sprintf(buffer, "%s (outer %d)", buffer, outer_dim);
	}

	return buffer;
}

/**
Creates a query operator of the requested type.

Interim implementation until an operator class hierarchy
is put in place.  This call is only a stop-gap measure
to that effect.

\param The Operator enumeration of the desired type of query operator.
*/
const QueryOperator * Descriptor::CreateQueryOperator(Operator type) const
{
	const char* string_rep[OPERATOR_ENUM_COUNT] = {
		"=",						// O_EQ
		"=ALL",						// O_EQ_ALL
		"=ANY",						// O_EQ_ANY
		"<>",						// O_NOT_EQ
		"<>ALL",					// O_NOT_EQ_ALL
		"<>ANY",					// O_NOT_EQ_ANY
		"<",						// O_LESS
		"<ALL",						// O_LESS_ALL
		"<ANY",						// O_LESS_ANY
		">",						// O_MORE
		">ALL",			            // O_MORE_ALL
		">ANY",				        // O_MORE_ANY
		"<=",						// O_LESS_EQ
		"<=ALL",					// O_LESS_EQ_ALL
		"<=ANY",					// O_LESS_EQ_ANY
		">=",						// O_MORE_EQ
		">=ALL",					// O_MORE_EQ_ALL
		">=ANY",					// O_MORE_EQ_ANY
		"IS NULL",					// O_IS_NULL
		"IS NOT NULL",				// O_NOT_NULL
		"BET.",						// O_BETWEEN
		"NOT BET.",					// O_NOT_BETWEEN
		"LIKE",						// O_LIKE
		"NOT LIKE",					// O_NOT_LIKE
		"IN",						// O_IN
		"NOT IN",					// O_NOT_IN
		"EXISTS",					// O_EXISTS
		"NOT EXISTS",				// O_NOT_EXISTS
		"FALSE",					// O_FALSE
		"TRUE",						// O_TRUE
		"ESCAPE"					// O_ESCAPE
	};

	return new QueryOperator(type, string_rep[type]);
}

/**
Simple string concatenation into a buffer.

\param The destination buffer.
\param The capacity of the destination buffer.
\param The string to concatenate.
\param The length of the string to be concatenated.
\param The destination offset within buffer for the concatenation.
*/
void Descriptor::AppendString(char *buffer, size_t buffer_size,
							  const char *string, size_t string_length, size_t offset) const
{
	if((offset + string_length) > buffer_size) {
		throw InternalRCException("Bounds Check in appendString");
	}
	strcpy(buffer + offset, string);
}

/**
Appends a string representation of a constant operator into a buffer.

\param The destination buffer.
\param The capacity of the destination buffer.
\param The QueryOperator object managed by this Descriptor object.
*/
void Descriptor::AppendConstantToString(char buffer[], size_t size,
										const QueryOperator *operator_object) const
{
	const std::string &constant = operator_object->AsString();

	size_t offset = strlen(buffer);

	AppendString(buffer, size, " ", 1, offset++);
	AppendString(buffer, size, constant.data(), constant.length(), offset);
}

/**
Appends a string representation of a unary operator (attr op)
into a buffer.

\param The destination buffer.
\param The capacity of the destination buffer.
\param The QueryOperator object managed by this Descriptor object.
*/
void Descriptor::AppendUnaryOperatorToString(char buffer[], size_t size,
											 const QueryOperator *operator_object) const
{
	attr.ToString(buffer, size, 0);
	AppendConstantToString(buffer, size, operator_object);
}

/**
Appends a string representation of a binary operator (attr op val1)
into a buffer.

\param The destination buffer.
\param The capacity of the destination buffer.
\param The QueryOperator object managed by this Descriptor object.
*/
void Descriptor::AppendBinaryOperatorToString(char buffer[], size_t size,
											  const QueryOperator* operator_object) const
{
	AppendUnaryOperatorToString(buffer, size, operator_object);
	AppendString(buffer, size, " ", 1, strlen(buffer));
	val1.ToString(buffer, size, 0);
}

/**
Appends a string representation of a special-case ternary
operator into a buffer.

\param The destination buffer.
\param The capacity of the destination buffer.
\param The QueryOperator object managed by this Descriptor object.
*/
void Descriptor::AppendTernaryOperatorToString(char buffer[], size_t size,
											   const QueryOperator* operator_object) const
{
	AppendBinaryOperatorToString(buffer, size, operator_object);
	size_t offset = strlen(buffer);
	AppendString(buffer, size, " AND ", 5, offset);
	val2.ToString(buffer, size, 0);
	if(sharp)
	{
		offset = strlen(buffer);
		AppendString(buffer, size, " (sharp)", 8, offset);
	}
}

bool Descriptor::DimensionUsed(int dim)
{
	set<int> dims;
	if(attr.vc) {
		dims = attr.vc->GetDimensions();
		if(dims.find(dim) != dims.end())
			return true;
	}
	if(val1.vc) {
		dims = val1.vc->GetDimensions();
		if(dims.find(dim) != dims.end())
			return true;
	}
	if(val2.vc) {
		dims = val2.vc->GetDimensions();
		if(dims.find(dim) != dims.end())
			return true;
	}
	return false;
}

void Descriptor::DimensionUsed(DimensionVector &dims)
{
	if(attr.vc)
		attr.vc->MarkUsedDims(dims);
	if(val1.vc)
		val1.vc->MarkUsedDims(dims);
	if(val2.vc)
		val2.vc->MarkUsedDims(dims);
}

void Descriptor::LockSourcePacks(const MIIterator& mit)
{
	if(attr.vc)
		attr.vc->LockSourcePacks(mit);
	if(val1.vc)
		val1.vc->LockSourcePacks(mit);
	if(val2.vc)
		val2.vc->LockSourcePacks(mit);
}

void Descriptor::UnlockSourcePacks()
{
	if(attr.vc)
		attr.vc->UnlockSourcePacks();
	if(val1.vc)
		val1.vc->UnlockSourcePacks();
	if(val2.vc)
		val2.vc->UnlockSourcePacks();
}

void Descriptor::EvaluatePack(MIUpdatingIterator& mit)
{
	AttributeType attr_t = attr.vc ? attr.vc->TypeName() : AttributeType();
	AttributeType val1_t = val1.vc && !attr.vc->Type().IsLookup() ? val1.vc->TypeName() : AttributeType();
	AttributeType val2_t = val2.vc && !attr.vc->Type().IsLookup() ? val2.vc->TypeName() : AttributeType();
	if(attr.vc && val1.vc) {
		if(!AreComparable(attr_t, val1_t) || (val2.vc && !AreComparable(attr_t, val2_t)))
			throw NotImplementedRCException("Parameters are not comparable.");
	}
	LockSourcePacks(mit);

	// Check if we can delegate evaluation of descriptor to physical column
	if(encoded)
		attr.vc->EvaluatePack(mit, *this);
	else if(RequiresUTFConversions(collation)) {
		InitUTFBuffers();
		while(mit.IsValid()) {
			if(CheckCondition_UTF(mit) == false)
				mit.ResetCurrent();
			++mit;
			if(mit.PackrowStarted())
				break;
		}
	} else {
		while(mit.IsValid()) {
			if(CheckCondition(mit) == false)
				mit.ResetCurrent();
			++mit;
			if(mit.PackrowStarted())
				break;
		}
	}
}

bool Descriptor::CheckCondition_UTF(const MIIterator& mit)
{
	if(op == O_TRUE)
		return true;
	else if(op == O_FALSE)
		return false;

	bool result = true;

	AttributeType attr_t = attr.vc ? attr.vc->TypeName() : AttributeType();
	AttributeType val1_t = val1.vc ? val1.vc->TypeName() : AttributeType();
	AttributeType val2_t = val2.vc ? val2.vc->TypeName() : AttributeType();
	if(attr.vc && val1.vc) {
		if(!AreComparable(attr_t, val1_t) || (val2.vc && !AreComparable(attr_t, val2_t)))
			throw NotImplementedRCException("Parameters are not comparable.");
	}
	// Assumption: LockSourcePaxternally done.
	if(op == O_EQ) {		// fast track for the most common operator
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(attr.vc && val1.vc && RequiresUTFConversions(collation));
		if(attr.vc->IsNull(mit) || val1.vc->IsNull(mit))
			return false;
		RCBString s1, s2;
		attr.vc->GetValueString(s1, mit);
		val1.vc->GetValueString(s2, mit);
		return CollationStrCmp(collation, s1, s2) == 0;
	} else if(op == O_NOT_NULL) {
		if(attr.vc->IsNull(mit))
			return false;
	} else if(op == O_IS_NULL) {
		if(!attr.vc->IsNull(mit))
			return false;
	} else if(op == O_EXISTS || op == O_NOT_EXISTS) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(dynamic_cast<SubSelectColumn*>(attr.vc));
		SubSelectColumn* sub = static_cast<SubSelectColumn*>(attr.vc);
		//sub->SetExistMode();
		bool is_empty = sub->IsEmpty(mit);
		if((op == O_EXISTS && is_empty) || (op == O_NOT_EXISTS && !is_empty))
			return false;
	} else if(op == O_BETWEEN || op == O_NOT_BETWEEN) {
		if(attr.vc->IsNull(mit))
			return false;
		// need to consider three value logic
		RCBString s1, s2, s3;
		attr.vc->GetValueString(s1, mit);
		val1.vc->GetValueString(s2, mit);
		val2.vc->GetValueString(s3, mit);
		bool attr_ge_val1 = CollationStrCmp(collation, s1, s2) >= 0;
		bool attr_le_val2 = CollationStrCmp(collation, s1, s3) <= 0;
		BHTribool val1_res = val1.vc->IsNull(mit) ? BHTRIBOOL_UNKNOWN : BHTribool(attr_ge_val1);
		BHTribool val2_res = val2.vc->IsNull(mit) ? BHTRIBOOL_UNKNOWN : BHTribool(attr_le_val2);
		if(op == O_BETWEEN) {
			if(val1_res != true || val2_res != true)
				return false;
		} else {
			if(val1_res != false && val2_res != false)
				return false;
		}
	} else if(IsSetOperator(op)) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(attr.vc && dynamic_cast<MultiValColumn*>(val1.vc));
		//if(attr.vc->IsNull(mit))
		//	return false;
		return CheckSetCondition_UTF(mit, op);
	} else if(op == O_LIKE || op == O_NOT_LIKE) {
		if(attr.vc->IsNull(mit) || val1.vc->IsNull(mit))
			return false;
		RCBString v, pattern;
		attr.vc->GetValueString(v, mit);
		val1.vc->GetValueString(pattern, mit);
		int x = collation.collation->coll->wildcmp(collation.collation, 
			v.val, v.val + v.len, 
			pattern.val, pattern.val + pattern.len, 
			'\\', '_', '%'); 
		result = (x == 0 ? true : false);
		if(op == O_LIKE)	
			return result;
		else
			return !result;
	} else { // all other logical operators: >, >=, <, <=
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(attr.vc && val1.vc);
		if(attr.vc->IsNull(mit) || val1.vc->IsNull(mit))
			return false;
		RCBString s1, s2;
		attr.vc->GetValueString(s1, mit);
		val1.vc->GetValueString(s2, mit);
		if(!CollationStrCmp(collation, s1, s2, op))
			return false;
		//ConvertToBinaryForm(attr.vc, mit, buf_attr, collation);
		//ConvertToBinaryForm(val1.vc, mit, buf_val1, collation);
		//if(!RCValueObject::compare(buf_attr, buf_val1, op))
		//	return false;
	}
	return result;
}

bool Descriptor::CheckCondition(const MIIterator& mit)
{
	if(op == O_TRUE)
		return true;
	else if(op == O_FALSE)
		return false;

	bool result = true;

	AttributeType attr_t = attr.vc ? attr.vc->TypeName() : AttributeType();
	AttributeType val1_t = val1.vc ? val1.vc->TypeName() : AttributeType();
	AttributeType val2_t = val2.vc ? val2.vc->TypeName() : AttributeType();
	if(attr.vc && val1.vc) {
		if(!AreComparable(attr_t, val1_t) || (val2.vc && !AreComparable(attr_t, val2_t)))
			throw NotImplementedRCException("Parameters are not comparable.");
	}
	// Assumption: LockSourcePacks externally done.
	if(op == O_EQ) {		// fast track for the most common operator
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(attr.vc && val1.vc);
		// nulls checked in operator ==
		if(!(attr.vc->GetValue(mit) == val1.vc->GetValue(mit)))
			return false;
	}/* else if(op == O_IN || op == O_NOT_IN) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(dynamic_cast<MultiValColumn*>(val1.vc));
		BHTribool res = static_cast<MultiValColumn*>(val1.vc)->Contains(mit, attr.vc->GetValue(mit, false));
		if(op == O_NOT_IN)
			res = !res;
		result = (res == true);
	}*/ else if(op == O_NOT_NULL) {
		if(attr.vc->IsNull(mit))
			return false;
	} else if(op == O_IS_NULL) {
		if(!attr.vc->IsNull(mit))
			return false;
	} else if(op == O_EXISTS || op == O_NOT_EXISTS) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(dynamic_cast<SubSelectColumn*>(attr.vc));
		SubSelectColumn* sub = static_cast<SubSelectColumn*>(attr.vc);
		//sub->SetExistMode();
		bool is_empty = sub->IsEmpty(mit);
		if((op == O_EXISTS && is_empty) || (op == O_NOT_EXISTS && !is_empty))
			return false;
	} else if(op == O_BETWEEN || op == O_NOT_BETWEEN) {
		if(attr.vc->IsNull(mit))
			return false;
		RCValueObject rcvo1 = attr.vc->GetValue(mit, false);
		// need to consider three value logic
		BHTribool val1_res = val1.vc->IsNull(mit) ? BHTRIBOOL_UNKNOWN : BHTribool(rcvo1 >= val1.vc->GetValue(mit, false));
		BHTribool val2_res = val2.vc->IsNull(mit) ? BHTRIBOOL_UNKNOWN : BHTribool(rcvo1 <= val2.vc->GetValue(mit, false));
		if(op == O_BETWEEN) {
			if(val1_res != true || val2_res != true)
				return false;
		} else {
			if(val1_res != false && val2_res != false)
				return false;
		}
	} else if(IsSetOperator(op)) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(attr.vc && dynamic_cast<MultiValColumn*>(val1.vc));
		//if(attr.vc->IsNull(mit))
		//	return false;
		return CheckSetCondition(mit, op);
	} else { // all other logical operators: >, >=, <, <=
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(attr.vc && val1.vc);
		if(attr.vc->IsNull(mit) || val1.vc->IsNull(mit))
			return false;
		if(!RCValueObject::compare(attr.vc->GetValue(mit), val1.vc->GetValue(mit), op))
			return false;
	}
	return result;
}

bool Descriptor::CheckSetCondition_UTF(const MIIterator& mit, Operator op)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(IsSetOperator(op));
	bool result = true;
	MultiValColumn* mvc = static_cast<MultiValColumn*>(val1.vc);
	RCBString s1;
	attr.vc->GetValueString(s1, mit);
	RCValueObject aggr;
	switch(op) {
		case O_EQ_ALL:
			for(MultiValColumn::Iterator it = mvc->begin(mit), end = mvc->end(mit); it != end; ++it) {
				RCBString s2 = it->GetString();
				//ConvertToBinaryForm(it->GetString(), buf_val1, buf_val1_len, collation.collation, false);
				if(s1.IsNull() || it->IsNull() || CollationStrCmp(collation, s1, s2) != 0)
					return false;
			}
			break;
		case O_IN:
		case O_EQ_ANY:
			if(s1.IsNull())
				result = false;
			else
				result = (mvc->Contains(mit, s1) == true);
			break;
		case O_NOT_IN:
		case O_NOT_EQ_ALL: {
			if((s1.IsNull() && mvc->NoValues(mit) != 0)) {
				result = false;
				break;
			}
			BHTribool res = mvc->Contains(mit, s1);
			res = !res;
			result = (res == true);
			break;
						   }
		case O_NOT_EQ_ANY:
			result = false;
			if(!s1.IsNull()) {
				for(MultiValColumn::Iterator it = mvc->begin(mit), end = mvc->end(mit); it != end; ++it) {
					//ConvertToBinaryForm(it->GetString(), buf_val1, buf_val1_len, collation.collation, false);
					if(!it->IsNull() && CollationStrCmp(collation, s1, it->GetString()) != 0)
						return true;
				}
			}
			break;
		case O_LESS_ALL:
		case O_LESS_EQ_ALL:
			Query::UnmarkAllAny(op);
			aggr = mvc->GetSetMin(mit);
			if(mvc->NoValues(mit) == 0)
				result = true; // op ALL (empty_set) is TRUE
			else if(s1.IsNull() || aggr.IsNull() || mvc->ContainsNull(mit) || !CollationStrCmp(collation, s1, aggr, op))
				result = false;
			break;
		case O_MORE_ANY:
		case O_MORE_EQ_ANY:
			Query::UnmarkAllAny(op);
			aggr = mvc->GetSetMin(mit);
			if(mvc->NoValues(mit) == 0)
				result = false; // op ANY (empty_set) is FALSE
			else if(s1.IsNull() || aggr.IsNull() || !CollationStrCmp(collation, s1, aggr, op))
				result = false;
			break;
		case O_LESS_ANY:
		case O_LESS_EQ_ANY:
			Query::UnmarkAllAny(op);
			aggr = mvc->GetSetMax(mit);
			if(mvc->NoValues(mit) == 0)
				result = false; // op ANY (empty_set) is FALSE
			else if(s1.IsNull() || aggr.IsNull() || !CollationStrCmp(collation, s1, aggr, op))
				result = false;
			break;
		case O_MORE_ALL:
		case O_MORE_EQ_ALL:
			Query::UnmarkAllAny(op);
			aggr = mvc->GetSetMax(mit);
			if(mvc->NoValues(mit) == 0)
				result = true; // op ALL (empty_set) is TRUE
			else if(s1.IsNull() || aggr.IsNull() || mvc->ContainsNull(mit) || !CollationStrCmp(collation, s1, aggr, op))
				result = false;
			break;
	}
	return result;
}

bool Descriptor::CheckSetCondition(const MIIterator& mit, Operator op)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(IsSetOperator(op));
	bool result = true;
	MultiValColumn* mvc = static_cast<MultiValColumn*>(val1.vc);
	RCValueObject val = attr.vc->GetValue(mit);
	RCValueObject aggr;
	switch(op) {
		case O_EQ_ALL:
			for(MultiValColumn::Iterator it = mvc->begin(mit), end = mvc->end(mit); it != end; ++it) {
				if(val.IsNull() || it->IsNull() || !RCValueObject::compare(val, it->GetValue(), O_EQ))
					return false;
			}
			break;
		case O_IN:
		case O_EQ_ANY:
			if(val.IsNull())
				result = false;
			else
				result = (mvc->Contains(mit, *val) == true);
			break;
		case O_NOT_IN:
		case O_NOT_EQ_ALL: {
			if((val.IsNull() && mvc->NoValues(mit) != 0)) {
				result = false;
				break;
			}
			BHTribool res = mvc->Contains(mit, *val);
			res = !res;
			result = (res == true);
			break;
		}
		case O_NOT_EQ_ANY:
			result = false;
			if(!val.IsNull()) {
				for(MultiValColumn::Iterator it = mvc->begin(mit), end = mvc->end(mit); it != end; ++it) {
					if(!it->IsNull() && RCValueObject::compare(val, it->GetValue(), O_NOT_EQ))
						return true;
				}
			}
			break;
		case O_LESS_ALL:
		case O_LESS_EQ_ALL:
			Query::UnmarkAllAny(op);
			aggr = mvc->GetSetMin(mit);
			if(mvc->NoValues(mit) == 0)
				result = true; // op ALL (empty_set) is TRUE
			else if(val.IsNull() || aggr.IsNull() || mvc->ContainsNull(mit) || !RCValueObject::compare(val, aggr, op))
				result = false;
			break;
		case O_MORE_ANY:
		case O_MORE_EQ_ANY:
			Query::UnmarkAllAny(op);
			aggr = mvc->GetSetMin(mit);
			if(mvc->NoValues(mit) == 0)
				result = false; // op ANY (empty_set) is FALSE
			else if(val.IsNull() || aggr.IsNull() || !RCValueObject::compare(val, aggr, op))
				result = false;
			break;
		case O_LESS_ANY:
		case O_LESS_EQ_ANY:
			Query::UnmarkAllAny(op);
			aggr = mvc->GetSetMax(mit);
			if(mvc->NoValues(mit) == 0)
				result = false; // op ANY (empty_set) is FALSE
			else if(val.IsNull() || aggr.IsNull() || !RCValueObject::compare(val, aggr, op))
				result = false;
			break;
		case O_MORE_ALL:
		case O_MORE_EQ_ALL:
			Query::UnmarkAllAny(op);
			aggr = mvc->GetSetMax(mit);
			if(mvc->NoValues(mit) == 0)
				result = true; // op ALL (empty_set) is TRUE
			else if(val.IsNull() || aggr.IsNull() || mvc->ContainsNull(mit) || !RCValueObject::compare(val, aggr, op))
				result = false;
			break;
	}
	return result;
}

void Descriptor::CoerceColumnType(VirtualColumn*& for_typecast)
{
	VirtualColumn* vc = attr.vc;
	TypeCastColumn* tcc = NULL;
	if(ATI::IsNumericType(vc->TypeName())) {
		if(ATI::IsTxtType(for_typecast->TypeName()))
			tcc = new String2IntCastColumn(for_typecast, vc->Type());
		else if(ATI::IsDateTimeType(for_typecast->TypeName()))
			tcc = new DateTime2IntCastColumn(for_typecast, vc->Type());
	} else if(ATI::IsDateTimeType(vc->TypeName())) {
		if(ATI::IsTxtType(for_typecast->TypeName()))
			tcc = new String2DateTimeCastColumn(for_typecast, vc->Type());
		else if(ATI::IsNumericType(for_typecast->TypeName()))
			tcc = new Int2DateTimeCastColumn(for_typecast, vc->Type());
	} else if(ATI::IsTxtType(vc->TypeName())) {
		if(ATI::IsDateTimeType(for_typecast->TypeName()))
			tcc = new DateTime2VarcharCastColumn(for_typecast, vc->Type());
		if(ATI::IsNumericType(for_typecast->TypeName()))
			tcc = new Num2VarcharCastColumn(for_typecast, vc->Type());
	} else
		return;

	if(tcc) {
		if(rccontrol.isOn())
			rccontrol.lock() << "Type conversion for VC:" << (for_typecast == val1.vc ? val1.vc_id : for_typecast == val2.vc ? val2.vc_id : val1.vc_id ) << unlock;
		table->AddVirtColumn(tcc);
		for_typecast = tcc;
	}
}

void Descriptor::CoerceColumnTypes() {

	if(val1.vc)
		if (MultiValColumn* mvc = dynamic_cast<MultiValColumn*>(val1.vc)) {
			if (InSetColumn* isc = dynamic_cast<InSetColumn*>(val1.vc)) {
				for(uint i = 0; i < isc->columns.size(); i++)
					CoerceColumnType(isc->columns[i]);
			}
			mvc->SetExpectedType(attr.vc->Type());
		} else
			CoerceColumnType(val1.vc);
	if(val2.vc)
		CoerceColumnType(val2.vc);
	CoerceCollation();
}

//void Descriptor::CopyVC(VirtualColumn*& target, VirtualColumn* source_c, Descriptor& source_d)
//{
//	if(source_d.IsLocalVC(source_c)) {
//		if (source_c->GetStatus() == VirtualColumn::VC_CONST && dynamic_cast<ConstColumn*>(source_c)) {  //TypeCastColumn not handled
//			VirtualColumn* vc = new ConstColumn(*(ConstColumn*)source_c);
//			vc_local.push_back(vc);
//			target = vc;
//		} else
//			assert(0 && "can copy only CONST columns");
//	} else
//		target = source_c;
//}
//
//

bool Descriptor::NullMayBeTrue()		// true, if the descriptor may give nontrivial answer if any of involved dimension is null
{
	if(op == O_IS_NULL || op == O_NOT_NULL)
		return true;

	// TODO: more precise conditions
	// Examples:
	//    (a is null) = 1
	//    b = ifnull(a, 0)
	//    c > (case when a is null then 10 else 20)
	//    d = (select count(*) from t where t.x = a)
	//    e IN (1, 2, a)
	//    f < ALL (1, 2, a)

	// For now, a simplistic version: any complex case is true.
	if(attr.vc && dynamic_cast<SingleColumn*>(attr.vc) == NULL && !attr.vc->IsConst())
		return true;
	if(val1.vc && dynamic_cast<SingleColumn*>(val1.vc) == NULL && !val1.vc->IsConst())
		return true;
	if(val2.vc && dynamic_cast<SingleColumn*>(val2.vc) == NULL && !val2.vc->IsConst())
		return true;
	return false;
}

bool Descriptor::IsParameterized() const
{
	return ( (attr.vc && attr.vc->IsParameterized()) ||
			 (val1.vc && val1.vc->IsParameterized()) ||
			 (val2.vc && val2.vc->IsParameterized()) );
}

bool Descriptor::IsDeterministic() const
{
	bool det = true;
	if(attr.vc)
		det = det && attr.vc->IsDeterministic();
	if(val1.vc)
		det = det && val1.vc->IsDeterministic();
	if(val2.vc)
		det = det && val2.vc->IsDeterministic();
	return det;
}

bool IsUnicode(DTCollation coll) 
{
	return (strcmp(coll.collation->csname, "utf8") == 0 || strcmp(coll.collation->csname, "ucs2") == 0);
}

bool IsBinary(DTCollation coll) 
{
	return (strcmp(coll.collation->csname, "binary") == 0);
}

extern bool use_charset;

void Descriptor::CoerceCollation()
{
	collation = attr.vc && attr.vc->Type().IsString() ? attr.vc->GetCollation() : DTCollation();
	if(!use_charset)
		return;
	if(val1.vc && val1.vc->GetCollation().collation != collation.collation && val1.vc->GetCollation().derivation <= collation.derivation &&
		val1.vc->Type().IsString()) {
		if(val1.vc->GetCollation().derivation < collation.derivation)
			collation = val1.vc->GetCollation();
		else { // equal coercibility
			if(IsUnicode(collation) && !IsUnicode(val1.vc->GetCollation())) {
				// do nothing: unicode has higher precedence 
			} else if(!IsUnicode(collation) && IsUnicode(val1.vc->GetCollation()))
				collation = val1.vc->GetCollation();
			else if(IsBinary(collation) && !IsBinary(val1.vc->GetCollation())) {
				// do nothing: binary has higher precedence
			} else if(!IsBinary(collation) && IsBinary(val1.vc->GetCollation()))
				collation = val1.vc->GetCollation();
			else {
				BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"Error: Incompatible collations!");
			}
		}
	}
	if(val2.vc && val2.vc->GetCollation().collation != collation.collation && val2.vc->GetCollation().derivation <= collation.derivation &&
		val2.vc->Type().IsString()) {
		if(val2.vc->GetCollation().derivation < collation.derivation)
			collation = val2.vc->GetCollation();
		else { // equal coercibility
			if(IsUnicode(collation) && !IsUnicode(val2.vc->GetCollation())) {
				// do nothing: unicode has higher precedence 
			} else if(!IsUnicode(collation) && IsUnicode(val2.vc->GetCollation()))
				collation = val2.vc->GetCollation();
			else if(IsBinary(collation) && !IsBinary(val2.vc->GetCollation())) {
				// do nothing: binary has higher precedence
			} else if(!IsBinary(collation) && IsBinary(val2.vc->GetCollation()))
				collation = val2.vc->GetCollation();
			else {
				BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"Error: Incompatible collations!");
			}
		}
	}
}

void Descriptor::InitUTFBuffers()
{
	int buf_attr_len = 0;
	int buf_val1_len = 0;
	if(attr.vc)
		buf_attr_len = CollationBufLen(collation, attr.vc->MaxStringSize());
	if(val1.vc)
		buf_val1_len = CollationBufLen(collation, val1.vc->MaxStringSize());
	buf_len = max(buf_attr_len, buf_val1_len);
	if(attr.vc)
		buf_attr = RCBString(NULL, buf_len, true);
	if(val1.vc)
		buf_val1 = RCBString(NULL, buf_len, true);
}

bool IsSetOperator(Operator op)
{
	return IsSetAnyOperator(op) || IsSetAllOperator(op) || op == O_IN || op == O_NOT_IN;
}

bool IsSetAllOperator(Operator op)
{
	return
		op == O_EQ_ALL ||
		op == O_NOT_EQ_ALL ||
		op == O_LESS_ALL ||
		op == O_MORE_ALL ||
		op == O_LESS_EQ_ALL ||
		op == O_MORE_EQ_ALL;
}

bool IsSetAnyOperator(Operator op)
{
	return
		op == O_EQ_ANY ||
		op == O_NOT_EQ_ANY ||
		op == O_LESS_ANY ||
		op == O_MORE_ANY ||
		op == O_LESS_EQ_ANY ||
		op == O_MORE_EQ_ANY;
}

bool ISTypeOfEqualOperator(Operator op)
{
	return op == O_EQ || op == O_EQ_ALL || op == O_EQ_ANY;
}

bool ISTypeOfNotEqualOperator(Operator op)
{
	return op == O_NOT_EQ || op == O_NOT_EQ_ALL || op == O_NOT_EQ_ANY;
}

bool ISTypeOfLessOperator(Operator op)
{
	return op == O_LESS || op == O_LESS_ALL || op == O_LESS_ANY;
}

bool ISTypeOfLessEqualOperator(Operator op)
{
	return op == O_LESS_EQ || op == O_LESS_EQ_ALL || op == O_LESS_EQ_ANY;
}

bool ISTypeOfMoreOperator(Operator op)
{
	return op == O_MORE || op == O_MORE_ALL || op == O_MORE_ANY;
}


bool ISTypeOfMoreEqualOperator(Operator op)
{
	return op == O_MORE_EQ || op == O_MORE_EQ_ALL || op == O_MORE_EQ_ANY;
}
