#ifndef _DESCRIPTOR_H_
#define _DESCRIPTOR_H_

#include "common/CommonDefinitions.h"
#include "CQTerm.h"

class ConstColumn;
class MIIterator;
class DimensionVector;
class Query;
class TempTable;

/*! DT_NON_JOIN - 1 dimension condition
 * 	DT_SIMPLE_JOIN - suitable for e.g. hash join
 *  DT_COMPLEX_JOIN - possibly requires nested loop join
 */
enum DescriptorJoinType	{ DT_NON_JOIN, DT_SIMPLE_JOIN, DT_COMPLEX_JOIN, DT_NOT_KNOWN_YET };

class Descriptor
{
public:
	CQTerm				attr;		// descriptor is usually "attr op val1 [val2]", e.g. "a = 7", "a BETWEEN 8 AND 12"
    Operator 			op;
    CQTerm 				val1;
    CQTerm 				val2;

    LogicalOperator		lop;		// OR is not implemented on some lower levels! Only valid in HAVING tree.
    bool 				sharp;		// used to indicate that BETWEEN contains sharp inequalities
    bool 				encoded;
    bool 				done;
    double 				evaluation;	// approximate weight ("hardness") of the condition; execution order is based on it
    int					outer_dim;	// this dimension is on the right side of LEFT JOIN, or -1
    bool				delayed;	// descriptor is delayed (usually because of outer joins) tbd after joins
    TempTable* 			table;

    Descriptor(TempTable* t);
    ~Descriptor();
    Descriptor(const Descriptor & desc);
    Descriptor & operator =(const Descriptor & desc);
    void swap(Descriptor& d);

    int operator ==(const Descriptor & sec);
    bool EqualExceptOuter(const Descriptor& sec);		// descriptors are equal, but one may be inner and other may be outer
    void Simplify(bool in_having = false);
    void SwitchSides();
    bool IsTrue() { return op == O_TRUE; }
    bool IsFalse() { return op == O_FALSE; }

    void EvaluatePack(MIUpdatingIterator& mit);	// Assumption: no locking needed, done inside

	void InitUTFBuffers();

	bool CheckCondition_UTF(const MIIterator& mit);
	bool CheckCondition(const MIIterator& mit);	// Assumption: LockSourcePacks done externally.

    void LockSourcePacks(const MIIterator& mit);
    void UnlockSourcePacks();

    bool DimensionUsed(int dim);
    void DimensionUsed(DimensionVector &dims);

    bool IsOneDimensional();
    std::set<int> GetDimensions();
    bool IsDeterministic() const;
    bool IsType_JoinSimple();
    bool IsType_AttrAttr();
    bool IsType_AttrVal();			// note: "attr IS NULL" is also regarded as AttrVal.
    bool IsType_AttrValVal();
    bool IsType_AttrValOrAttrValVal() const;
    bool IsType_AttrMultiVal() const;
    bool IsType_AnyVal();
    bool IsType_AnyValAny();
    bool IsType_AnyAnyVal();

	// Note: CalculateJoinType() must be executed before checking a join type
	void CalculateJoinType();
	DescriptorJoinType GetJoinType() const 	{ return desc_t; }
    bool IsType_Complex(std::vector<int> table_aliases) {
        return IsType_JoinComplex(table_aliases) || IsType_NonJoinComplex(table_aliases);
    }
    bool IsType_JoinComplex(std::vector<int> table_aliases);
    bool IsType_JoinComplex();
    bool IsType_NonJoinComplex(std::vector<int> table_aliases);
    bool IsType_Join()					{ return (IsType_JoinSimple() || IsType_JoinComplex()); }

    bool IsDelayed() const				{ return delayed; }
	bool IsParameterized() const;
    bool IsInner()						{ return outer_dim == -1; } // join_type == JO_INNER; }
    bool IsOuter()						{ return outer_dim != -1; } // (IsLeftJoin() || IsRightJoin() || IsFullOuterJoin()); }

    char* ToString(char buf[], size_t buf_ct);
    void CoerceColumnTypes();

    bool NullMayBeTrue();				// true, if the descriptor may give nontrivial answer if any of involved dimension is null

    const QueryOperator *CreateQueryOperator(Operator type) const;
    void CoerceColumnType(VirtualColumn*& to_be_casted);
	DTCollation GetCollation() { return collation; }
private:
	/*! \brief Checks condition for set operator, e.g., <ALL
	* \pre LockSourcePacks done externally.
	* \param mit - iterator on MultiIndex
	* \param op - operator to check
	* \return bool
	*/
	bool CheckSetCondition(const MIIterator& mit, Operator op);
	bool CheckSetCondition_UTF(const MIIterator& mit, Operator op);

	void AppendString(char *buffer, size_t bufferSize, const char *string, size_t stringLength, size_t offset) const;
    void AppendConstantToString(char buffer[], size_t size, const QueryOperator *operatorObject) const;
    void AppendUnaryOperatorToString(char buffer[], size_t size, const QueryOperator *operatorObject) const;
    void AppendBinaryOperatorToString(char buffer[], size_t size, const QueryOperator *operatorObject) const;
    void AppendTernaryOperatorToString(char buffer[], size_t size, const QueryOperator *operatorObject) const;
    bool ExtractInvolvedTables(std::set<int> & tables, VirtualColumn *vc, std::vector<int> & table_aliases);
	void CoerceCollation();

    //! make the type of to_be_casted to be comparable to attr.vc by wrapping to_be_casted in a TypeCastColumn
    DescriptorJoinType desc_t;
	DTCollation collation;
	// buffers for UTF values converted to binary form
	RCBString buf_attr, buf_val1;
	int buf_len;
};

class SortDescriptor
{
public:
	VirtualColumn*	vc;
	int			dir;		// ordering direction: 0 - ascending, 1 - descending
	SortDescriptor() : vc(NULL), dir(0) {};
	int operator==(const SortDescriptor& sec);

};

DescriptorType GetDescOperType(const Descriptor& d);
bool IsSetOperator(Operator op);
bool IsSetAllOperator(Operator op);
bool IsSetAnyOperator(Operator op);
bool ISTypeOfEqualOperator(Operator op);
bool ISTypeOfNotEqualOperator(Operator op);
bool ISTypeOfLessOperator(Operator op);
bool ISTypeOfLessEqualOperator(Operator op);
bool ISTypeOfMoreOperator(Operator op);
bool ISTypeOfMoreEqualOperator(Operator op);

#endif
