/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _FILTER_H_
#define _FILTER_H_

#include <assert.h>
#ifdef __WIN__
#include <winsock2.h>
#endif
#include <boost/pool/pool.hpp>
#include <boost/pool/object_pool.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/static_assert.hpp>

#include "common/stdafx.h"
#include "common/CommonDefinitions.h"
#include "compress/BitstreamCompressor.h"
#include "system/RCSystem.h"
#include "system/MemoryManagement/TrackableObject.h"
#include "system/IBFile.h"

class Filter {
public:
	class Bit;
	class ConstBit;
	template<typename const_qual_t>
	class Iterator
	{
	public:
		struct TYPE
		{
			typedef enum
			{
				D_SET, D_CLEAR, D_BOTH
			} type_t;
		};
	private:
		Filter* _owner;
		_int64 _idx;
		typename TYPE::type_t _type;
	public:
		Iterator(void) :
			_owner(NULL), _idx(0), _type(TYPE::D_BOTH)
		{ /* Look at mysys/my_bitmap.c in MySQL */
		}
		Iterator(Iterator const& it) :
			_owner(it._owner), _idx(it._idx), _type(it._type)
		{
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ( _idx >= 0 ) && ( (_owner == NULL) || (_idx <= _owner->NoObj()) ) );
		}
		Iterator& operator=(Iterator const& it)
		{
			if(&it != this) {
				_owner = it._owner;
				_idx = it._idx;
				_type = it._type;
			}
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ( _idx >= 0 ) && ( _idx <= _owner->NoObj() ) );
			return (*this);
		}
		bool operator==(Iterator const& it) const
		{
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT( _owner == it._owner );
			return (_idx == it._idx);
		}
		bool operator!=(Iterator const& it) const
		{
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT( _owner == it._owner );
			return (_idx != it._idx);
		}
		Iterator& operator++(void);
		Iterator operator++(int)
		{
			Iterator it(*this);
			operator ++();
			return (it);
		}
		Iterator& operator--(void)
		{
			--_idx;
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! "not implemented" );
			return (*this);
		}
		Iterator operator--(int)
		{
			Iterator it(*this);
			operator --();
			return (it);
		}
		const_qual_t operator*(void) const
		{
			return (const_qual_t(_owner, _idx));
		}
	private:
		friend class Filter;
		Iterator(Filter*, _int64, typename Iterator::TYPE::type_t const&);
	};

	typedef Iterator<Bit> iterator;
	typedef Iterator<ConstBit> const_iterator;

//	Filter(int len_block, ushort len_bit);
	Filter(_int64 no_obj, bool all_ones = false);
	Filter(const char* fname);
	Filter(const Filter& filter);
	// Data size: len_block - number of 64k blocks, len_bit - the remaining bits (in [len_block+1]-th block)
	~Filter();

	// Delayed operations on filter
	void SetDelayed(uint b, ushort pos);
	void SetDelayed(_uint64 n);
	void ResetDelayed(uint b, ushort pos);
	void ResetDelayed(_uint64 n);
	void CommitPack(uint block);
	void Commit();

	// Random and massive access to filter
	void Set();					// all = 1
	void SetBlock(int b);		// all in block = 1
	void Set(_int64 n)			// set position n to 1
	{ Set((int)(n >> 16), (ushort)(n & 65535));	}
	void Set(int b, ushort n);	// set position n in block b to 1
	void SetBetween(_int64 n1,_int64 n2);				// set objects n1-n2 to 1
	void SetBetween(int b1, ushort  n1, int b2, ushort n2);	// set 1 between...
	void SetCurrent()		// set the position indicated by the last GetNext... command (faster!)
	{ Set(iterator_b, iterator_n); }
	void Reset();
	void Reset(_int64 n)
	{ Reset((int)(n >> 16), (ushort)(n & 65535)); }
	void ResetBlock(int b);
	void Reset(int b, ushort n);
	void ResetBetween(_int64 n1,_int64 n2);
	void ResetBetween(int b1, ushort n1, int b2, ushort n2);
	void ResetCurrent()	// reset the position indicated by the last GetNext... command (faster!)
	{ Reset(iterator_b, iterator_n); }
	void ResetCurrentDelayed()	// reset the position indicated by the last GetNext... command (faster!)
	{ ResetDelayed(iterator_b, iterator_n); }
	void Reset(Filter &f2);			// reset all positions where f2 is 1
	bool Get(int b, ushort n);
	bool Get(_int64 n)
	{
		if(no_blocks == 0)
			return false;
		return Get((int)(n >> 16), (ushort)(n & 65535));
	}
	bool IsEmpty();
	bool IsEmpty(int b);	// block b
	bool IsEmptyBetween(_int64 n1, _int64 n2);	// true if there are only 0 between n1 and n2, inclusively
	bool IsFull();
	bool IsFull(int b);		// block b
	bool IsFullBetween(_int64 n1, _int64 n2);	// true if there are only 1 between n1 and n2, inclusively

	void Copy(Filter& f, int block);

	// Logical operations on filter
	bool IsEqual(Filter &sec);
	void And(Filter &f2);
	void Or(Filter &f2);
	void Not();

	// Iterative access to filter
	void GetNext1(int &b, ushort &n);	// the next iterator position
	_int64 GetNext1()								// the next iterator position
	{
		int b = 0;
		ushort n = 0;
		this->GetNext1(b, n);
		if(b < 0)
			return b;
		return (_int64)((_int64)b<<16) + n;
	}
	void GetNext0(int &b, ushort &n);	// the next iterator position
	_int64 GetNext0();								// the next iterator position
	_int64 GetCurrent() const;		// the current (defined by the previous GetNext...) iterator position
	void Rewind()			{iterator_b = iterator_n = -1;}
	void RewindTo(_int64 n);	// rewind iterator to the position n
								// Note: if on n-th position it is 1, then GetNext1() will return n.
	void Rewind(int b);		// rewind iterator to the beginning of the block b
	_int64 Find1(_int64 k);	// find the position of the k-th "1" in the filter (or -1 if k>=NoOnes),
							// k=0 for the first existing "1"

	// Statistics etc.
	_int64 NoOnes() const;
	uint NoOnes(int b);	// block b
	_int64 NoOnesBetween(_int64 n1, _int64 n2);	// no of 1 between n1 and n2, inclusively
	_int64 NoObj() const;			// number of objects (table size)
	int NoBlocks() const;
	int NoAddBits() const;
	int DensityWeight();			// = 65537 for empty filter or a filter with only one nonempty block.
									// Otherwise it is an average number of ones in nonempty blocks.

	// Maintenance
	char* GetFilePrefix();
	void Remove();
	void Grow(_int64 grow_size, bool value = true);
	void Store(const char *fname);
	void MoveBlockToEnd(int b);
	void MoveBitsToEnd(int block, int no_of_bits_to_move);

	iterator begin( iterator::TYPE::type_t const& );
	iterator end( iterator::TYPE::type_t const& );
	iterator find( _uint64, iterator::TYPE::type_t const& );
	const_iterator begin( const_iterator::TYPE::type_t const& ) const;
	const_iterator end( const_iterator::TYPE::type_t const& ) const;
	const_iterator find( _uint64, const_iterator::TYPE::type_t const& ) const;

	////////////////////////////////////////////////////////////////////////////////
	boost::pool<> *GetBitBlockPool() { return bit_block_pool; };

	friend class FilterOnesIterator;
	friend class FilterOnesIteratorOrdered;

	class Block;
	Block* GetBlock(int b){assert(b < no_blocks); return blocks[b];};
	uchar	GetBlockStatus(int i){assert(i < no_blocks); return block_status[i];};
	static const uchar FB_FULL = 0;		// block status: full
	static const uchar FB_EMPTY	= 1;		// block status: empty
	static const uchar FB_MIXED	= 2;		// block status: neither full nor empty

private:
	void Construct(bool all_ones = false);
	void ConstructPool();
	void AddNewBlocks(int new_blocks, bool value);
//	Filter& operator=(const Filter& filter) {};
//	const static ushort version = 1;
	int iterator_b, iterator_n;				// the current iterator position (i.e. the value returned by the last GetNext... function)
	int no_blocks;							//number of blocks
	int no_of_bits_in_last_block;	//number of bits in the last block
	void InitDelayedStats(); // initializes delayed_stats
	char* save_path_prefix;
	
	Block** blocks;	
	// Ghostbuster version

	uchar *block_status;

    enum { BIT_BLOCK_SIZE=8192,
           COMPRESSED_BIT_BLOCK_SIZE=BIT_BLOCK_SIZE+10 };

	static const int bitBlockSize;
	static const int compressedBitBlockSize;
	static const uchar file_format_v1;

	boost::pool<> *bit_block_pool;		//default allocator uses system heap
	boost::object_pool<Block> *block_object_pool;
	int* delayed_stats; // array of int's used for delayed Set/Reset: -1 = initial value; -2 = delayed not possible; >=0 = actual set position

#if  defined(__WIN__) || (__GNUC__ >= 4)
#pragma pack(push, 1)
#elif defined(__GNUC__) && (__GNUC__ == 3 &&  __GNUC_MINOR__ == 4)
#pragma pack(1)
#endif

	struct Header
	{
		uint nob;					// number of blocks
		uint bil;					// number of bits in the last block
		uchar v;					// format version
	};
	BOOST_STATIC_ASSERT(sizeof(Header) == 9);

	struct BHeader
	{
		uint len;
		uint nob;
		uint no1;
	};
	BOOST_STATIC_ASSERT(sizeof(BHeader) == 12);

#if  defined(__WIN__) || (__GNUC__ >= 4)
#pragma pack(pop)
#elif defined(__GNUC__) && (__GNUC__ == 3 &&  __GNUC_MINOR__ == 4)
#pragma pack()
#endif
public:
	class Block {
	public:
		Block(Filter* owner, int _no_obj = 65536, bool set_full = false);
		Block(Block const& block, Filter* owner);
		Block(Filter* owner, IBFile* f);
		~Block();
		bool Set(ushort n);	// the object = 1
		bool Set(ushort  n1, ushort n2);	// set 1 between..., true => block is full
		bool Reset(ushort n);
		bool Reset(ushort n1, ushort n2);	// true => block is empty
		bool Reset(Block &b2);				// reset at all position where b2 is 1
		bool Get(ushort n)
		{ return (block_table[n/32] & 1 << n%32) ? true : false; }
		bool IsEqual(Block &b2);
		bool And(Block &b2);					// true => block is empty
		bool Or(Block &b2);					// true => block is full
		void Not();
		uint NoOnes()
		{ return no_set_bits; }
		bool IsEmptyBetween(ushort n1, ushort n2);
		bool IsFullBetween(ushort n1, ushort n2);
		uint NoOnesBetween(ushort n1, ushort n2);

		bool GetNext1(int &n);	// the next iterator position
		bool GetNext0(int &n);	// the next iterator position
		int Find1(int k);

		uint NoObj(){return no_obj + 1;};

		bool IsCompressUpToDate(){return compress_up_to_date;}
		void GrowBlock(int grow_size, bool value);
		void ShrinkBlock(int new_size);
		void Store(IBFile &f);

		friend class FilterOnesIterator;
		friend class FilterOnesIteratorOrdered;

		bool IsNumberOfOnesInHeaderProper(int& tst);

	private:
		void Clean(bool value);
		void InitBits(ushort  n1, ushort n2, bool value);
		void CommonInit(Filter *owner);
		
		ushort no_obj;		                //index of the last used position
		ushort block_size;					//size of the block_table table in uint
		uint *block_table;						//objects table
		ushort no_set_bits;
		uchar dirty ;
		uchar compress_up_to_date ;
		uint compress_buffer_size;
		uint b, lastn, bln, bitsLeft;
		static const int posOf1 [];
		Filter* owner;
//		No idea how to create an allocator for the pool below to use BH heap, due to static methods in allocator
		boost::pool<> *bit_block_pool;
	};
};

//////////////////////////////////////////////////////////////////////

class FilterOnesIterator
{
public:
    FilterOnesIterator();
    FilterOnesIterator(Filter *ff);
    void Init(Filter *ff);
    virtual void Rewind();
    void Rewind(int pack);
    virtual void NextPack();
    bool IsValid()
    {
        return valid;
    }

    _int64 operator *() const
    {
        BHASSERT_WITH_NO_PERFORMANCE_IMPACT(valid);
        return cur_position;
    }

    virtual FilterOnesIterator & operator ++();
    _int64 GetPackSizeLeft();
    _int64 GetTotalSize()
    {
        return f->NoOnes();
    }

    bool InsideOnePack()
    {
        return inside_one_pack;
    }

    int GetCurrPack()
    {
        BHASSERT_WITH_NO_PERFORMANCE_IMPACT(valid);
        return iterator_b;
    }

    int GetCurrInPack()
    {
        BHASSERT_WITH_NO_PERFORMANCE_IMPACT(valid);
        return iterator_n;
    }
    virtual FilterOnesIterator* Copy();

protected:
    bool valid;
    Filter *f;
    _int64 cur_position;
    int iterator_b;
    int iterator_n;
    uint b, lastn, bln, bitsLeft;
    int ones_left_in_block;
    int prev_block;
    bool inside_one_pack;
    void Reset();
};

//////////////////////////////////////////////////////////////////////

class PackOrderer;

class FilterOnesIteratorOrdered	: public FilterOnesIterator	// iterating through all values 1 in a filter with given order of blocks (datapacks)
{
public:
	FilterOnesIteratorOrdered();		// empty initialization: no filter attached
	FilterOnesIteratorOrdered(Filter *ff, PackOrderer* po);

	void Init(Filter *ff, PackOrderer* po);		// like constructor - reset iterator position
	virtual void Rewind();				// iterate from the beginning (the first 1 in Filter)
//	void Rewind(int pack);		// iterate from the beginning of the given pack (the first 1 in Filter)
//								// note: if the pack is empty, set IsValid() to false

	//! move to the next 1 position within an associated Filter
	virtual FilterOnesIteratorOrdered& operator++();
	virtual void NextPack();			// iterate to a first 1 in the next nonempty pack
    virtual FilterOnesIteratorOrdered* Copy();

private:
	PackOrderer* po;
};


//////////////////////////////////////////////////////////////////////
class Filter::ConstBit {
	Filter& _owner;
	_uint64 _idx;
public:
	operator bool ( void ) const;
	_uint64 GetBlockNo( void ) const;
	_uint64 GetNoInBlock( void ) const;
	_uint64 GetIndex( void ) const;
	bool operator == ( bool const& ) const;
	bool operator != ( bool const& ) const;
	bool operator | ( bool const& );
	bool operator & ( bool const& );
	bool operator ^ ( bool const& );
private:
	ConstBit( Filter*, _uint64 );
	friend class Filter::Iterator<ConstBit>;
};

class Filter::Bit {
	Filter& _owner;
	_uint64 _idx;
public:
	operator bool ( void ) const;
	_uint64 GetBlockNo( void ) const;
	_uint64 GetNoInBlock( void ) const;
	_uint64 GetIndex( void ) const;
	bool operator = ( bool const& );
	bool operator == ( bool const& ) const;
	bool operator != ( bool const& ) const;
	Bit& operator |= ( bool const& );
	Bit& operator &= ( bool const& );
	Bit& operator ^= ( bool const& );
	bool operator | ( bool const& );
	bool operator & ( bool const& );
	bool operator ^ ( bool const& );
private:
	Bit( Filter*, _uint64 );
	friend class Filter::Iterator<Bit>;
};

typedef boost::shared_ptr<Filter> FilterPtr;
typedef boost::scoped_ptr<Filter> FilterScopedPtr;
typedef std::auto_ptr<Filter> FilterAutoPtr;

#endif

