/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "JoinerHash.h"
#include "vc/VirtualColumn.h"
#include "common/bhassert.h"
#include "system/ConnectionInfo.h"
#include "core/RoughMultiIndex.h"
#include "core/tools.h"

using namespace std;

JoinerHash::JoinerHash( MultiIndex *_mind, RoughMultiIndex *_rmind,
			bool _count_only_now,
			 _int64 _output_limit) : TwoDimensionalJoiner( _mind, _rmind, _count_only_now, _output_limit)
{
	traversed_dims = DimensionVector(_mind);
	matched_dims = DimensionVector(_mind);
	for(int i = 0; i < _mind->NoDimensions(); i++)
		traversed_hash_column.push_back(-1);
	no_of_traversed_dims = 0;
	vc1 = NULL;
	vc2 = NULL;
	cond_hashed = 0;
	other_cond_exist = false;
	packrows_omitted = 0;
	packrows_matched = 0;
}

JoinerHash::~JoinerHash()
{
	delete [] vc1;
	delete [] vc2;
}

int JoinerHash::ExecuteJoinConditions(vector<Descriptor> &desc)
{
	MEASURE_FET("JoinerHash::ExecuteJoinConditions(...)");
	why_failed = FAIL_HASH;

	vector<int> hash_descriptors;
	/////////////////// Prepare all descriptor information /////////////
	bool first_found = true;
	DimensionVector dims1(mind);		// Initial dimension descriptions
	DimensionVector dims2(mind);
	for(int i = 0; i < desc.size(); i++) {
		if(desc[i].IsType_JoinSimple() && desc[i].op == O_EQ
/*			&&		// Compatible types? - compatibility should be checked below in ColumnBinEncoder.
			(desc[i].attr.vc->Type() == desc[i].val1.vc->Type() ||
			(desc[i].attr.vc->Type().IsString() && desc[i].val1.vc->Type().IsString()) ||
			(desc[i].attr.vc->Type().IsFixed() && desc[i].val1.vc->Type().IsFixed() && desc[i].attr.vc->Type().GetScale() == desc[i].val1.vc->Type().GetScale()))
*/			// type compatibility checked. Now check the dimension compatibility.
			) {
			if(first_found) {
				hash_descriptors.push_back(i);
				desc[i].attr.vc->MarkUsedDims(dims1);
				desc[i].val1.vc->MarkUsedDims(dims2);
				mind->MarkInvolvedMaterialClasses(dims1);
				mind->MarkInvolvedMaterialClasses(dims2);
				first_found = false;
			} else {
				DimensionVector sec_dims1(mind);		// Make sure the local descriptions are compatible
				DimensionVector sec_dims2(mind);
				desc[i].attr.vc->MarkUsedDims(sec_dims1);
				desc[i].val1.vc->MarkUsedDims(sec_dims2);
				if(dims1.Includes(sec_dims1) && dims2.Includes(sec_dims2)) {
					hash_descriptors.push_back(i);
				} else if(dims1.Includes(sec_dims2) && dims2.Includes(sec_dims1)) {
					desc[i].SwitchSides();
					hash_descriptors.push_back(i);
				}
			}
		}
		else {
			other_cond_exist = true;
			other_cond.push_back(desc[i]);
		}
	}
	cond_hashed = hash_descriptors.size();
	if(cond_hashed == 0) {
		why_failed = FAIL_HASH;
		return 0;
	} else {	// else is needed here: we will not need these MIIterators any further
		// Check the proper direction: the smaller dimension should be traversed, the larger one should be matched
		MIIterator mit1(mind, dims1);
		MIIterator mit2(mind, dims2);
		if(mit1.NoTuples() > mit2.NoTuples()) {				// switch sides of joining conditions
			for(int i = 0; i < cond_hashed; i++)
				desc[hash_descriptors[i]].SwitchSides();
		}
	}
	//////////////////////////////////////////////////////////////////////////////////
	// jhash is a class field, initialized as empty
	vc1 = new VirtualColumn * [cond_hashed];
	vc2 = new VirtualColumn * [cond_hashed];
	bool compatible = true;
	for(int i = 0; i < cond_hashed; i++)		// add all key columns
	{
		vc1[i] = desc[hash_descriptors[i]].attr.vc;
		vc2[i] = desc[hash_descriptors[i]].val1.vc;
		vc1[i]->MarkUsedDims(traversed_dims);			// Prepare final descriptions
		vc2[i]->MarkUsedDims(matched_dims);
		compatible = jhash.AddKeyColumn(vc1[i], vc2[i]);
	}
	// check whether we should take into account more dims
	mind->MarkInvolvedMaterialClasses(traversed_dims);
	mind->MarkInvolvedMaterialClasses(matched_dims);
	if(desc[0].IsOuter()) {	// (t1 lj t2) join (t3 lj t4) - make sure the materialized classes are involved on left size
		if(matched_dims[desc[0].outer_dim])
			mind->MarkMaterialClasses(traversed_dims);
		if(traversed_dims[desc[0].outer_dim])
			mind->MarkMaterialClasses(matched_dims);
	}
	if(traversed_dims.Intersects(matched_dims) ||	// both materialized - we should rather use a simple loop
		!compatible) {								// could not prepare common encoding
		why_failed = FAIL_HASH;
		return 0;
	}
	// prepare columns for traversed dimension numbers in hash table
	no_of_traversed_dims = 0;
	for(int i = 0; i < mind->NoDimensions(); i++) {
		if(traversed_dims[i]) {
			traversed_hash_column[i] = cond_hashed + no_of_traversed_dims; // jump over the joining key columns
			no_of_traversed_dims++;
			int bin_index_size = 4;
			if(mind->OrigSize(i) > 0x000000007FFFFF00)
				bin_index_size = 8;
			jhash.AddTupleColumn(bin_index_size);
		}
	}
	// jhash prepared, perform the join
	ExecuteJoin();
	why_failed = NOT_FAILED;
	return desc.size();			// all done
}

void JoinerHash::ExecuteJoin()
{
	MEASURE_FET("JoinerHash::ExecuteJoin(...)");
	_int64 joined_tuples = 0;

	// preparing the new multiindex tables
	MIIterator tr_mit(mind, traversed_dims);
	MIIterator match_mit(mind, matched_dims);	// this iterator is actually just for counting NoTuples() below
	_uint64 dim1_size = tr_mit.NoTuples();
	_uint64 dim2_size = match_mit.NoTuples();

	_int64 approx_join_size = dim1_size;
	if(dim2_size > approx_join_size)
		approx_join_size = dim2_size;

	jhash.Initialize(_int64(dim1_size * 1.5));
	MINewContents new_mind(mind);
	new_mind.SetDimension(traversed_dims);
	new_mind.SetDimension(matched_dims);
	if(!count_only_now)
		new_mind.Init(approx_join_size);
	// joining itself

	_int64 traversed_tuples = 0;

	if(dim1_size > 0 && dim2_size > 0)
		while(tr_mit.IsValid()) {
			traversed_tuples += TraverseDim(tr_mit);
			if(tr_mit.IsValid())
				rccontrol.lock(m_conn.GetThreadID()) << "Traversed " << traversed_tuples << "/" << dim1_size << " rows." << unlock;
			else
				rccontrol.lock(m_conn.GetThreadID()) << "Traversed all " << dim1_size << " rows." << unlock;

			joined_tuples += MatchDim(new_mind);
			rccontrol.lock(m_conn.GetThreadID()) << "Produced " << joined_tuples << " tuples." << unlock;
			if(output_limit != -1 && output_limit <= joined_tuples)
				break;
		}
	// revert multiindex to the updated tables
	if(packrows_omitted > 0)
		rccontrol.lock(m_conn.GetThreadID()) << "Roughly omitted " << int(packrows_omitted / double(packrows_matched) * 10000.0) / 100.0 << "% packrows." << unlock;
	if(count_only_now)
		new_mind.CommitCountOnly(joined_tuples);
	else
		new_mind.Commit(joined_tuples);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

_int64 JoinerHash::TraverseDim(MIIterator &mit)
{
	MEASURE_FET("JoinerHash::TraverseMaterialDim(...)");
	jhash.ClearAll();

	_int64 hash_row = 0;				// hash_row = 0, otherwise deadlock for null on the first position
	_int64 traversed_rows = 0;
	while(mit.IsValid()) {
		if(m_conn.killed())
			throw KilledRCException();
		if(mit.PackrowStarted()) {
			for(int i = 0; i < cond_hashed; i++)
				vc1[i]->LockSourcePacks(mit);
		}
		bool null_found = false;
		for(int i = 0; i < cond_hashed; i++) {
			if(vc1[i]->IsNull(mit)) {
				null_found = true;
				break;
			}
			jhash.PutKeyValue(i, mit);
		}
		if(!null_found) {  // else go to the next row - equality cannot be fulfilled
			hash_row = jhash.FindAndAddCurrentRow();
			if(hash_row == NULL_VALUE_64)
				return traversed_rows;			// no space left - stop for now and then restart from the current row
			// Put the tuple column. Note: needed also for count_only_now, because another conditions may need it.
			for(int i = 0; i < mind->NoDimensions(); i++)
				if(traversed_dims[i])
					jhash.PutTupleValue(traversed_hash_column[i], hash_row, mit[i]);
		}
		++mit;
		traversed_rows++;
	}
	return traversed_rows;
}

_int64 JoinerHash::MatchDim(MINewContents &new_mind)
{
	MEASURE_FET("JoinerHash::MatchMaterialDim(...)");
	_int64 joined_tuples = 0;
	_int64 hash_row;
	_int64 no_of_matching_rows;
	MIIterator mit(mind, matched_dims);
	MIDummyIterator combined_mit(mind);		// a combined iterator for checking non-hashed conditions, if any
	while(mit.IsValid()) {
		if(m_conn.killed())
			throw KilledRCException();
		// Rough and locking part
		bool omit_this_packrow = false;
		bool packrow_uniform = false;				// if the packrow is uniform, process it massively
		if(mit.PackrowStarted()) {
			packrow_uniform = true;
			for(int i = 0; i < cond_hashed; i++) {
				if(vc2[i]->Type().IsLookup() || vc2[i]->Type().IsString()) {
					packrow_uniform = false;
				} else {
					_int64 local_min = vc2[i]->GetMinInt64(mit);
					_int64 local_max = vc2[i]->GetMaxInt64(mit);
					if(local_min != NULL_VALUE_64 && local_max != NULL_VALUE_64 &&
						jhash.ImpossibleValues(i, local_min, local_max)) {
						omit_this_packrow = true;
						break;
					}
					if(local_min == NULL_VALUE_64 || local_max == NULL_VALUE_64 ||
							other_cond_exist || local_min != local_max ||
							vc2[i]->NullsPossible()) {
						packrow_uniform = false;
					}
				}
			}
			packrows_matched++;
			if(packrow_uniform && !omit_this_packrow) {
				for(int i = 0; i < cond_hashed; i++) {
					_int64 local_min = vc2[i]->GetMinInt64(mit);
					jhash.PutMatchedValue(i, local_min);
				}
				no_of_matching_rows = jhash.InitCurrentRowToGet() * mit.GetPackSizeLeft();
				if(!count_only_now)
					while((hash_row = jhash.GetNextRow()) != NULL_VALUE_64) {
						MIIterator mit_this_pack(mit);
						do {
							SubmitJoinedTuple(hash_row, mit_this_pack, new_mind);
							++mit_this_pack;
						} while(mit_this_pack.IsValid() && !mit_this_pack.PackrowStarted());
					}
				joined_tuples += no_of_matching_rows;
				omit_this_packrow = true;
			}
			if(omit_this_packrow) {
				mit.NextPackrow();
				packrows_omitted++;
				continue;				// here we are jumping out for impossible or uniform packrow
			}
			for(int i = 0; i < cond_hashed; i++)
				vc2[i]->LockSourcePacks(mit);
		}
		// Exact part - make the key row ready for comparison
		bool null_found = false;
		bool non_matching_sizes = false;
		for(int i = 0; i < cond_hashed; i++) {
			if(vc2[i]->IsNull(mit)) {
				null_found = true;
				break;
			}
			jhash.PutMatchedValue(i, vc2[i], mit);
		}
		if(!null_found && !non_matching_sizes) {	// else go to the next row - equality cannot be fulfilled
			no_of_matching_rows = jhash.InitCurrentRowToGet();
			// Find all matching rows
			if(!other_cond_exist) {
				// Basic case - just equalities
				if(!count_only_now)
					while((hash_row = jhash.GetNextRow()) != NULL_VALUE_64)
						SubmitJoinedTuple(hash_row, mit, new_mind);	// use the multiindex iterator position
				joined_tuples += no_of_matching_rows;
			} else {
				// Complex case: different types of join conditions mixed together
				combined_mit.Combine(mit);
				while((hash_row = jhash.GetNextRow()) != NULL_VALUE_64) {
					bool other_cond_true = true;
					for(int i = 0; i < mind->NoDimensions(); i++) {
						if(traversed_dims[i])
							combined_mit.Set(i, jhash.GetTupleValue(traversed_hash_column[i], hash_row));
					}
					for(int j = 0; j < other_cond.size(); j++) {
						other_cond[j].LockSourcePacks(combined_mit);
						if(other_cond[j].CheckCondition(combined_mit) == false) {
							other_cond_true = false;
							break;
						}
					}
					if(other_cond_true) {
						if(!count_only_now)
							SubmitJoinedTuple(hash_row, mit, new_mind);	// use the multiindex iterator position
						joined_tuples++;
					}
				}
			}
			if(output_limit != -1 && output_limit <= joined_tuples)
				return joined_tuples;
		}
		++mit;
	}
	return joined_tuples;
}

void JoinerHash::SubmitJoinedTuple(_int64 hash_row, MIIterator &mit, MINewContents &new_mind)
{
	// assumption: SetNewTableValue is called once for each dimension involved (no integrity checking)
	for(int i = 0; i < mind->NoDimensions(); i++) {
		if(matched_dims[i])
			new_mind.SetNewTableValue(i, mit[i]);
		else if(traversed_dims[i])
			new_mind.SetNewTableValue(i, jhash.GetTupleValue(traversed_hash_column[i], hash_row));
	}
}

