/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "JoinerHashTable.h"
#include "core/tools.h"

_int64				JoinerHashTable::max_total_size		= 0;

// This is a list of prime steps used to make the probability of persistent conflicts reasonably low.
// The list should not contain 67 (the minimal size of the buffer).

const int prime_steps_no = 13;
int prime_steps[prime_steps_no] = { 1, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47 };

JoinerHashTable::JoinerHashTable( )
{
	// Memory settings (global for all GroupTables)
	if(max_total_size == 0) {	// not set yet
		max_total_size = 128 * 1024 * 1024;						// default settings for less than 2.5 GB reserved
		int mem_scale = TrackableObject::MemorySettingsScale();
		if(mem_scale == 0)		// low settings (not more than 0.5 GB)
			max_total_size = 64 * 1024 * 1024;
		else if(mem_scale == 3)		// high settings (when 2.5 - 5 GB is reserved in 64-bit version)
			max_total_size = 256 * 1024 * 1024;
		else if(mem_scale == 4)		// high settings (when 5 - 10 GB is reserved in 64-bit version)
			max_total_size = 512 * 1024 * 1024;
		else if(mem_scale > 4)		// high settings (when at least 10 GB is reserved in 64-bit version)
			max_total_size = 1024 * 1024 * 1024;
	}

	for_count_only = false;

	initialized = false;
	no_of_occupied = 0;
	input_buffer = NULL;
	column_offset = NULL;
	t = NULL;
	key_buf_width = 0;
	total_width = 0;
	mult_offset = 0;
	no_attr = 0;
	no_key_attr = 0;
}

JoinerHashTable::~JoinerHashTable()
{
	dealloc(t);
	delete [] input_buffer;
	delete [] column_offset;
}

bool JoinerHashTable::AddKeyColumn(VirtualColumn *vc, VirtualColumn *vc_matching)
{
	int cur_col = encoder.size();
	encoder.push_back(ColumnBinEncoder(ColumnBinEncoder::ENCODER_IGNORE_NULLS));		// non-monotonic, non-decodable
	bool success = encoder[cur_col].PrepareEncoder(vc, vc_matching);
	size.push_back(encoder[cur_col].GetPrimarySize());
	return success;
}

void JoinerHashTable::AddTupleColumn(int bin_size)
{
	size.push_back(bin_size);
}

void JoinerHashTable::Initialize(_int64 max_table_size)
{
	// Input buffer format:		<key_1><key_2>...<key_n>
	// Table format:			<key_1><key_2>...<key_n><tuple_1>...<tuple_m><multiplier>
	if(max_table_size < 2)					// otherwise strange things may occur
		max_table_size = 2;
	if(initialized)
		return;
	AddTupleColumn(8);				// add multiplier column
	no_key_attr = encoder.size();
	no_attr = size.size();
	if(no_key_attr==0)
		return;
	if(no_attr - no_key_attr == 1)		// i.e. no tuple columns, just a multiplier
		for_count_only = true;

	column_offset 	= new int [no_attr];

	// Prepare buffers and mappings
	key_buf_width = 0;
	for(int i = 0; i < no_key_attr; i++) {
		encoder[i].SetPrimaryOffset(key_buf_width);
		key_buf_width += size[i];
	}
	key_buf_width = 4 * ( (key_buf_width + 3) / 4);				// e.g. 1->4, 12->12, 19->20
	total_width = key_buf_width;
	for(int i = no_key_attr; i < no_attr; i++) {
		column_offset[i] = total_width;
		total_width += size[i];
	}
	mult_offset = column_offset[no_attr - 1];

	/////////// create buffers ////////////////////

	no_rows = max_total_size / total_width;		// memory limitation
	if( (max_table_size + 1) * 1.25 < no_rows)
		no_rows = _int64((max_table_size + 1) * 1.25);	// make sure that rows_limit>=max_no_groups

	// calculate vertical size (not dividable by a set of some prime numbers)
	if( no_rows < 67 )								// too less groups => high collision probability; 67 is prime.
		no_rows = 67;

	// make sure that the size is not dividable by any of prime_steps (>1)
	bool increased = false;
	do {
		increased = false;
		for( int i = 1; i < prime_steps_no; i++ )
			if( no_rows % prime_steps[i] == 0 ) {
				increased = true;
				no_rows++;
				break;
			}
	} while( increased );
	rows_limit = _int64(no_rows * 0.9);		// rows_limit is used to determine whether the table is full

	t = (unsigned char*)alloc(((total_width / 4) * no_rows) * 4, BLOCK_TEMPORARY);	// no need to cache on disk
	input_buffer = (unsigned char*)new unsigned int [ key_buf_width / 4 ];

	// initialize everything
	ClearAll();
	initialized = true;
}

//////////////////////////////////////////////////////////////////////////////////////////

_int64 JoinerHashTable::FindAndAddCurrentRow()	// a position in the current JoinerHashTable, NULL_VALUE_64 if no more space
{
	//MEASURE_FET("JoinerHashTable::FindAndAddCurrentRow()");
	if(no_of_occupied >= rows_limit)		// no more space
		return NULL_VALUE_64;

	unsigned int crc_code = CRC32(input_buffer, key_buf_width);
	_int64 row = crc_code % no_rows;
	// vertical table size is not dividable by the following step values, so they will eventually
	// iterate through the whole table
	_int64 iterate_step = prime_steps[crc_code % prime_steps_no];
	_int64 startrow = row;
	do {
		_int64 *multiplier = (_int64*)(t + row * total_width + mult_offset);
		if(*multiplier != 0) {
			if(memcmp(t + row * total_width, input_buffer, key_buf_width) == 0) {
				// i.e. identical row found
				_int64 last_multiplier = *multiplier;
				assert(last_multiplier>0);
				(*multiplier)++;
				if(for_count_only)
					return row;
				// iterate several steps forward to find some free location
				row = (row + iterate_step * last_multiplier) % no_rows;
			}
			else {									// some other value found
				// iterate one step forward
				row = (row + iterate_step) % no_rows;
			}
		} else {
			memcpy(t + row * total_width, input_buffer, key_buf_width);
			*multiplier = 1;
			no_of_occupied++;
			return row;
		}
	} while(row != startrow);
	return NULL_VALUE_64;			// search deadlock (we returned to the same position)
}

_int64 JoinerHashTable::InitCurrentRowToGet()
{
	//MEASURE_FET("JoinerHashTable::InitCurrentRowToGet()");
	unsigned int crc_code = CRC32( input_buffer, key_buf_width );
	current_row = crc_code % no_rows;
	// vertical table size is not dividable by the following step values, so they will eventually
	// iterate through the whole table
	current_iterate_step = prime_steps[crc_code % prime_steps_no];
	_int64 startrow = current_row;
	to_be_returned = 0;
	do {
		_int64 *multiplier = (_int64*)(t + current_row * total_width + mult_offset);
		if(*multiplier != 0) {
			if(memcmp(t + current_row * total_width, input_buffer, key_buf_width) == 0) {
				// i.e. identical row found
				to_be_returned = *multiplier;
				return *multiplier;
			}
			else {									// some other value found
				// iterate one step forward
				current_row = (current_row + current_iterate_step) % no_rows;
			}
		} else {
			current_row = NULL_VALUE_64;	// not found
			return 0;
		}
	} while(current_row != startrow);
	current_row = NULL_VALUE_64;			// not found at all
	return 0;
}

_int64 JoinerHashTable::GetNextRow()
{
	//MEASURE_FET("JoinerHashTable::InitCurrentRowToGet()");
	if(to_be_returned == 0)
		return NULL_VALUE_64;
	_int64 row_to_return = current_row;
	to_be_returned--;
	if(to_be_returned == 0)
		return row_to_return;
	// now prepare the next row to be returned
	_int64 startrow = current_row;
	current_row = (current_row + current_iterate_step) % no_rows;
	do {
		assert( *((int*)(t + current_row * total_width + mult_offset)) != 0 );
		if( memcmp( t + current_row * total_width, input_buffer, key_buf_width ) == 0 ) {
			// i.e. identical row found - keep the current_row
			return row_to_return;
		}
		else {	// some other value found - iterate one step forward
			current_row = (current_row + current_iterate_step) % no_rows;
		}
	} while( current_row != startrow );
	return row_to_return;
}

//////////////////////////////////////////////////////////////////////////////////////////

void JoinerHashTable::ClearAll()
{
	memset( t, 0, total_width * no_rows );
	memset( input_buffer, 0, key_buf_width );
	no_of_occupied = 0;
	for(int i = 0; i < no_key_attr; i++)
		encoder[i].ClearStatistics();
}

bool JoinerHashTable::ImpossibleValues(int col, _int64 local_min, _int64 local_max)
{
	return encoder[col].ImpossibleValues(local_min, local_max);
}

/*
bool JoinerHashTable::IsColumnSizeValid(int col, int size)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(col >=0 && col < no_attr - 1);
	return (use_CRC[col] || column_offset[col + 1] - column_offset[col] - 2 >= size);
}
*/
