/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "MultiIndex.h"
#include "MIIterator.h"
#include "GroupDistinctTable.h"
#include "system/ConnectionInfo.h"
#include "tools.h"

using namespace std;

#define MATERIAL_TUPLES_LIMIT   100000000000LL			// = 400-800 GB of cache for one dimension
#define MATERIAL_TUPLES_WARNING   2500000000LL			// = 10-20 GB of cache for one dimension

MultiIndex::MultiIndex()
	:	m_conn(&ConnectionInfoOnTLS.Get())
{
	// update Clear() on any change
	no_dimensions = 0;
	no_tuples = 0;
	no_tuples_too_big = false;
	material_no_tuples = 0;
	mat_max_no_tuples = 0;
	f = NULL;
	t = NULL;
	index_type = IT_VIRTUAL;
	iterator_m=0;
	dim_size = NULL;
	iterator_lock = 0;
}

MultiIndex::MultiIndex(const MultiIndex &s)
	:	m_conn(s.m_conn), _noOpLocks( s._noOpLocks )
{
	no_dimensions = s.no_dimensions;
	no_tuples = s.no_tuples;
	material_no_tuples = s.material_no_tuples;
	mat_max_no_tuples = s.mat_max_no_tuples;
	no_tuples_too_big = s.no_tuples_too_big;
	index_type = s.index_type;
	iterator_m = 0;
	if(no_dimensions>0)
	{
		f = new Filter * [no_dimensions];
		t = new IndexTable * [no_dimensions];
		dim_size = new _int64 [no_dimensions];
		used_in_output = s.used_in_output;
		for(int i = 0; i < no_dimensions; i++)
		{
			if(s.f[i]) {
				f[i] = new Filter(*s.f[i]);
				//NOTE: f[i] does not have to be unlocked to be inserted into this memory manager
			} else
				f[i] = NULL;

			if(s.t[i]) {
				(s.t[i])->Lock();
				t[i] = new IndexTable(*s.t[i]);
				(s.t[i])->Unlock();
				//NOTE: t[i] does not have to be unlocked to be inserted into this memory manager
			}	else
				t[i] = NULL;

			dim_size[i]=s.dim_size[i];
			can_be_distinct.push_back(s.can_be_distinct[i]);
			forget_now.push_back(false);
		}
	} else {
		f = NULL;
		t = NULL;
		dim_size = NULL;
	}
	iterator_lock = 0;
}

void MultiIndex::Copy(MultiIndex &s)			// copy the object s
{
	if(f)
		for(int i = 0; i < no_dimensions; i++) delete f[i];
	if(t)
		for(int i = 0; i < no_dimensions; i++) delete t[i];
	delete [] f;
	delete [] t;
	delete [] dim_size;
	/////////////////////////////////////////
	_noOpLocks = s._noOpLocks;
	no_dimensions = s.no_dimensions;
	no_tuples = s.no_tuples;
	material_no_tuples = s.material_no_tuples;
	mat_max_no_tuples = s.mat_max_no_tuples;
	no_tuples_too_big = s.no_tuples_too_big;
	index_type = s.index_type;
	iterator_m=0;
	if(no_dimensions > 0) {
		f = new Filter * [no_dimensions];
		t = new IndexTable * [no_dimensions];
		dim_size = new _int64 [no_dimensions];
		used_in_output = s.used_in_output;
		can_be_distinct = s.can_be_distinct;
		forget_now.clear();
		for(int i = 0; i < no_dimensions; i++) {
			if(s.f[i]) {
				f[i] = new Filter(*s.f[i]);
				//NOTE: f[i] does not have to be unlocked to be inserted into this memory manager
			} else
				f[i] = NULL;

			if(s.t[i]) {
				(s.t[i])->Lock();
				t[i] = new IndexTable(*s.t[i]);
				(s.t[i])->Unlock();
				//NOTE: t[i] does not have to be unlocked to be inserted into this memory manager
			} else
				t[i] = NULL;

			dim_size[i] = s.dim_size[i];
			forget_now.push_back(false);
		}
	} else {
		f = NULL;
		t = NULL;
		dim_size = NULL;
	}
}

void MultiIndex::Canibalize(MultiIndex &s)		// copy the object s by moving pointers, then reset s
{
	if(f)
		for(int i = 0; i < no_dimensions; i++) delete f[i];
	if(t)
		for(int i = 0; i < no_dimensions; i++) delete t[i];
	delete [] f;
	delete [] t;
	delete [] dim_size;
	/////////////////////////////////////////
	_noOpLocks = s._noOpLocks;
	no_dimensions = s.no_dimensions;
	no_tuples = s.no_tuples;
	material_no_tuples = s.material_no_tuples;
	mat_max_no_tuples = s.mat_max_no_tuples;
	no_tuples_too_big = s.no_tuples_too_big;
	f = s.f;
	t = s.t;
	dim_size = s.dim_size;
	index_type = s.index_type;
	can_be_distinct=s.can_be_distinct;
	used_in_output = s.used_in_output;
	forget_now = s.forget_now;
	iterator_m=0;
	////////////////////////////////////////
	s._noOpLocks.clear();
	s.no_dimensions = 0;
	s.no_tuples = 0;
	s.material_no_tuples = 0;
	s.mat_max_no_tuples=0;
	s.no_tuples_too_big = false;
	s.f = NULL;
	s.t = NULL;
	s.index_type = IT_VIRTUAL;
	s.iterator_m=0;
	s.dim_size = NULL;
	s.can_be_distinct.clear();
	s.used_in_output.clear();
	s.forget_now.clear();
}

MultiIndex::~MultiIndex()
{
	// update Clear() on any change
	try {
		if(f)
			for(int i = 0; i < no_dimensions; i++) delete f[i];
		if(t)
			for(int i = 0; i < no_dimensions; i++) delete t[i];
	} catch ( ... ) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! "exception from destructor" );
	}
	delete [] f;
	delete [] t;
	delete [] dim_size;
}

void MultiIndex::Clear()
{
	try {
		if(f)
			for(int i = 0; i < no_dimensions; i++) delete f[i];
		if(t)
			for(int i = 0; i < no_dimensions; i++) delete t[i];
	} catch ( ... ) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT( ! "exception from destructor" );
	}
	delete [] f;
	delete [] t;
	delete [] dim_size;
	no_dimensions = 0;
	no_tuples = 0;
	no_tuples_too_big = false;
	material_no_tuples = 0;
	mat_max_no_tuples = 0;
	f = NULL;
	t = NULL;
	index_type = IT_VIRTUAL;
	iterator_m=0;
	dim_size = NULL;
	iterator_lock = 0;
}

void MultiIndex::Empty(int dim_to_make_empty)
{
	bool mat_involved = (dim_to_make_empty == -1 || t[dim_to_make_empty]);
	bool still_material = false;
	for(int i = 0; i < no_dimensions; i++) {
		if(dim_to_make_empty == -1 || dim_to_make_empty == i || (t[i] && mat_involved)) {
			if(t[i]) {
				f[i]=new Filter( dim_size[i] );
				delete t[i];
				t[i] = NULL;
				material_no_tuples = 0;
			}
			f[i]->Reset();
			can_be_distinct[i] = true;
			used_in_output[i] = true;
			forget_now[i] = false;
		}
		if(t[i])
			still_material = true;
	}
	no_tuples = 0;
	no_tuples_too_big = false;
	index_type = (still_material ? IT_MIXED : IT_VIRTUAL);
}

void MultiIndex::AddDimension()
{
	// First stage: enlarge tables of dimensions
	no_dimensions++;
	_noOpLocks.push_back( 0 );
	Filter		 **nf = new Filter * [no_dimensions];
	IndexTable	 **nt = new IndexTable * [no_dimensions];
	_int64		  *ns = new _int64 [no_dimensions];
	for(int i = 0; i < no_dimensions - 1; i++)
	{
		ns[i] = dim_size[i];
		nf[i] = f[i];
		nt[i] = t[i];
	}
	delete [] dim_size;
	delete [] f;
	delete [] t;
	f = nf;
	t = nt;
	dim_size = ns;
	dim_size[no_dimensions - 1] = 0;
	f[no_dimensions - 1] = NULL;
	t[no_dimensions - 1] = NULL;
	return;		// Note: other functions will use AddDimension() to enlarge tables
}

void MultiIndex::AddDimension_constant(_uint64 const_val, _uint64 size) {
	AddDimension();
	int bytes_per_value = ( size>0xFFFF ? (size>0xFFFFFFFF ? 8 : 4 ) : 2);
	if( const_val==0 || ( size > (8*bytes_per_value*mat_max_no_tuples) )) {
		// compare sizes of Filter and IndexTable (in bits)
		if(index_type==IT_VIRTUAL) index_type=IT_MIXED;
		t[no_dimensions - 1] = new IndexTable(mat_max_no_tuples, bytes_per_value,0);
		t[no_dimensions - 1]->Lock(); //TrackableObject is not auto-locked at the construction time
		for(_int64 i = 0; i < material_no_tuples; i++)
			t[no_dimensions - 1]->Set64(i, const_val);
		t[no_dimensions - 1]->Unlock();
	} else {
		if(index_type==IT_ORDERED) index_type=IT_MIXED;
		if( size>0 ) {
			f[no_dimensions - 1] = new Filter(size);
			f[no_dimensions - 1]->Set(const_val-1);		// just one 1 in a filter; note that 0 is reserved for null
		}
		else {
			size = 1;
			f[no_dimensions - 1] = new Filter(1);
			f[no_dimensions - 1]->Reset();
		}
	}
	can_be_distinct.push_back(false);
	used_in_output.push_back(true);
	forget_now.push_back(false);
	dim_size[no_dimensions - 1] = size;
}

void MultiIndex::AddDimension_cross(_uint64 size) {
	if(index_type==IT_ORDERED) index_type=IT_MIXED;
	AddDimension();
	used_in_output.push_back(true);
	forget_now.push_back(false);
	if(no_dimensions > 1)
		MultiplyNoTuples( size );
	else
		no_tuples = size;
	if(size > 0) {
		dim_size[no_dimensions - 1] = size;
		if(index_type != IT_COUNT) {
			f[no_dimensions - 1] = new Filter(size);
			f[no_dimensions - 1]->Set();
		}
	}
	else {		// size == 0   => prepare a dummy dimension with empty filter
		dim_size[no_dimensions - 1] = 1;
		if(index_type != IT_COUNT) {
			f[no_dimensions - 1] = new Filter(1);
			f[no_dimensions - 1]->Reset();
		}
	}
	can_be_distinct.push_back(true);		// may be modified below
	CheckIfVirtualCanBeDistinct();
}

void MultiIndex::AddDimension_cross(const Filter& new_f)
{
	if(index_type==IT_ORDERED) index_type=IT_MIXED;
	AddDimension();
	f[no_dimensions - 1] = new Filter(new_f);
	if(no_dimensions > 1)
		MultiplyNoTuples( new_f.NoOnes() );
	else
		no_tuples = new_f.NoOnes();
	dim_size[no_dimensions - 1] = new_f.NoObj();
	used_in_output.push_back(true);
	forget_now.push_back(false);
	can_be_distinct.push_back(true);
	CheckIfVirtualCanBeDistinct();
}

void MultiIndex::AddDimension_cross(MultiIndex &new_f)
{
	int old_dimensions=no_dimensions;
	for(int d=0;d<new_f.no_dimensions;d++)
	{
		AddDimension();
		dim_size[ old_dimensions+d ] = new_f.dim_size[d];
	}
	if(new_f.index_type==IT_VIRTUAL) index_type=IT_VIRTUAL;
	bool is_pure_virtual = true;	// actual state, not just declared one
	for(int d=0;d<old_dimensions;d++)
	{
		if(t[d]!=NULL) is_pure_virtual=false;
	}
	if(is_pure_virtual)
	{
		material_no_tuples=new_f.material_no_tuples;
		mat_max_no_tuples=new_f.mat_max_no_tuples;
		for ( int d=0; d<new_f.no_dimensions; d++ ) {
			// copy all
			if(new_f.f[d])
				f[d+old_dimensions]=new Filter(*new_f.f[d]);

			if ( new_f.t[d] ){
				(new_f.t[d])->Lock();
				t[d+old_dimensions]=new IndexTable(*new_f.t[d]);
				(new_f.t[d])->Unlock();
			}
			used_in_output.push_back(new_f.used_in_output[d]);
			can_be_distinct.push_back(false);		// just in case - there are many possible situations
			forget_now.push_back(false);
		}
	}	else {
		_uint64 new_mat_tuples=material_no_tuples*new_f.material_no_tuples;
		if( new_f.material_no_tuples==0 )
			new_mat_tuples=material_no_tuples;
		if( material_no_tuples==0 )
			new_mat_tuples=new_f.material_no_tuples;
		if(new_mat_tuples>MATERIAL_TUPLES_LIMIT) {
			throw OutOfMemoryRCException("Out of resources in JOIN.");
		}
		if(new_mat_tuples>material_no_tuples) {
			mat_max_no_tuples=RoundUpTuples(new_mat_tuples);	// add some unused tuples for further enlargement
			for(int d=0;d<old_dimensions;d++) {
				if(t[d]) {
					// leave all virtual, enlarge the rest
					t[d]->Lock();
					IndexTable *t_new_local=new IndexTable(mat_max_no_tuples , t[d]->BytesPerValue(),0);
					t_new_local->CopyLockStatus(*t[d]);
					_uint64 j_start,j_stop,val;
					for(_uint64 i=0;i<material_no_tuples;i++)	{
						j_start=i*new_f.material_no_tuples;
						j_stop=(i+1)*new_f.material_no_tuples;
						val=t[d]->Get64(i);
						for(_uint64 j=j_start;j<j_stop;j++)
							t_new_local->Set64(j,val);		// translate "abcd" into "aaa...aaabbb...bbbccc...cccddd...ddd"
					}
					delete t[d];
					t[d]=t_new_local;
					t[d]->Unlock();
					can_be_distinct[d] = false;
					used_in_output[d] = true;
					forget_now[d] = false;
				}
			}
		}
		for(int d=0;d<new_f.no_dimensions;d++) {
			int dd=d+old_dimensions;
			if(new_f.f[d]) {
				f[dd]=new Filter(*new_f.f[d]);
				can_be_distinct.push_back(false);
				used_in_output.push_back(true);
				forget_now.push_back(false);
			}
			else if(new_f.t[d]) {
				(new_f.t[d])->Lock();
				t[dd]=new IndexTable(mat_max_no_tuples , new_f.t[d]->BytesPerValue(),0);
				t[dd]->Lock();
				_uint64 j_start,j_stop;
				for(_uint64 i=0;i<material_no_tuples;i++) {
					j_start=i*new_f.material_no_tuples;
					j_stop=(i+1)*new_f.material_no_tuples;
					for(_uint64 j=j_start;j<j_stop;j++)
						t[dd]->Set64(j,new_f.t[d]->Get64(j-j_start));		// translate "abcd" into "abcdabcdabcd....abcd"
				}
				(new_f.t[d])->Unlock();
				t[dd]->Unlock();
				can_be_distinct.push_back(false);
				used_in_output.push_back(true);
				forget_now.push_back(false);
			}
		}
		material_no_tuples=new_mat_tuples;
	}
	MultiplyNoTuples( new_f.no_tuples );
}

void MultiIndex::MultiplyNoTuples( _uint64 factor )
{
	no_tuples = SafeMultiplication(no_tuples, factor);
	if(no_tuples == NULL_VALUE_64)
		no_tuples_too_big = true;
}

////////////////////////////////////////////////////////////////////////////

void MultiIndex::CheckIfVirtualCanBeDistinct()	// updates can_be_distinct table in case of virtual multiindex
{
	// check whether can_be_distinct can be true
	// it is possible only when there are only 1-object dimensions
	// and one multiobject (then the multiobje+ct one can be distinct, the rest cannot)
	if(no_dimensions>1)
	{
		int non_one_found = 0;
		for(int i=0;i<no_dimensions;i++)
		{
			if(dim_size[i]>1)
				non_one_found++;
		}
		if(non_one_found==1)
			for(int j=0;j<no_dimensions;j++)
				if(dim_size[j]==1)
					can_be_distinct[j] = false;
				else
					can_be_distinct[j] = true;
		if(non_one_found>1)
			for(int j=0;j<no_dimensions;j++)
				can_be_distinct[j] = false;
	}
}


void MultiIndex::AddTuple(_uint64 *v)
{
	if(no_dimensions==1 && f[0]!=NULL && f[0]->Get(v[0])==false)
	{
		f[0]->Set((_int64)v[0]);
		return;
	}
	Materialize();
	no_tuples++;
	if(no_tuples>mat_max_no_tuples)
		mat_max_no_tuples=RoundUpTuples(no_tuples);
	for(int d=0;d<no_dimensions;d++)
	{
		t[d]->Lock();
		t[d]->ExpandTo(mat_max_no_tuples);
		if(v[d]==NULL_VALUE_64)
			t[d]->Set64(no_tuples-1,0);
		else
			t[d]->Set64(no_tuples-1,v[d]+1);
		t[d]->Unlock();
		can_be_distinct[d]=false;
	}
	material_no_tuples=no_tuples;
}

void MultiIndex::LockForGetIndex( int dim ) {
	++ _noOpLocks[ dim ];
	if ( t[dim] )
		t[dim]->Lock();
}

void MultiIndex::UnlockFromGetIndex( int dim ) {
	-- _noOpLocks[ dim ];
	if ( t[dim] )
		t[dim]->Unlock();
}


_uint64 MultiIndex::DimSize(int dim)				// the size of one dimension: material_no_tuples for materialized, NoOnes for virtual
{
	assert(index_type!=IT_COUNT);
	if(f[dim])
	{
		_uint64 result = f[dim]->NoOnes();
		return result;
	}
	return material_no_tuples;
}


void MultiIndex::LockAllForUse()
{
	for (int dim = 0; dim < no_dimensions; dim++)
		LockForGetIndex(dim);
}

void MultiIndex::UnlockAllFromUse()
{
	for (int dim = 0; dim < no_dimensions; dim++)
		UnlockFromGetIndex(dim);
}

void MultiIndex::SetFilter(Filter *f_in, int dim, bool &nulls_found)  // set 1 in the filter on every position present in this MultiIndex
{
	assert(index_type!=IT_COUNT);
	if(f[dim]!=NULL)
	{
		f_in->Or(*(f[dim]));
	}
	else if(t[dim])
	{
		t[dim]->Lock();
		for(_uint64 tuple=0;tuple<material_no_tuples;tuple++)
			if(t[dim]->Get64(tuple)!=0)					// 0 is null object (not present in the filter)
				f_in->Set(t[dim]->Get64(tuple)-1);
			else
				nulls_found = true;
		t[dim]->Unlock();
	}
}

bool MultiIndex::SetFilterUnique(Filter *f_in, int dim)  // set 1 in the filter on every position present in this MultiIndex
// stop adding and return false in case of any repeating tuple number, or null
{
	assert(index_type!=IT_COUNT);
	if(f[dim]!=NULL)
	{
		f_in->Or(*(f[dim]));
		return true;
	}
	else if(t[dim])
	{
		t[dim]->Lock();
		for(_uint64 tuple = 0; tuple < material_no_tuples; tuple++) {
			_uint64 obj = t[dim]->Get64(tuple);
			if(obj != 0) {						// 0 is null object (not present in the filter)
				obj -= 1;
				if(f_in->Get(obj))
					return false;				// repetition found
				f_in->Set(obj);
			} else
				return false;
		}
		t[dim]->Unlock();
	}
	return true;
}

bool MultiIndex::NullsExist(int dim)				// return true if there exist any 0 value (always false for virtual dimensions)
{
	assert(index_type!=IT_COUNT);	// NOTE: rather slow, so it is better to find nulls by hand, e.g. in MultiIndex::SetFilter
	if(f[dim]) return false;
	if(t[dim]==NULL)	return false;
	t[dim]->Lock();
	for(_uint64 tuple=0;tuple<material_no_tuples;tuple++)
		if(t[dim]->Get64(tuple)==0)
		{
			t[dim]->Unlock();
			return true;
		}
	t[dim]->Unlock();
	return false;
}

void MultiIndex::Materialize()
{
	DimensionVector dimensions(this);
	for(int i = 0; i < no_dimensions; i++)
		dimensions[i] = true;

	Materialize(dimensions);
}

void MultiIndex::Materialize(DimensionVector &dims_to_materialize)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("MultiIndex::Materialize()");
#endif
	assert(index_type!=IT_COUNT);
	if(index_type==IT_ORDERED) return;

	bool materialization_needed = false;
	for(int i = 0; i < no_dimensions; i++) {
		if(t[i] != NULL)
			dims_to_materialize[i] = true;
		if(f[i] != NULL && dims_to_materialize[i])
			materialization_needed = true;
	}
	if(!materialization_needed)
		return;

	_int64 dims_size = 0;
	vector<IndexTable*> new_t;
	{	// this block is needed to ensure mit destruction in a proper place
		MIIterator mit(this, dims_to_materialize);
		dims_size = mit.NoTuples();
		for(int i = 0; i < no_dimensions; i++) {
			new_t.push_back(NULL);
		}

		if(dims_size == NULL_VALUE_64 || dims_size > (1024 * GBYTE) / (dims_to_materialize.NoDimsUsed() * 8))	// = 1 TB cache size, note also that max. Filter size is 64k x sizeof(int)
		{
			throw OutOfMemoryRCException("Out of resources in JOIN.");
			assert(0);
		}
		if(dims_size > MATERIAL_TUPLES_WARNING)
			rccontrol.lock(m_conn->GetThreadID()) << "Warning: a big intermediate object created (~" << ((mit.NoTuples()/8)*no_dimensions)/1000000000LL << " GB)" << unlock;
		mat_max_no_tuples = RoundUpTuples(dims_size);			// add some unused tuples for further enlargement

		// Special case: the empty materialized multiindex
		if(dims_size == 0) {
			index_type = IT_ORDERED;
			no_tuples = 0;
			material_no_tuples = 0;
			no_tuples_too_big = false;
			for(int d = 0; d < no_dimensions; d++) {	// scan through all dimensions, including these outside the scope
				if(f[d] && !dims_to_materialize[d]) {
					index_type = IT_MIXED;
					continue;							// delete only materialized dims, if out of scope
				}
				int bytes_per_value = 2;
				if(t[d]) {
					bytes_per_value = t[d]->BytesPerValue();
					delete t[d];
					t[d] = NULL;
				}
				if(f[d]) {
					bytes_per_value = ( ((_uint64)f[d]->NoObj()+1)>0xFFFF ? (((_uint64)f[d]->NoObj()+1)>0xFFFFFFFF ? 8 : 4 ) : 2);
					delete f[d];
					f[d] = NULL;
				}
				if(!forget_now[d]) {
					t[d]=new IndexTable(mat_max_no_tuples , bytes_per_value, 0);
					t[d]->CopyLocksToIT(_noOpLocks[d]);
				}
			}
			return;
		}

		for(int d = 0; d < no_dimensions; d++)		// scan through all dimensions, including these outside the scope
		{
			if(no_dimensions > 1)
				can_be_distinct[d] = false;
			if(t[d] != NULL)				// enlarge all materialized dimensions, even outside the scope
			{
				if(!forget_now[d]) {
					new_t[d] = new IndexTable(mat_max_no_tuples, t[d]->BytesPerValue(), 0);
					new_t[d]->CopyLockStatus(*t[d]);
					mit.Rewind();
					_uint64 j = 0;
					while(mit.IsValid()) {
						new_t[d]->Set64(j, (mit[d] == NULL_VALUE_64 ? 0 : mit[d] + 1));
						j++;
						++mit;
					}
					new_t[d]->Unlock();
				}
			}
			else if(f[d] != NULL && dims_to_materialize[d])		// leave virtual filters which are outside the scope, materialize the rest
			{
				assert(t[d] == NULL);
				if(!forget_now[d]) {
					int bytes_per_value = ( ((_uint64)f[d]->NoObj()+1)>0xFFFF ? (((_uint64)f[d]->NoObj()+1)>0xFFFFFFFF ? 8 : 4 ) : 2);
					new_t[d] = new IndexTable(mat_max_no_tuples , bytes_per_value, 0);
					new_t[d]->CopyLocksToIT(_noOpLocks[d]);
					mit.Rewind();
					_uint64 j = 0;
					while(mit.IsValid()) {
						new_t[d]->Set64(j, mit[d] + 1);		// +1, because 0 is null object, not possible for virtual dimensions
						j++;
						++mit;
					}
					new_t[d]->Unlock();
				}
			}
			else if(!forget_now[d] && dims_to_materialize[d]) {
				assert(0);	// both are nulls and we don't know why!
			}
		}
	}	// this block is needed to ensure mit destruction in a proper place
	// replace t[d] and cleanup
	material_no_tuples = dims_size;
	index_type = IT_ORDERED;
	for(int d = 0; d < no_dimensions; d++) {
		if(t[d]) {
			delete t[d];
			t[d] = new_t[d];
		} else if(f[d] && dims_to_materialize[d]) {
			delete f[d];
			f[d] = NULL;
			t[d] = new_t[d];
		} else if(f[d])
			index_type = IT_MIXED;
	}
}

void MultiIndex::MakeCountOnly(_int64 mat_tuples)
{
	DimensionVector dim_out(this);
	if(mat_tuples == -1)
		mat_tuples = no_tuples;
	for(int i = 0; i < no_dimensions; i++)
		if(f[i])
			dim_out[i] = true;
	MakeCountOnly(mat_tuples, dim_out);
}


void MultiIndex::MakeCountOnly(_int64 mat_tuples, DimensionVector &dims_to_materialize)
{
	if(index_type==IT_COUNT) {
		no_tuples_too_big = false;
		no_tuples = material_no_tuples = mat_tuples;
	}
	else
	{
		for(int i = 0; i < no_dimensions; i++)
			if(f[i] && dims_to_materialize[i]) {
				delete f[i];
				f[i] = NULL;
			}
		material_no_tuples = mat_tuples;
		index_type = IT_MIXED;		// for UpdateMultiIndex
		UpdateNoTuples();
		index_type = IT_COUNT;
	}
	for(int i=0;i<no_dimensions;i++) if(t[i])
	{
		delete t[i];
		t[i]=NULL;
	}
}

void MultiIndex::MakeDimensionNull(int dim)	// transform the given dimension into a table of null objects
{
	if( t[dim] ) {
		for( _int64 i=0; i<material_no_tuples; i++ )
			t[dim]->Set64( i, 0 );
	}
	else {
		if( material_no_tuples == 0 ) {
			material_no_tuples = 1;
			mat_max_no_tuples = RoundUpTuples( 1 );			// add some unused tuples for further enlargement
		}
		int bytes_per_value = ( (OrigSize(dim)+1)>0xFFFF ? ((OrigSize(dim)+1)>0xFFFFFFFF ? 8 : 4 ) : 2);
		t[dim] = new IndexTable(mat_max_no_tuples , bytes_per_value, 0);
		if( f[dim] ) {
			t[dim]->CopyLocksToIT(_noOpLocks[dim]);
			delete f[dim];
			f[dim] = NULL;
		}
		for( _int64 i=0; i<material_no_tuples; i++ )
			t[dim]->Set64( i, 0 );
		index_type = IT_MIXED;		// note: may be switched to IT_ORDERED inside UpdateNoTuples()
		UpdateNoTuples();
	}
}

void MultiIndex::UpdateNoTuples()
{
	// assumptions:
	// - no_material_tuples is correct, even if all t[...] are NULL (forgotten).
	//   However, if all f[...] are not NULL, then the index is set to IT_VIRTUAL and no_material_tuples = 0
	// - IT_MIXED or IT_VIRTUAL may be in fact IT_ORDERED (must be set properly on output)
	// - if no_material_tuples > 0, then new index_type is not IT_VIRTUAL
	no_tuples_too_big = false;
	if(index_type == IT_VIRTUAL && (no_dimensions == 0 || f[0] == NULL))	// empty index - WHEN IS IT USED?
	{
		no_tuples=0;
		material_no_tuples=0;
	}
	else if(index_type == IT_MIXED || index_type == IT_VIRTUAL)
	{
		no_tuples=1;
		_uint64 loc_no_tuples;
		bool virtual_found = false;
		bool material_found = false;
		for(int d=0;d<no_dimensions;d++)
			if(f[d])
			{
				virtual_found = true;
				loc_no_tuples = f[d]->NoOnes();

				MultiplyNoTuples( loc_no_tuples );
				if( loc_no_tuples == 0 )
					break;
			}
			else {
				material_found = true;
/*				if( index_type == IT_VIRTUAL ) {
					no_tuples = 0;
					no_tuples_too_big = false;
				}*/
			}
		if(material_found) {
			index_type = IT_MIXED;
			if( !virtual_found )
				index_type = IT_ORDERED;
			MultiplyNoTuples(material_no_tuples );	// we have to assume that material_no_tuples is correct
		}
		else {
			material_no_tuples = 0;
			index_type = IT_VIRTUAL;
		}
	}
	else if(index_type == IT_ORDERED)
	{
		no_tuples_too_big = false;
		no_tuples = material_no_tuples;		// no other way to determine the real number of tuples
	}	// else IT_COUNT
}

_uint64 MultiIndex::RoundUpTuples(_uint64 n)
{
	if(n<256) return 256;
	if(n<2048) return 2048;
	n=(n>>16)+1;
	return n<<16;		// rounded up for the next 65536
}

//////////////////////////////// Logical operations //////////////////////////////

void MultiIndex::MIFilterAnd(MIIterator &mit, Filter &fd)	// limit the MultiIndex by excluding all tuples which are not present in fd, in order given by mit
{
	assert(index_type!=IT_COUNT);
	if(index_type == IT_VIRTUAL && no_dimensions == 1) {
		FilterOnesIterator fit(f[0]);
		_int64 cur_pos = 0;
		while(fit.IsValid()) {
			if(!fd.Get(cur_pos))
				f[0]->ResetDelayed(*fit);
			++fit;
			cur_pos++;
		}
		f[0]->Commit();
		UpdateNoTuples();
		return;
	}

	vector<int>  dim_order = mit.DimOrder();
	DimensionVector dim_used(this);
	vector<IndexTable*> new_t;
	for(int i = 0; i < no_dimensions; i++)
		new_t.push_back(NULL);

	bool material_included = false;
	for(vector<int>::iterator dit = dim_order.begin(); dit != dim_order.end(); ++dit) {		// iterate through dimensions
		if(t[*dit])
			material_included = true;
		dim_used[*dit] = true;
	}
	if(material_included)
		for(int i = 0; i < no_dimensions; i++)
			if(t[i])
				dim_used[i] = true;

	_uint64 new_multiplier = fd.NoOnes();
	_uint64 new_mat_no_tuples = new_multiplier;
	_uint64 mat_multiplier = 1;
	if(!material_included && material_no_tuples > 0) {
		// there exists a materialized part, but not included in iterator => must be multiplied!
		mat_multiplier = material_no_tuples;
		new_mat_no_tuples = SafeMultiplication(new_mat_no_tuples, mat_multiplier);
		if(new_mat_no_tuples == NULL_VALUE_64) {
			no_tuples_too_big = true;
			no_tuples = NULL_VALUE_64;
			return;
		}
	}
	_uint64 new_mat_max_no_tuples = RoundUpTuples(new_mat_no_tuples);			// add some unused tuples for further enlargement
	_int64 iter_obj, loc_obj;

	// first phase: rewrite used dims
	for(int d = 0; d < no_dimensions; d++)
		if(dim_used[d] && !forget_now[d]) {
			loc_obj = 0;
			int byte_width = 0;
			if(t[d]) {
				byte_width = t[d]->BytesPerValue();
			} else { // filter-based dimension
				byte_width = ((f[d]->NoObj()+1)>0xFFFF ? ((f[d]->NoObj()+1)>0xFFFFFFFF ? 8 : 4 ) : 2);
			}
			new_t[d] = new IndexTable(new_mat_max_no_tuples , byte_width, 0);
			if(t[d])
				new_t[d]->CopyLockStatus(*t[d]);
			else
				new_t[d]->CopyLocksToIT(_noOpLocks[d]);

			for(_uint64 pass = 0; pass < mat_multiplier; pass++) {		// in case of multipliers, make it abcdabcdabcd
				mit.Rewind();
				iter_obj = 0;
				while(mit.IsValid()) {
					if(fd.Get(iter_obj)) {
						new_t[d]->Set64(loc_obj, (mit[d] == NULL_VALUE_64 ? 0 : mit[d] + 1));
						loc_obj++;
					}
					++mit;
					iter_obj++;
				}
			}
		}
	// second phase: multiply unused material dims, if any
	if(mat_multiplier > 1)
		for(int d = 0; d < no_dimensions; d++)
			if(!dim_used[d] && !forget_now[d] && t[d]) {
				loc_obj = 0;
				t[d]->Lock();
				int byte_width = t[d]->BytesPerValue();
				new_t[d] = new IndexTable(new_mat_max_no_tuples , byte_width, 0);
				new_t[d]->CopyLockStatus(*t[d]);
				// in case of multipliers, make it aaaabbbbcccdddd
				for(_int64 orig_obj = 0; orig_obj < mat_multiplier; orig_obj++) {
					for(_uint64 pass = 0; pass < new_multiplier; pass++) {
						new_t[d]->Set64(loc_obj, t[d]->Get64(orig_obj));
						loc_obj++;
					}
				}
				t[d]->Unlock();
			}
	// third phase: replace and cleanup
	for(int d = 0; d < no_dimensions; d++) {
		if(t[d]) {
			delete t[d];
			t[d] = new_t[d];
		}
		if(f[d] && dim_used[d]) {
			delete f[d];
			f[d] = NULL;
			t[d] = new_t[d];
		}
	}
	material_no_tuples = new_mat_no_tuples;
	index_type = IT_MIXED;		// note: may be switched to IT_ORDERED inside UpdateNoTuples()
	UpdateNoTuples();
}

void MultiIndex::DontForget()					// Revoke current optimizations (dims may be still needed)
{
	for(int i = 0; i < no_dimensions; i++)
		if(forget_now[i] && (t[i] || f[i]))		// not forgotten yet?
			forget_now[i] = false;
}


void MultiIndex::MarkTuplesPresent(::Filter &f, MultiIndex &sec)	// f - a filter (empty by assumption) equal in size with sec iterated by default MIIterator
{																	// result: mark in f all tuples of sec which are present in this.
	MIIterator mit(this);
	MIIterator mit_sec(&sec);
	Filter read_again(mit.NoTuples());
	read_again.Set();
	GroupDistinctTable dist_table;
	dist_table.InitializeB(sizeof(_int64) * NoDimensions(), mit.NoTuples());
	do {
		mit.Rewind();
		_int64 position = 0;
		dist_table.Clear();
		while(mit.IsValid()) {
			if(read_again.Get(position)) {
				GDTResult res = dist_table.Add((unsigned char*)(*mit));
				if(res != GDT_FULL)
					read_again.Reset(position);
			}
			++mit;
			position++;
			if(m_conn->killed())			// from time to time...
				throw KilledRCException();
		}
		mit_sec.Rewind();
		position = 0;
		while(mit_sec.IsValid()) {
			GDTResult res = dist_table.Add((unsigned char*)(*mit_sec));	// using Add() to check whether the value was added
			if(res == GDT_EXISTS)
				f.Set(position);
			++mit_sec;
			position++;
			if(m_conn->killed())			// from time to time...
				throw KilledRCException();
		}
	} while(!read_again.IsEmpty());
}

void MultiIndex::Not()
{
	assert(index_type!=IT_COUNT);
	if(no_tuples == 0)
	{
		for(int i = 0; i < no_dimensions; i++)
			if(dim_size[i] == 0)
				return;									// cross product contains the empty set => NOT is trivial
		for(int i = 0; i < no_dimensions; i++)
		{
			if(f[i]==NULL)
			{
				f[i]=new Filter(dim_size[i]);
			}
			f[i]->Set();
			if(t[i])
			{
				delete t[i];
			}
		}
		index_type = IT_VIRTUAL;
		material_no_tuples = 0;
		UpdateNoTuples();
		return;
	}
	bool simple=true;		// simple case: only one nontrivial dimension and it is virtual
	bool was_nontrivial=false;
	int nontrivial=-1;
	for(int i=0;i<no_dimensions;i++)
	{
		if(t[i])
		{
			simple=false;
			break;
		}
		if(f[i])
		{
			if(f[i]->IsFull()==false && was_nontrivial)
			{
				simple=false;
				break;
			}
			if(f[i]->IsFull()==false)
			{
				nontrivial=i;
				was_nontrivial=true;
			}
		}
	}
	if(simple && was_nontrivial==false)		// i.e. all f[i] were full
	{
		Empty();
		return;
	}
	if(simple)
	{
		assert(nontrivial>-1);
		f[nontrivial]->Not();
		material_no_tuples = 0;
		UpdateNoTuples();
	}
	else
	{
		// TODO: use a similar mechanism like for Or, And:
		// MultiIndex new_mind = the same dimensions, but full
		// MarkTuplesPresent(filter, new_mind);
		// filter.Not();
		// create new t[d] basing on new_mind filtered by filter
		throw NotImplementedRCException("NOT is too complex.");
	}
}

void MultiIndex::Or(MultiIndex &s)	// enlarge the multiindex by adding all tuples from the second one (excluding repetitions)
{
	assert(index_type!=IT_COUNT);
	// check compatibility
	if(no_dimensions!=s.no_dimensions)
		throw NotImplementedRCException("MultiIndex::Or() incompatible arguments.");
	if(s.ZeroTuples())
		return;						// empty, do not add anything
	if(  ZeroTuples()) {
		Copy(s);
		return;
	}
	//////////// simplified version: virtual dimension only

	if(index_type == IT_VIRTUAL && s.index_type == IT_VIRTUAL)	{
		if(no_dimensions==1) {
			assert(f[0] && s.f[0]);
			LockAllForUse();
			s.LockAllForUse();
			f[0]->Or(*s.f[0]);
			no_tuples = f[0]->NoOnes();
			UnlockAllFromUse();
			s.UnlockAllFromUse();
			return;
		} else {
			// check whether all filters but one are equal, in this case make Or() on the differing filters and left the rest of them unchanged
			int dim_different = -1;
			bool simple_or = true;
			LockAllForUse();
			s.LockAllForUse();
			for(int i=0;i<no_dimensions;i++) if(f[i]) {
				if( f[i]->IsEqual(*(s.f[i]))==false ) {
					if(dim_different!=-1) {
						simple_or = false;
						break;
					}
					dim_different = i;
				}
			}
			if(simple_or && dim_different!=-1)		// sometimes all are just equal and nothing is to be changed
			{
				f[dim_different]->Or(*s.f[dim_different]);
				UpdateNoTuples();
			}
			UnlockAllFromUse();
			s.UnlockAllFromUse();
			if(simple_or)
				return;			// done!
		}
	}
	//////////// make a filter of non-repeated objects

	_int64 approx_size = no_tuples + s.no_tuples;
	if(approx_size > MATERIAL_TUPLES_LIMIT)
		throw OutOfMemoryRCException("Attempting to create too large intermediate object.");
	if(approx_size > MATERIAL_TUPLES_WARNING)
		rccontrol.lock(m_conn->GetThreadID()) << "Warning: a big intermediate object created." << unlock;

	Filter fs(s.no_tuples);
	fs.Reset();
	MarkTuplesPresent(fs, s);
	fs.Not();		// now fs points to new tuples (to be added by Or)

	////////// enlarge the MultiIndex by tuples pointed by a filter

	Materialize();			// in case we did not make it earlier
	_uint64 new_no_tuples = no_tuples + fs.NoOnes();
	_uint64 new_mat_max_no_tuples = RoundUpTuples(new_no_tuples);			// add some unused tuples for further enlargement
	for(int d = 0; d < no_dimensions; d++) if(t[d])	{
		t[d]->Lock();
		t[d]->ExpandTo(new_mat_max_no_tuples);
		MIIterator mit(&s);
		_uint64 cur_obj = no_tuples;
		_int64 position = 0;
		while(mit.IsValid()) {
			if(fs.Get(position)) {
				t[d]->Set64(cur_obj, (mit[d] == NULL_VALUE_64 ? 0 : mit[d] + 1));
				cur_obj++;
			}
			++mit;
			position++;
		}
		t[d]->Unlock();
		can_be_distinct[d] = (can_be_distinct[d] && s.can_be_distinct[d]);
	}
	mat_max_no_tuples = new_mat_max_no_tuples;
	no_tuples = new_no_tuples;
	material_no_tuples = no_tuples;
}

void MultiIndex::And(MultiIndex &s)	// limit the multiindex by excluding all tuples from the second one
{
	assert(index_type!=IT_COUNT);
	// check compatibility
	if(no_dimensions!=s.no_dimensions)
	{
		throw NotImplementedRCException("MultiIndex::And() incompatible arguments.");
		assert(0);
		return;		// incompatible
	}
	if( ZeroTuples() || s.ZeroTuples() )
	{
		Empty();	// empty set
		return;
	}

	if( index_type==IT_VIRTUAL && s.index_type==IT_VIRTUAL )
	{
		no_tuples_too_big = false;
		no_tuples = 1;
		for(int i=0;i<no_dimensions;i++)		// virtual case: just cross the filters
		{
			if(f[i] && s.f[i])
			{
				f[i]->And(*s.f[i]);
				MultiplyNoTuples( f[i]->NoOnes() );
				if( ZeroTuples() ) break;
			}
			else
				assert(f[i]==NULL && s.f[i]==NULL);		// both must be present or not present in the same time!
		}
		if( ZeroTuples() ) Empty();				// delete all other filters
		return;
	}

	//////////// make a filter of non-repeated objects

	Materialize();
	Filter fs(no_tuples);
	fs.Reset();
	s.MarkTuplesPresent(fs, *this);
	MIIterator mit(this);
	MIFilterAnd(mit, fs);	// leave only tuples present in the second multiindex
}

/////////////////////////////////////////////////////////////////////////////////////
/*
void MultiIndex::AddNewMaterializedTuple(int dim, _int64 val)
// extends all mat. dimensions by one row, set 0 everywhere except one dimension
{
	_int64 obj = material_no_tuples;
	if( obj==0 ) {
		for(int i = 0; i < no_dimensions; i++ )				// find empty dimensions and transform them into t[]
			if( !forget_now[ i ] && f[i] && f[i]->NoOnes()==0 )
				MakeDimensionNull(i);						// inside material_no_tuples is set to 1
		material_no_tuples = 1;								// just in case it is not set above
	}
	else
		material_no_tuples++;
	for( int i=0; i<no_dimensions; i++ ) if( t[ i ] && !forget_now[ i ]){
		if( obj>=t[ i ]->N() )
		{
			t[ i ]->ExpandTo( obj<2048 ? 2048 : _int64(obj*2) );	// enlarge with safe backup
		}
		t[ i ]->Set64( obj, 0 );			// null objects for all dimensions...
	}
	if( t[ dim ] && !forget_now[ dim ] && val != NULL_VALUE_64 )
		t[ dim ]->Set64( obj, val+1 );		// ...except this one. Note that val is a "natural" row number, so val = 0 is not a null
}

void MultiIndex::AddNewMaterializedTuple(vector<int> dims, vector<_uint64> row)
{
	_int64 obj = material_no_tuples;
	if( obj==0 ) {
		for(int i = 0; i < no_dimensions; i++ )				// find empty dimensions and transform them into t[]
			if( !forget_now[ i ] && f[i] && f[i]->NoOnes()==0 )
				MakeDimensionNull(i);						// inside material_no_tuples is set to 1
		material_no_tuples = 1;								// just in case it is not set above
	}
	else
		material_no_tuples++;

	for(int i = 0; i < no_dimensions; i++ ) {
		if( t[ i ] && !forget_now[ i ]) {
			if( obj>=t[ i ]->N() ) {
				t[ i ]->ExpandTo( obj<2048 ? 2048 : _int64(obj*2) );	// enlarge with safe backup
			}
			t[ i ]->Set64( obj, 0 );			// null objects for all dimensions...
		}
	}
	int cur_dim = 0;
	for(int i = 0; i < dims.size(); i++ ) {
		cur_dim = dims[i];
		if( t[cur_dim] && !forget_now[cur_dim] )
			t[cur_dim]->Set64( obj, row[i] );	// note that row[i] = 0 for null, i.e. this value should be taken fom t[]->Get64
	}
}
*/
void MultiIndex::AddNewMaterializedTuple(DimensionVector &dims, const _int64 *row)
{
	if(m_conn->killed())
		throw KilledRCException();
	_int64 obj = material_no_tuples;
	if(obj == 0) {
		for(int i = 0; i < no_dimensions; i++ )				// find empty dimensions and transform them into t[]
			if(!forget_now[i] && f[i] && f[i]->NoOnes()==0 && dims[i])
				MakeDimensionNull(i);						// inside material_no_tuples is set to 1
		material_no_tuples = 1;								// just in case it is not set above
	}
	else
		material_no_tuples++;

	for(int i = 0; i < no_dimensions; i++ ) {
		if(t[i] && !forget_now[i]) {
			if(obj >= t[i]->N()) {
				t[i]->ExpandTo(obj < 2048 ? 2048 : _int64(obj * 2));	// enlarge with safe backup
			}
			if(dims[i] && row[i] != NULL_VALUE_64)
				t[i]->Set64(obj, row[i] + 1);	// note that 0 means null in t[i], and actual numbers start with 1
			else
				t[i]->Set64(obj, 0);		// null objects for all other dimensions
		}
	}
}

bool MultiIndex::MarkInvolvedMaterialClasses(DimensionVector &v)	// if any materialized dimension is marked, then mark the rest of this class
{
	bool mat_found = false;
	bool added = false;
	for(int i = 0; i < no_dimensions; i++)
		if(v[i] && t[i]) {
			mat_found = true;
			break;
		}
	if(mat_found)
		for(int i = 0; i < no_dimensions; i++)
			if(t[i] && !v[i]) {
				added = true;
				v[i] = true;
			}
	return added;
}

void MultiIndex::MarkMaterialClasses(DimensionVector &v)
{
	for(int i = 0; i < no_dimensions; i++)
		if(t[i])
			v[i] = true;
}

bool MultiIndex::ContainsMaterialClasses(DimensionVector &v)
{
	for(int i = 0; i < no_dimensions; i++)
		if(t[i] && v[i])
			return true;
	return false;
}

//////////////////////////////////////////////////////////////////////////////////////
DimensionVector::DimensionVector() : size(0), v(NULL)
{}

DimensionVector::DimensionVector(MultiIndex *mind) : v(NULL)
{
	assert(mind);
	size = mind->NoDimensions();
	if(size > 0)
		v = new bool [size];
	for(int i = 0; i < size; i++)
		v[i] = false;
}

DimensionVector::DimensionVector(int no_dims) : v(NULL)
{
	size = no_dims;
	if(size > 0)
		v = new bool [size];
	for(int i = 0; i < size; i++)
		v[i] = false;
}

DimensionVector::DimensionVector(const DimensionVector &sec) : v(NULL)
{
	size = sec.size;
	if(size > 0)
		v = new bool [size];
	for(int i = 0; i < size; i++)
		v[i] = sec.v[i];
}

DimensionVector::~DimensionVector()
{
	delete [] v;
}

DimensionVector &DimensionVector::operator=(const DimensionVector &sec)
{
	if(&sec != this) {
		if(size != sec.size) {
			size = sec.size;
			delete [] v;
			v = (size > 0) ? new bool [size] : NULL;
		}
		for(int i = 0; i < size; i++)
			v[i] = sec.v[i];
	}
	return *this;
}

void DimensionVector::Clean()
{
	for(int i = 0; i < size; i++)
		v[i] = false;
}

void DimensionVector::SetAll()
{
	for(int i = 0; i < size; i++)
		v[i] = true;
}

bool DimensionVector::Intersects(DimensionVector &sec)
{
	assert(size == sec.size);
	for(int i = 0; i < size; i++)
		if(v[i] && sec.v[i])
			return true;
	return false;
}

void DimensionVector::Minus(DimensionVector &sec)
{
	assert(size == sec.size);
	for(int i = 0; i < size; i++)
		if(sec.v[i])
			v[i] = false;
}

void DimensionVector::Plus(DimensionVector &sec)
{
	assert(size == sec.size);
	for(int i = 0; i < size; i++)
		if(sec.v[i])
			v[i] = true;
}

bool DimensionVector::Includes(DimensionVector &sec)
{
	assert(size == sec.size);
	for(int i = 0; i < size; i++)
		if(sec.v[i] && !v[i])
			return false;
	return true;
}

int DimensionVector::NoDimsUsed()						// return a number of present dimensions
{
	int res = 0;
	for(int i = 0; i < size; i++)
		if(v[i])
			res++;
	return res;
}

//////////////////////////////////////////////////////////////////////////////////////

IndexTable::IndexTable(_int64 _size, int _bytes_per_value, int mem_modifier)
	:	CacheableItem("JW", "INT"), m_conn(ConnectionInfoOnTLS.Get())
{
    ConstructionLock();
	// Note: buffer size should be 2^n
	int mem_scale = MemorySettingsScale() + mem_modifier;
	if(mem_scale<=0)		// low settings (below 0.5 GB)
		max_buffer_size_in_bytes = 32 * 1024 * 1024;				// 32 MB = 2^25
	else if(mem_scale==1 || mem_scale==2)		// (0.5 - 2.5 GB)
		max_buffer_size_in_bytes = 64 * 1024 * 1024;
	else if(mem_scale==3)		// high settings (2.5 - 5 GB)
		max_buffer_size_in_bytes = 128 * 1024 * 1024;				// 128 MB = 2^27
	else if(mem_scale==4)		// high settings (5 - 10 GB)
		max_buffer_size_in_bytes = 256 * 1024 * 1024;
	else						// high settings (when at least 10 GB is reserved)
		max_buffer_size_in_bytes = 512 * 1024 * 1024;

	buffer_size_in_bytes = max_buffer_size_in_bytes;
	CI_SetDefaultSize(max_buffer_size_in_bytes);
	size = _size;
	bytes_per_value = _bytes_per_value;					// values allowed: 2, 4, 8
	assert(bytes_per_value==2 || bytes_per_value==4 || bytes_per_value==8);
	if( size == 0 )
		size = 1;
	unsigned int values_per_block = buffer_size_in_bytes / bytes_per_value;
	block_shift = CalculateBinSize(values_per_block)-1;	// e.g. BinSize(16)==5, but shift by 4. WARNING: should it be (val...-1)? Now it works, because v... is only 2^25, 2^24, 2^23
	block_mask = (_uint64(1)<<block_shift)-1;

	if(size*bytes_per_value<buffer_size_in_bytes)
		buffer_size_in_bytes = int(size*bytes_per_value);		// the whole table in one buffer
	buf = (unsigned char*)alloc(buffer_size_in_bytes, BLOCK_UNCOMPRESSED);
	memset(buf, 0, buffer_size_in_bytes);
	cur_block = 0;
	max_block_used = 0;
	block_changed = false;
    ConstructionUnlock();
}

IndexTable::IndexTable(IndexTable &sec)
	:	CacheableItem("JW", "INT"), max_buffer_size_in_bytes(sec.max_buffer_size_in_bytes),
		bytes_per_value(sec.bytes_per_value), max_block_used(sec.max_block_used),
 		buffer_size_in_bytes(sec.buffer_size_in_bytes),	block_shift(sec.block_shift),
		block_mask(sec.block_mask), size(sec.size), cur_block(0),
		block_changed(false), m_conn(sec.m_conn)
{
	Lock();
	sec.Lock();
	CI_SetDefaultSize(max_buffer_size_in_bytes);
	buf = (unsigned char*)alloc(buffer_size_in_bytes, BLOCK_UNCOMPRESSED);
	_int64 used_size = size;
	if(used_size > ((_int64(max_block_used) + 1) << block_shift))
		used_size = ((_int64(max_block_used) + 1) << block_shift);
	if (max_block_used == 0) {
		memcpy(buf,sec.buf, size*bytes_per_value);
	} else
		for(uint k = 0; k < used_size; k++)
			Set64(k, sec.Get64(k));
	sec.Unlock();
	Unlock();
}

IndexTable::~IndexTable()
{
	DestructionLock();
	dealloc(buf);
}

void IndexTable::Set64(_uint64 n,_uint64 val)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( IsLocked() );
	assert(n<size);	// warning C4018: '<' : signed/unsigned mismatch
	int b = int(n>>block_shift);
	if(b != cur_block)
		LoadBlock(b);
	block_changed = true;
	if(bytes_per_value==8)		((_uint64*)buf)[n & block_mask] 		   = val;
	else if(bytes_per_value==4)		((unsigned int*)buf)[n & block_mask]   = (unsigned int)val;
	else if(bytes_per_value==2)	((unsigned short*)buf)[n & block_mask] = (unsigned short)val;
	else
		assert(0);
}

void IndexTable::LoadBlock(int b)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( IsLocked() );
	if(buf==NULL) 									// possible after block caching on disk
	{
		buf = (unsigned char*)alloc(buffer_size_in_bytes, BLOCK_UNCOMPRESSED);
	}
	else if(block_changed)
		CI_Put(cur_block,buf);
	CI_Get(b,buf);
	if(m_conn.killed())			// from time to time...
		throw KilledRCException();
	max_block_used = max(max_block_used, b);
	cur_block = b;
	block_changed = false;
}

void IndexTable::ExpandTo(_int64 new_size)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT( IsLocked() );
	if(new_size<=size) return;
	if(size*bytes_per_value==_int64(buffer_size_in_bytes))			// the whole table was in one buffer
	{
		int new_buffer_size_in_bytes;
		if(new_size*bytes_per_value<max_buffer_size_in_bytes)
			new_buffer_size_in_bytes = int(new_size*bytes_per_value);
		else
			new_buffer_size_in_bytes = max_buffer_size_in_bytes;
		buf = (unsigned char*)rc_realloc( buf, new_buffer_size_in_bytes, BLOCK_UNCOMPRESSED );
		buffer_size_in_bytes = new_buffer_size_in_bytes;
	}
	// else: the table is buffered anyway, so we don't need to do anything!?
	size=new_size;
}

int IndexTable::CurrentSize()
{
	if(buf == NULL) return 100;
	return buffer_size_in_bytes + 100;
}

int IndexTable::Compress()
{
	return Collapse();			// the same operation - no special compressed mode
}

int IndexTable::Collapse()
{
	if(buf && block_changed)
		CI_Put(cur_block,buf);
	dealloc(buf);
	buf = NULL;
	cur_block = -1;
	block_changed = false;
	return 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////

MINewContents::MINewContents(MultiIndex *m) :	mind(m), t_new(NULL),
												new_cur_used(NULL), all_virtual(true),
												optimized_dim_stay(-1), f_opt(NULL)
{
	dim_involved = new bool [mind->NoDimensions()];
	t_new = new IndexTable * [mind->NoDimensions()];
	new_cur_used = new _int64 [mind->NoDimensions()];
	for(int i = 0; i < mind->NoDimensions(); i++) {
		t_new[i] = NULL;
		new_cur_used[i] = 0;
		dim_involved[i] = false;
	}
}

MINewContents::~MINewContents()
{
	for(int i = 0; i < mind->NoDimensions(); i++)
		delete t_new[i];
	delete [] t_new;
	delete [] new_cur_used;
	delete f_opt;
	delete [] dim_involved;
}

void MINewContents::SetDimension(int dim)			// mark this dimension as to be involved
{
	dim_involved[dim] = true;
	if(all_virtual && mind->GetIndTable(dim)) {		// one material involved => all material involved
		all_virtual = false;
		for(int i = 0; i < mind->NoDimensions(); i++)
			if(mind->GetIndTable(i))
				dim_involved[i] = true;
	}
}

void MINewContents::SetDimension(DimensionVector &dims)	// mark all these dimensions as to be involved
{
	bool material_not_involved = false;
	for(int i = 0; i < mind->NoDimensions(); i++) {
		if(dims[i])
			dim_involved[i] = true;
		if(dim_involved[i] && mind->GetIndTable(i))
			all_virtual = false;
		if(!dim_involved[i] && mind->GetIndTable(i))
			material_not_involved = true;
	}
	if(!all_virtual && material_not_involved) {		// one material involved => all material involved
		for(int i = 0; i < mind->NoDimensions(); i++)
			if(mind->GetIndTable(i))
				dim_involved[i] = true;
	}
}

void MINewContents::Init(_int64 initial_size)		// initialize temporary structures (set approximate size)
{
	// check for special (filter) case
	if(all_virtual) {
		for(int dim = 0; dim < mind->NoDimensions(); dim++)
			if(dim_involved[dim] && !mind->IsForgotten(dim)) {
				if(optimized_dim_stay != -1) {		// more than one unforgotten found
					optimized_dim_stay = -1;
					break;
				}
				optimized_dim_stay = dim;
			}	// optimized_dim_stay>-1 if there is exactly one unforgotten dimension
	}

	// general case
	for(int dim = 0; dim < mind->NoDimensions(); dim++)
		if(dim_involved[dim] || (!all_virtual && mind->GetIndTable(dim))) {
			// for all involved dimensions, or all materialized (except the case of both virtual)
			new_cur_used[dim] = 0;
			delete t_new[dim];
			t_new[dim] = NULL;
			if(mind->IsForgotten(dim))
				continue;
			if(dim != optimized_dim_stay)
				InitTnew(dim, initial_size);
			else {
				f_opt = new Filter(mind->GetFilter(dim)->NoObj());
				f_opt->Reset();
			}
		}
}

void MINewContents::InitTnew(int dim, _int64 initial_size)
{
	int bytes_per_value;
	::Filter *f = mind->GetFilter(dim);
	IndexTable *t = mind->GetIndTable(dim);
	if(f)
		bytes_per_value = ( (f->NoObj()+1)>0xFFFF ? ((f->NoObj()+1)>0xFFFFFFFF ? 8 : 4 ) : 2);
	else
		bytes_per_value = t->BytesPerValue();

	if( initial_size < 8 )
		initial_size = 8;
	t_new[dim] = new IndexTable(initial_size, bytes_per_value, 0);
	if(f)
		t_new[dim]->CopyLocksToIT(mind->_noOpLocks[dim]);
	else {
		if(t)
			t_new[dim]->CopyLockStatus( *t );
	}
	new_cur_used[dim] = 0;
}

void MINewContents::CommitCountOnly(_int64 joined_tuples)		// commit changes to multiindex - must be called at the end, or changes will be lost
{
	DimensionVector dim_out(mind);
	for(int dim = 0; dim < mind->NoDimensions(); dim++)
		if(dim_involved[dim] && mind->GetIndTable(dim) == NULL)
			dim_out[dim] = true;
	if(all_virtual && mind->NoMaterialTuples() > 0) // multiply by the existing (additional) materialized dimensions, if any
		joined_tuples *= mind->NoMaterialTuples();
	mind->MakeCountOnly(joined_tuples, dim_out);	// inside: UpdateNoTuples();
}

void MINewContents::Commit(_int64 joined_tuples)		// commit changes to multiindex - must be called at the end, or changes will be lost
{
	_int64 multiply_by = 1;
	if(all_virtual) {		// joining two virtual to some other already materialized dimensions
		multiply_by = mind->NoMaterialTuples();
		if(multiply_by < 1)
			multiply_by = 1;
	}

	if(optimized_dim_stay != -1) {	// optimized version: just exchange filters
		for(int dim = 0; dim < mind->NoDimensions(); dim++) {
			if(dim_involved[dim] && (mind->IsForgotten(dim) || dim == optimized_dim_stay)) {
				delete mind->t[dim];
				delete mind->f[dim];
				mind->f[dim] = NULL;
				mind->t[dim] = NULL;
			}
		}
		mind->f[optimized_dim_stay] = f_opt;
		f_opt = NULL;
		mind->index_type = mind->IT_MIXED;
		mind->material_no_tuples = multiply_by;		// special version: leave one object in the non-existing t_new, to be logically multiplied by f_opt
		mind->UpdateNoTuples();
		return;
	}
	// normal version: find all unrelated already materialized dimensions and multiply them
	if(all_virtual) {		// joining two virtual to some other already materialized dimensions
		for(int dim = 0; dim < mind->NoDimensions(); dim++)
			if(mind->t[dim] && t_new[dim] == NULL)
			{
				if(mind->IsForgotten(dim))	{
					delete mind->t[dim];
					mind->t[dim] = NULL;
					continue;
				}
				mind->t[dim]->Lock();
				t_new[dim] = mind->t[dim];
				mind->t[dim] = new IndexTable(multiply_by * joined_tuples,
											t_new[dim]->BytesPerValue(),
											0);
				mind->t[dim]->CopyLockStatus(*t_new[dim]);
				_int64 i = 0;
				for(_int64 k = 0; k < mind->NoMaterialTuples(); k++)
					for(_int64 j = 0; j < joined_tuples; j++)
						mind->t[dim]->Set64(i++, t_new[dim]->Get64(k));		// abcd  ->  aaaabbbbccccdddd
				delete t_new[dim];
				t_new[dim] = NULL;
				mind->t[dim]->Unlock();
			}
	}
	// now we should exchange existing joined dimensions into the newly calculated ones
	mind->index_type = mind->IT_MIXED;
	mind->material_no_tuples = joined_tuples * multiply_by;
	for(int dim = 0; dim < mind->NoDimensions(); dim++) {
		if(dim_involved[dim] || (mind->t[dim] && mind->IsForgotten(dim)))
		{
			delete mind->t[dim];
			delete mind->f[dim];
			mind->f[dim] = NULL;
			mind->t[dim] = NULL;
			if(mind->IsForgotten(dim))
				continue;
			assert(new_cur_used[dim] == joined_tuples);
			if(multiply_by>1) {
				mind->t[dim] = new IndexTable(	mind->material_no_tuples,
												t_new[dim]->BytesPerValue(),
												0);
				mind->t[dim]->CopyLockStatus(*t_new[dim]);
				_int64 i=0;
				for(_int64 k = 0; k < multiply_by; k++)	// multiplication in case of existing other (previously materialized) dimensions
					for(_int64 j = 0; j < joined_tuples; j++)
						mind->t[dim]->Set64(i++, t_new[dim]->Get64(j));		// abcd  ->  abcdabcdabcdabcdabcd
				delete t_new[dim];
			}
			else {
				mind->t[dim] = t_new[dim];
			}
			t_new[dim] = NULL;
			// note that we assume that all materialized dimensions are updated at the same time
			// or that the original material_no_tuples was read and stored externally before the first update
		}
	}
	mind->UpdateNoTuples();
}

void MINewContents::DisableOptimized()
{
	assert(optimized_dim_stay != -1);
	if(!mind->IsForgotten(optimized_dim_stay)) {
		InitTnew(optimized_dim_stay, f_opt->NoObj());
		FilterOnesIterator it(f_opt);
		_int64 obj = 0;
		while(it.IsValid()) {
			t_new[optimized_dim_stay]->Set64(obj++, *it + 1);
			++it;
		}
		new_cur_used[optimized_dim_stay] = obj;
	} else
		new_cur_used[optimized_dim_stay] = f_opt->NoOnes();
	delete f_opt;
	f_opt = NULL;
	optimized_dim_stay = -1;
}

void MINewContents::SetNewTableValue(int dim, _int64 val)	// set a value (NULL_VALUE_64 is a null object index); if the index is larger than the current size, enlarge table automatically
{
	if(dim == optimized_dim_stay) {
		if(val == NULL_VALUE_64 || f_opt->Get(val)) {	// repetition or null object found
			DisableOptimized();
			SetNewTableValue(dim, val);					// set again, this time to t_new
		}
		else
			f_opt->Set(val);
		return;
	}
	_int64 obj = new_cur_used[dim];
	new_cur_used[dim]++;
	if(!t_new[dim])						// == forget_now[dim], but it is faster (not a vector)
		return;
	if(obj >= t_new[dim]->N()) {
		t_new[dim]->ExpandTo(obj<2048 ? 2048 : _int64(obj * 2));	// enlarge with safe backup
	}
	if(val == NULL_VALUE_64)
		t_new[dim]->Set64(obj, 0);
	else
		t_new[dim]->Set64(obj, val + 1);
}
