/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "OuterJoiner.h"
#include "common/bhassert.h"
#include "CQTerm.h"
#include "MIUpdatingIterator.h"
#include "core/tools.h"

using namespace std;

OuterJoiner::OuterJoiner(MultiIndex& mind, vector<Descriptor>& desc, bool _count_only)
	: left_dims(&mind), old_mind(mind)	// copy Mulittindex
{
	bool is_null_only = (desc[0].op == O_IS_NULL);	// all the conditions are "IS NULL"
	for(int i = 0; i < desc.size(); i++) {
		desc[i].DimensionUsed(left_dims);		// mark all dims used in conditions
		if(desc[i].op != O_IS_NULL)
			is_null_only = false;

//		if(desc[i].NullMayBeTrue())
//			throw NotImplementedRCException("Outer join for this operator is not implemented");
	}
	outer_dim = desc[0].outer_dim;
	assert(outer_dim != -1);
	left_dims[outer_dim] = false;

	// now left_dims contain only relevant (used in conditions) left side dimensions

	reference_dim = -1;		// try to find reference_dim (see the header for algorithm description)
	no_of_left_dims = left_dims.NoDimsUsed();
	add_old_nulls = false;	// do not add old nulls until proven we need it

	if(no_of_left_dims == 1) {
		for(int i = 0; i < old_mind.NoDimensions(); i++)
			if(left_dims[i])
				reference_dim = i;
		reference_filter = FilterAutoPtr(new Filter(old_mind.OrigSize(reference_dim)));
		reference_filter->Reset();
		old_mind.SetFilter(reference_filter.get(), reference_dim, add_old_nulls);
		if(is_null_only)
			add_old_nulls = false;				// treat nulls as meeting join condition (=> don't add after join)
//		for(int i = 0; i < outer_dim; i++)
//			left_dims[i] = true;
		old_mind.MarkInvolvedMaterialClasses(left_dims);		// all left side dimensions

	} else {
		// try to find a key dimension (the larger one)
		_int64 max_dim_size = 0;
		int max_dim = -1;
		reference_dim = -1;
		for(int i = 0; i < outer_dim; i++)
			left_dims[i] = true;
		old_mind.MarkInvolvedMaterialClasses(left_dims);		// all left side dimensions
		for(int i = 0; i < old_mind.NoDimensions(); i++)
			if(left_dims[i]) {
				if(old_mind.OrigSize(i) > max_dim_size) {
					max_dim_size = old_mind.OrigSize(i);
					max_dim = i;
				}
				if(left_dims.NoDimsUsed() > 1 && old_mind.GetFilter(i) && old_mind.GetFilter(i)->NoOnes() > 1)
					// Mixed virtual and materialized dimensions
					throw NotImplementedRCException("Outer join is too complex");
			}
		if(max_dim > -1) {				// take the largest dimension and try to use it as a key
			reference_dim = max_dim;
			reference_filter = FilterAutoPtr(new Filter(max_dim_size));
			reference_filter->Reset();
			bool dim_unique = old_mind.SetFilterUnique(reference_filter.get(), reference_dim);
			if(!dim_unique)	{	// not a key
				for(int i = 0; i < old_mind.NoDimensions(); i++)
					if(i != max_dim && left_dims[i]) {
						reference_dim = i;
						reference_filter = FilterAutoPtr(new Filter(old_mind.OrigSize(i)));
						reference_filter->Reset();
						if(dim_unique = old_mind.SetFilterUnique(reference_filter.get(), reference_dim))
							break;
					}
				if(!dim_unique)
					throw NotImplementedRCException("Outer join is too complex");
			}
		} else
			throw NotImplementedRCException("Outer join is too complex");
	}
	no_of_left_dims = left_dims.NoDimsUsed();	// again, this time including all additional materialized dims
	count_only = _count_only;
}

OuterJoiner::~OuterJoiner()
{
}

void OuterJoiner::AddOuterTuples(MultiIndex &new_mind)
{
	_int64 tuples_for_count_only = 0;
	if(count_only)
		tuples_for_count_only = new_mind.NoTuples();

	// Analyze what to add
	if(reference_dim != -1 && !(new_mind.GetFilter(outer_dim) && new_mind.GetFilter(outer_dim)->IsEmpty())) { 
		// special cases: a) result is a product of nonempty and empty filter (=> reference_filter should be full)
		//				  b) some other dimensions are empty (=> normal reference_filter generation)
		MIIterator mit(&new_mind, reference_dim);
		while(mit.IsValid()) {
			if(mit[reference_dim] != NULL_VALUE_64)
				reference_filter->Reset(mit[reference_dim]);
			++mit;
		}
	}

	// now reference_filter contains only rows which are missing after join
	if(reference_dim != -1 && (add_old_nulls || !reference_filter->IsEmpty())) {

		// prepare a set of dimensions to be materialized and added
		DimensionVector dims_to_materialize(left_dims);
		dims_to_materialize[outer_dim] = true;
		// special case (smk_10): (t1 left join t2) join (t3 left join t4)
		new_mind.MarkMaterialClasses(dims_to_materialize);		// materialize together with any existing mat. dims
		new_mind.Materialize(dims_to_materialize);

		// now all the involved dimensions are already materialized
		if(old_mind.GetIndTable(outer_dim) == NULL)
			old_mind.MarkMaterialClasses(left_dims);		// iterate also all additional materialized dims
		MIIterator mit(&old_mind, left_dims);
		new_mind.LockAllForUse();
		while(mit.IsValid()) {
			_uint64 old_obj = mit[reference_dim];
			if((add_old_nulls && old_obj == NULL_VALUE_64) || 	// will not work for IS NULL, but otherwise we may add
					(old_obj != NULL_VALUE_64 && reference_filter->Get(old_obj))) {
				tuples_for_count_only++;
				// add this tuple
				new_mind.AddNewMaterializedTuple(left_dims, *mit);
			}
			++mit;
		}
		new_mind.UnlockAllFromUse();
		if(count_only) {			// multiply by omitted dims
			for(int d = 0; d < new_mind.NoDimensions(); d++)
				if(!dims_to_materialize[d] && new_mind.GetFilter(d))
					tuples_for_count_only *= new_mind.GetFilter(d)->NoOnes();
		}
	}

	if(count_only)
		new_mind.MakeCountOnly(tuples_for_count_only);
}

///////////////////////////////////////////////////////////////////////////////
/*

_int64 OuterJoiner::AddOuterTuples(int dim)
{
	MEASURE_FET("OuterJoiner::AddOuterTuples(...)");
	_int64 added_tuples = 0;
	int orig_dim = dims_info[dim].dim_no;
	// Analyze what to add
	if( dims_info[dim].filter.get() ) {
		MIIterator mit(&mind, orig_dim);
		while(mit.IsValid()) {
			if(mit[orig_dim] != NULL_VALUE_64)
				dims_info[dim].filter->Reset(mit[orig_dim]);
			++mit;
		}

		if (dims_info[dim].dim_is_filter) {
			added_tuples = dims_info[dim].filter->NoOnes();
			if( added_tuples > 0 && !count_only ) {		// anything to add?
				// NOTE: valid only for two-dimensional joining!
				int second_dim = dims_info[ 1 - dim ].dim_no;	// Assuming we have only two dims_info (0 and 1)

				if(   mind.GetFilter( orig_dim ) &&				// still virtual on one or both sides
					( mind.GetFilter( second_dim ) || mind.GetIndTable( second_dim ) ) ) {
					int max_dim = max( dims_info[0].dim_no , dims_info[1].dim_no );
					mind.Materialize( max_dim+1 );		// +1, because Materialize() needs a number of dims, not a number of max. one
				}
				if( mind.GetFilter( orig_dim ) ) {		// rare case: virtual + forgotten after join
					Filter *f = mind.GetFilter( orig_dim );
					f->Or( *(dims_info[dim].filter) );
					mind.MakeDimensionNull( second_dim );
				} else {
					FilterOnesIterator it(dims_info[dim].filter.get());
					while(it.IsValid()) {
						mind.AddNewMaterializedTuple( orig_dim, *it );
						++it;
					}
				}
			}
		} else {
			// first dimension as IndexTable
			_uint64 i;
			int no_of_old_dims = materialized_rows->dims.size();
			// checking if mind needs materialization
			int max_dim = max( dims_info[0].dim_no , dims_info[1].dim_no );
			mind.Materialize( max_dim+1 );		// +1, because Materialize() needs a number of dims, not a number of max. one
			// adding tuples
			vector<_uint64> row;
			row.assign( no_of_old_dims, 0 );
			materialized_rows->i_tables[materialized_rows->joining_ndx[dim]]->Lock();
			for(int k = 0; k < no_of_old_dims; k++ )
				materialized_rows->i_tables[k]->Lock();
			for (i= 0; i< materialized_rows->materialized_before_join; i++) {	//[0] must exist, all have same size
				_int64 cur_obj_before_join = materialized_rows->i_tables[materialized_rows->joining_ndx[dim]]->Get64(i);
				if( cur_obj_before_join == 0 ) {		// note that the object number is shifted by one
					// add NULL rows
					added_tuples++;
					if( !count_only ) {
						for(int k = 0; k < no_of_old_dims; k++ )
							row[k] = materialized_rows->i_tables[k]->Get64(i);
						mind.AddNewMaterializedTuple(materialized_rows->dims, row);
					}
				} else if( dims_info[dim].filter->Get( cur_obj_before_join - 1 ) ) {
						added_tuples++;
						if( !count_only ) {
							for(int k = 0; k < no_of_old_dims; k++ )
								row[k] = materialized_rows->i_tables[k]->Get64(i);
							mind.AddNewMaterializedTuple(materialized_rows->dims, row);
						}
				}
			}
			for(int k = 0; k < no_of_old_dims; k++ )
				materialized_rows->i_tables[k]->Unlock();
			materialized_rows->i_tables[materialized_rows->joining_ndx[dim]]->Unlock();
		}
	}
	return added_tuples;
}

OuterJoiner::MaterializedRows::MaterializedRows(int jdim1, int jdim2, MultiIndex& mind)
{
	MEASURE_FET("OuterJoiner::MaterializedRows::MaterializedRows(...)");
	materialized_before_join = mind.NoMaterialTuples();
	for(int i = 0; i < mind.NoDimensions(); i++) {
		if(mind.GetIndTable(i) != NULL) {
			dims.push_back(i);
			i_tables.push_back(new IndexTable(*(mind.GetIndTable(i))));
			if(jdim1 == i)
				joining_ndx[0] = dims.size()-1;
			else if(jdim2 == i)
				joining_ndx[1] = dims.size()-1;

		}
	}
}

OuterJoiner::MaterializedRows::~MaterializedRows()
{
	for (int i = 0; i < dims.size(); i++ )
		delete i_tables[i];
	i_tables.clear();
	dims.clear();
}
*/
