/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "core/PackOrderer.h"
#include "core/MIIterator.h"

PackOrderer::PackOrderer(VirtualColumn* vc, RSValue* r_filter, OrderType order)
{
	Init(vc, order, r_filter);
}

PackOrderer::CompIntLess PackOrderer::comp_int_less;
PackOrderer::CompIntGreater PackOrderer::comp_int_greater;

bool PackOrderer::Init(VirtualColumn* vc, OrderType order, RSValue* r_filter)
{
		
	if(otype != NotSpecified)
		return false;

	otype = order;

	if(vc->Type().IsFixed() || vc->Type().IsDateTime()) {
		mmtype = MMT_Fixed;
	} else
		mmtype = MMT_String;

	lastly_left = true;
	int d = vc->GetDim();
	assert(d != -1); //works only for vcolumns based on a single table

	MIIterator mit(vc->GetMultiIndex(), d);

	MMTU mid(0);
	while(mit.IsValid()) {
		int pack = mit.GetCurPackrow(d);
		if(!r_filter || r_filter[pack] != RS_NONE) {
			if(mmtype == MMT_Fixed) {
				_int64 min = vc->GetMinInt64(mit);
				_int64 max = vc->GetMaxInt64(mit);
				switch(otype) {
					case RangeSimilarity:
						mid.i = min == NULL_VALUE_64 ? MINUS_INF_64 : (max-min)/2;
						break;
					case MinAsc:
					case MinDesc:
						mid.i = min;
						break;
					case MaxAsc:
					case MaxDesc:
						mid.i = max;
						break;
				}
			} else
				// not implemented for strings & doubles/floats
				//keep the original order
				mid.i = mit.GetCurPackrow(d);
			packs.push_back(make_pair(mid, pack));
		}
		mit.NextPackrow();
	}

	if(mmtype == MMT_Fixed) {
		switch(otype) {
		case RangeSimilarity:
		case MinAsc:
		case MaxAsc:
			sort(packs.begin(), packs.end(), comp_int_less);
			break;
		case MinDesc:
		case MaxDesc:
			sort(packs.begin(), packs.end(), comp_int_greater);

		}
	}

	prevndx = curndx = INIT_VAL;
	return true;
}

PackOrderer& PackOrderer::operator++()
{
	if(curndx == END)
		return *this;
	switch(otype) {
		case RangeSimilarity: {
			if(lastly_left) {
				if(prevndx < packs.size() - 1) {
					lastly_left = !lastly_left;
					int tmp = curndx;
					curndx = prevndx + 1;
					prevndx = tmp;
				} else {
					if(curndx > 0)
						--curndx;
					else
						curndx = END;
				}
			} else if(prevndx > 0) {
				lastly_left = !lastly_left;
				int tmp = curndx;
				curndx = prevndx-1;
				prevndx = tmp;
			} else {
				if(curndx < packs.size() - 1)
					++curndx;
				else
					curndx = END;
			}
			break;
		}
		default:
			if(curndx != END){
				if(curndx < (int)packs.size() - 1)
					++curndx;
				else
					curndx = END;
			}
	}
	return *this;			// TODO: this line is added only to make compilation possible
}


void PackOrderer::RewindToMatch(VirtualColumn* vc, MIIterator& mit)
{
	assert(vc->GetDim() != -1);
	assert(otype == RangeSimilarity);

	if(mmtype == MMT_Fixed) {
		_int64 min = vc->GetMinInt64(mit);
		_int64 max = vc->GetMaxInt64(mit);
		_int64 mid = min == NULL_VALUE_64 ? MINUS_INF_64 : (max-min)/2;

		vector<PackPair>::iterator it = lower_bound(packs.begin(), packs.end(), make_pair(mid,0), comp_int_less);
		if(it == packs.end())
			curndx = packs.size() - 1;
		else
			curndx = distance(packs.begin(), it);
	} else
		// not implemented for strings & doubles/floats
		curndx = 0;

	if(packs.size() == 0)
		curndx = END;
	prevndx = curndx;
}

void PackOrderer::RewindToCurrent()
{
	if(curndx == INIT_VAL || curndx == END )
		curndx = 0;
	if(packs.size() == 0)
		curndx = END;
	prevndx = curndx;
}

/*!
 * Based on the given set of virtual columns (all from the same dimension) with information how they will be used (e.g. calculate MIN)
 * this function should find the best ordering of packrows for this dimension.
 *
 * Currently implemented only - MAX and  MIN separately.
 * TODO:
 * -MAX, MIN together (give min,max,second min, second max,..),
 * -Any combination of MINs,MAXs
 * -smallest range first (for joins),
 *  ...
 */
void PackOrderer::ChoosePackOrderer(PackOrderer& po, const vector<OrderingInfo>& oi)
{
	bool is_max = false, is_min = false;
	vector<OrderingInfo>::const_iterator oi_it;
	VirtualColumn* vc_max = NULL;
	VirtualColumn* vc_min = NULL;
	for(oi_it = oi.begin(); oi_it != oi.end(); oi_it++) {
		if(oi_it->op == MAX) {
			if(is_min) {
				is_min = false;
				break;
			} else {
				is_max = true;
				vc_max = oi_it->vc;
			}
		} else if(oi_it->op == MIN) {
			if(is_max) {
				is_max = false;
				break;
			} else {
				is_min = true;
				vc_min = oi_it->vc;
			}
		}
	}
	if(is_max)
		po.Init(vc_max, MaxDesc);
	else if(is_min)
		po.Init(vc_min, MinAsc);
}
