/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// This is a part of RCBase implementation concerned with the query optimization and rough set mechanisms
/////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "Query.h"
#include "RoughMultiIndex.h"
#include "vc/VirtualColumn.h"
#include "vc/SingleColumn.h"
#include "core/PackOrderer.h"

using namespace std;

double Query::EvaluateConditionNonJoinWeight(Descriptor &d)
{
	// Interpretation of weight:
	// an approximation of logarithm of answer size
	// 0 -> time is very short (constant).
	// high weight -> schedule this query to be executed later
	double eval = 0.0;
	_uint64 no_distinct, no_distinct2;
	_uint64 answer_size;
	if(d.IsTrue() || d.IsFalse())
		eval = 0;				// constant time
	else if(d.IsType_AttrValOrAttrValVal()) {		// typical condition: attr=val
		if(!d.encoded)
			return 9999; // temporarily till EncodeCondition is finished
		SingleColumn* col = static_cast<SingleColumn*>(d.attr.vc);
		answer_size = col->ApproxAnswerSize(d);
		eval = log(1 + double(answer_size));		// approximate size of the result
	} else if(d.GetJoinType() == DT_NON_JOIN && d.attr.vc && d.val1.vc ) {	// attr=attr on the same table
		_uint64 no_obj1 = d.attr.vc->NoTuples();		// changed to uint64 to prevent negative logarithm for NULL_VALUE_64
		_uint64 no_obj2 = d.val1.vc->NoTuples();
		if(d.op == O_EQ) {
			no_distinct = d.attr.vc->GetApproxDistVals(false);
			if(no_distinct == 0)
				no_distinct = 1;
			no_distinct2 = d.val1.vc->GetApproxDistVals(false);
			if(no_distinct2 == 0)
				no_distinct2 = 1;
			if(no_distinct2 > no_distinct)
				no_distinct = no_distinct2;					// find the attribute with smaller abstract classes
			eval = log(1 + double(no_obj1) / no_distinct);	// size of the smaller abstract class
		} else {
			if(d.attr.vc->GetDimensions().size() > 0)
				eval = log(1 + double(no_obj1) / 2);	// other operators filter potentially a half of objects
			else
				eval = log(1 + double(no_obj2) / 2);	// other operators filter potentially a half of objects
		}
	} else	// other types, incl. joins (to be calculated separately)
		eval = 99999;
	return eval;
}

double Query::EvaluateConditionJoinWeight(Descriptor &d, MultiIndex &mind)
{
	double eval = 0.0;
	int dim1 = d.attr.vc->GetDim();
	int dim2 = d.val1.vc->GetDim();
	if(dim1 > -1 && dim2 > -1) {
		// optimize tips for multiindex in case of TempTable
		/*
		if(desc[i].attr.vc->GetVarMap()[0].tab->TableType() == TEMP_TABLE)
			mind.SetTempTablePageSize(dim1, ((TempTable*)(desc[i].attr.vc->GetVarMap()[0].tab))->CalculatePageSize());
		if(desc[i].val1.vc->GetVarMap()[0].tab->TableType() == TEMP_TABLE)
			mind.SetTempTablePageSize(dim2, ((TempTable*)(desc[i].val1.vc->GetVarMap()[0].tab))->CalculatePageSize());
		*/
		/////////////////////////////////////////
		_uint64 no_obj1 = mind.OrigSize(dim1);		// changed to uint64 to prevent negative logarithm for NULL_VALUE_64
		_uint64 no_obj2 = mind.OrigSize(dim2);
		if(mind.GetFilter(dim1))
			no_obj1 = mind.DimSize(dim1);
		if(mind.GetFilter(dim2))
			no_obj2 = mind.DimSize(dim2);
		if(no_obj1 == 0 || no_obj2 == 0)
			return 1;

		_uint64 no_distinct1, no_distinct2;
		double r_select1 = 0.5 + 0.5 * d.attr.vc->RoughSelectivity(); 	// 1 if no rough exclusions, close to 0.5 if good rough performance
		double r_select2 = 0.5 + 0.5 * d.val1.vc->RoughSelectivity();
		double c_select1 = double(no_obj1) / mind.OrigSize(dim1);	// condition selectivity (assuming 1:n)
		double c_select2 = double(no_obj2) / mind.OrigSize(dim2);
		double bigger_table_size = max(no_obj1, no_obj2);

		if(d.op == O_EQ) {
			no_distinct1 = d.attr.vc->GetApproxDistVals(false);
			no_distinct2 = d.val1.vc->GetApproxDistVals(false);
			if(no_distinct1 >= 0.99 * mind.DimSize(dim1)) {	// potentially 1:n join
				eval = double(no_obj2) * r_select2 * c_select1
					 + double(no_obj1) / 100;
				eval = log(1 + eval);		// result will be of size of the second table filtered by the value of the first one
			}
			else if(no_distinct2 >= 0.99 * mind.DimSize(dim2)) { // potentially n:1 join
				eval = double(no_obj1) * r_select1 * c_select2
					 + double(no_obj2) / 100;
				eval = log(1 + eval);		// result will be of size of the first table filtered by the value of the second one
			}
			else {
				// approximated size: (no_of_classes) * (ave_class_size1 * ave_class_size2) = no_of_classes * (no_obj1 / no_dist1) * (no_obj2 / no_dist2)
				//                    where no_of_classes = min(no_dist1,no_dist2) , because they must have common values
				no_distinct1 = no_distinct1 * double(no_obj1) / mind.DimSize(dim1);
				no_distinct2 = no_distinct2 * double(no_obj2) / mind.DimSize(dim2);
				_int64 no_of_classes = min(no_distinct1, no_distinct2);

				eval = log(double(no_of_classes < 1 ? 1 : no_of_classes))	// instead of multiplications
					+ (log(1 + double(no_obj1)) - log(double(no_distinct1)))
					+ (log(1 + double(no_obj2)) - log(double(no_distinct2)));
			}
		} else
			eval = log(1 + double(no_obj1)) + log(1 + double(no_obj2));			// these join operators generate potentially large result

		eval += (eval - log(1 + bigger_table_size)) * 2;				// bonus for selective strength
		if(!mind.IsUsedInOutput(dim1) || !mind.IsUsedInOutput(dim2))	// bonus for not used
			eval *= 0.7;

		eval += 20;
	} else	// other types of joins
		eval = 99999;
	return eval;
}


///////////////////////////////////////////////////////////////////////

bool Query::RoughUpdateMultiIndex(RoughMultiIndex &rmind,
								  MultiIndex &mind,
								  vector<Descriptor> &desc)
{
	bool is_nonempty = true;
	// init by previous values of mind (if any nontrivial)
	for(int i = 0; i < mind.NoDimensions(); i++) {
		Filter *loc_f = mind.GetFilter(i);
		rmind.UpdateGlobalRoughFilter(i, loc_f);	// if the filter is nontrivial, then copy pack status
	}
	// one-dimensional conditions
	for(int i = 0; i < desc.size(); i++) {
		if(!desc[i].done && !desc[i].IsDelayed() && desc[i].IsInner() &&
			desc[i].GetJoinType() == DT_NON_JOIN) {
			VirtualColumn* vc = desc[i].attr.vc;
			if(vc && !vc->IsConst()) {
				int dim = vc->GetDim();
				if(dim == -1)
					continue;
				RSValue *rf = rmind.GetLocalDescFilter(dim, i); // rough filter for a descriptor
				MIIterator mit(&mind, dim);
				while(mit.IsValid()) {
					int p = mit.GetCurPackrow(dim);
					if(p >= 0 && rf[p] != RS_NONE)
						rf[p] = vc->RoughCheck(mit, desc[i]);
					mit.NextPackrow();
					if(m_conn->killed()) throw KilledRCException();
				}
				bool this_nonempty = rmind.UpdateGlobalRoughFilter(dim, i); // update the filter using local information
				is_nonempty = (is_nonempty && this_nonempty);
			}
		}
	}

	// now we may check all joins (not performed earlier, because it is better to filter all dimensions first with 1-D conditions)
/*
	for(int i = 0; i < desc.size(); i++) {
		if(is_nonempty && !desc[i].done && !desc[i].IsDelayed() && desc[i].IsInner() && desc[i].IsType_JoinSimple()) {
			int dim1 = desc[i].attr.dim;
			int dim2 = desc[i].val1.dim;
			RCTable *cur_t1 = (RCTable *)desc[i].attr.tab_p;
			RCTable *cur_t2 = (RCTable *)desc[i].val1.tab_p;
			if( desc[i].IsLeftJoin() || desc[i].IsFullOuterJoin() )
				rmind.SetP2PLeftJoin( dim1, dim2 );
			if( desc[i].IsRightJoin() || desc[i].IsFullOuterJoin() )
				rmind.SetP2PLeftJoin( dim2, dim1 );
			// Note: if any nontrivial joining condition is made, we must change all RS_ALL to RS_SOME
			for(int p2=0; p2<cur_t2->NoPacks(); p2++)
				if(rmind.GetPackStatus(dim2,p2)==RS_ALL)
					rmind.SetPackStatus(dim2,p2,RS_SOME);
			bool p2p_exists = rmind.UseP2P(dim1, dim2);
			for(int p1=0; p1<cur_t1->NoPacks(); p1++)
			{
				if(rmind.GetPackStatus(dim1,p1)==RS_ALL)
					rmind.SetPackStatus(dim1,p1,RS_SOME);
				if(p2p_exists) {
					for(int p2=0; p2<cur_t2->NoPacks(); p2++)			// resolve two-dimensional rough condition
					{
						if(rmind.GetPackStatus(dim1,p1)==RS_NONE || rmind.GetPackStatus(dim2,p2)==RS_NONE)
							rmind.SetCurrentP2PStatus(p1, p2, RS_NONE);
						else
							rmind.SetCurrentP2PStatus(p1, p2, cur_t1->RoughCheck(p1, p2, desc[i]));
					}
				}
			}
		}
	}
*/
	// Recalculate all multidimensional dependencies only if there are 1-dim descriptors which can benefit
	// from the projection
	if(DimsWith1dimFilters(desc, mind))
		RoughMakeProjections(rmind, mind, desc);

	///////////////////////////////////////////////////////////////////////////////////
	is_nonempty = true;
	for(uint i = 0; i < rmind.NoDimensions(); i++)		// update classical multiindex
	{
		Filter *f = mind.GetFilter(i);
		if(f)
		{
			for(int b=0;b<rmind.NoPacks(i);b++)
			{
				if(rmind.GetPackStatus(i,b) == RS_NONE)
					f->ResetBlock(b);
			}
			if(f->IsEmpty())
				is_nonempty = false;
		}
	}

	///////////////////////////////////////////////////////////////////////////////////
	// Displaying statistics

	if(rccontrol.isOn()) {
		int pack_full = 0, pack_some = 0, pack_all = 0, pack_local_some = 0;
		rccontrol.lock(m_conn->GetThreadID()) << "Packs/packrows after KN evaluation:" << unlock;
		for(uint dim = 0; dim < rmind.NoDimensions(); dim++)
		{
			pack_full = 0;
			pack_some = 0;
			pack_local_some = 0;
			pack_all  = rmind.NoPacks(dim);
			for(int b = 0; b < pack_all; b++)
			{
				if(rmind.GetPackStatus(dim,b) == RS_ALL)		pack_full++;
				else if(rmind.GetPackStatus(dim,b) != RS_NONE)	pack_some++;
			}
			for(int local_desc_num = 0; local_desc_num < rmind.NoConditions(dim); local_desc_num++)
			{
				int desc_num = rmind.GlobalDescNum(dim,local_desc_num);
				RSValue *loc = rmind.GetLocalDescFilter(dim,desc_num,true);
				for(int b = 0; b < pack_all; b++)
					if(loc[b] != RS_ALL && loc[b] != RS_NONE)
						pack_local_some++;
			}
			rccontrol.lock(m_conn->GetThreadID()) << "(t" << dim << ") Pckrows: " << pack_all << ", susp. " << pack_some << " (" << pack_all-(pack_full+pack_some) << " empty " << pack_full << " full). Packs opened in " << rmind.NoConditions(dim) << " cond.: " << pack_local_some << unlock;
		}
		for(uint dim1 = 0; dim1 < rmind.NoDimensions(); dim1++)
			for(uint dim2 = dim1 + 1; dim2 < rmind.NoDimensions(); dim2++)
			{
				double dens = rmind.GetP2PDensity(dim1, dim2);
				if(dens != -1)
					rccontrol.lock(m_conn->GetThreadID()) << "(t" << dim1 << "-t" << dim2 << ") Pack2Pack pairs: " << int(dens*10000)/100.0 << "% possible." << unlock;
			}
	}
	return is_nonempty;
}

void Query::RoughMakeProjections(	RoughMultiIndex &rmind,
									MultiIndex &mind,
									vector<Descriptor> &desc)
{
	for(int dim=0; dim < mind.NoDimensions(); dim++)
		RoughMakeProjections(rmind, mind, desc, dim, false);

	rmind.UpdateReducedDimension();
}

void Query::RoughMakeProjections(	RoughMultiIndex &rmind,
									MultiIndex &mind,
									vector<Descriptor> &desc,
									int to_dim,
									bool update_reduced)
{
	int total_excluded = 0;
//	for(dim1=0; dim1< mind.NoDimensions(); dim1++ ){
	vector<int> dims_reduced = rmind.GetReducedDimensions();
	vector<int>::const_iterator dim1;
	for(dim1 = dims_reduced.begin(); dim1 != dims_reduced.end(); ++dim1) {
		if(*dim1 == to_dim)
			continue;
		// find all descriptors which may potentially influence other dimensions
		vector<Descriptor> local_desc;
		for(int i = 0; i < desc.size(); i++) {
			if(desc[i].done || desc[i].IsDelayed())
				continue;
			if(desc[i].IsOuter())
				return;					// do not make any projections in case of outer joins present
			if(!desc[i].attr.vc || desc[i].attr.vc->GetDim() == -1 || !desc[i].val1.vc || desc[i].val1.vc->GetDim() == -1)
				continue; //only SingleColumns processed here
			DimensionVector dims(&mind);
			desc[i].DimensionUsed(dims);
			if(desc[i].attr.vc && dims[*dim1] && dims[to_dim] && dims.NoDimsUsed() == 2)
				local_desc.push_back(desc[i]);
		}

		// make projection to another dimension
		for(int i = 0; i < local_desc.size(); i++) {
			// find the other dimension
			Descriptor& ld = local_desc[i];
			DimensionVector dims(&mind);
			ld.DimensionUsed(dims);
			PackOrderer po;
			VirtualColumn* matched_vc;
			MIDummyIterator local_mit(&mind);
			if(*dim1 == ld.attr.vc->GetDim()) {
				po.Init(ld.attr.vc, PackOrderer::RangeSimilarity, rmind.GetRSValueTable(ld.attr.vc->GetDim()));
				matched_vc = ld.val1.vc;
			} else {
				po.Init(ld.val1.vc, PackOrderer::RangeSimilarity, rmind.GetRSValueTable(ld.val1.vc->GetDim()));
				matched_vc = ld.attr.vc;
			}
			// for each dim2 pack, check whether it may be joined with anything nonempty on dim1
			for(int p2 = 0; p2 < rmind.NoPacks(to_dim); p2++) {
				if(rmind.GetPackStatus(to_dim, p2) != RS_NONE) {
					bool pack_possible = false;
					local_mit.SetPack(to_dim, p2);

//					if(po.IsValid())
						po.RewindToMatch(matched_vc, local_mit);
//					else
//						po.RewindToCurrent();

					while(po.IsValid()) {
//						for(int p1 = 0; p1 < rmind.NoPacks(dim1); p1++)
//							if(rmind.GetPackStatus(dim1, p1) != RS_NONE) {
						if(rmind.GetPackStatus(*dim1, po.Current()) != RS_NONE) {
							local_mit.SetPack(*dim1, po.Current());	// set a dummy position, just for transferring pack number
							if(ld.attr.vc->RoughCheck(local_mit, ld) != RS_NONE) {
								pack_possible = true;
								break;
							}
						}
						++po;
					}
					if(!pack_possible) {
						rmind.SetPackStatus(to_dim, p2, RS_NONE);
						total_excluded++;
					}
				}
			}
		}
		if(update_reduced)
			rmind.UpdateReducedDimension(*dim1);
		local_desc.clear();
	}
	if(total_excluded > 0) {
		rccontrol.lock(m_conn->GetThreadID()) << "Packrows excluded by rough multidimensional projections: " << total_excluded << unlock;
		rmind.UpdateLocalRoughFilters(to_dim);

		for(dim1 = dims_reduced.begin(); dim1 != dims_reduced.end(); ++dim1) {
			Filter *f = mind.GetFilter(*dim1);
			if(f) {
				for(int b = 0; b < rmind.NoPacks(*dim1); b++) {
					if(rmind.GetPackStatus(*dim1, b) == RS_NONE)
						f->ResetBlock(b);
				}
			}
		}
		mind.UpdateNoTuples();
	}
}

bool Query::DimsWith1dimFilters(vector<Descriptor> &desc, MultiIndex& mind)
{
	DimensionVector dv1(&mind), dv2(&mind);
	bool first = true;
	for(int i = 0; i < desc.size(); i++) {
		if(!desc[i].done && desc[i].IsInner() && !desc[i].IsType_Join() && !desc[i].IsDelayed()) {
			if(first) {
				desc[i].DimensionUsed(dv1);
				first = false;
			} else {
				desc[i].DimensionUsed(dv2);
				if(!dv1.Intersects(dv2))
					return true;
			}
		}
	}
	return false;
}
