/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

////////////////////////////////////////////////////////////////////////////////////////
// This is a part of RCAttr implementation concerned with the query execution mechanisms
////////////////////////////////////////////////////////////////////////////////////////

#include "RCAttr.h"
#include "RCAttrPack.h"
#include "CQTerm.h"
#include "RCAttrTypeInfo.h"
#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "PackGuardian.h"
#include "tools.h"
#include "ValueSet.h"
#include "vc/VirtualColumn.h"
#include "vc/ConstColumn.h"
#include "vc/InSetColumn.h"
#include "common/bhassert.h"

using namespace std;

/////////////////////////////////////////////////////////////////////////////////////////////////////

void RCAttr::EvaluatePack(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("RCAttr::UpdateFilterPack(...)");
#endif
	BHASSERT(d.encoded, "Descriptor is not encoded!");
	if(d.op == O_NOT_NULL)
		EvaluatePack_NotNull(mit, dim);
	else if(d.op == O_IS_NULL)
		EvaluatePack_IsNull(mit, dim);
	else if(d.op == O_LIKE || d.op == O_NOT_LIKE) {
		if(RequiresUTFConversions(d.GetCollation()))
			EvaluatePack_Like_UTF(mit, dim, d);
		else
			EvaluatePack_Like(mit, dim, d);
	} else if(PackType() == PackS && (d.op == O_IN || d.op == O_NOT_IN)) {
		if(RequiresUTFConversions(d.GetCollation()))
			EvaluatePack_InString_UTF(mit, dim, d);
		else
			EvaluatePack_InString(mit, dim, d);
	} else if(PackType() == PackN && (d.op == O_IN || d.op == O_NOT_IN))
		EvaluatePack_InNum(mit, dim, d);
	else if(PackType() == PackS && (d.op == O_BETWEEN || d.op == O_NOT_BETWEEN)) {
		if(RequiresUTFConversions(d.GetCollation()))
			EvaluatePack_BetweenString_UTF(mit, dim, d);
		else
			EvaluatePack_BetweenString(mit, dim, d);
	} else if(PackType() == PackN && !ATI::IsRealType(TypeName()) && (d.op == O_BETWEEN || d.op == O_NOT_BETWEEN))
		EvaluatePack_BetweenInt(mit, dim, d);
	else if(PackType() == PackN &&  ATI::IsRealType(TypeName()) && (d.op == O_BETWEEN || d.op == O_NOT_BETWEEN))
		EvaluatePack_BetweenReal(mit, dim, d);
	else {
		assert(0);	// case not implemented!
	}
}

void RCAttr::EvaluatePack_IsNull(MIUpdatingIterator &mit, int dim)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.NextPackrow();
		return;
	}
	if(pack_mode[pack] != 0 && loc_no_nulls[pack] != 0) {		// nontrivial pack exists
		do {
			if(mit[dim] != NULL_VALUE_64 && !packs[pack]->IsNull(mit.GetCurInpack(dim)))
				mit.ResetCurrent();
			++mit;
		} while(mit.IsValid() && !mit.PackrowStarted());
	} else {													// pack is trivial - uniform or null only
		if(GetPackOntologicalStatus(pack) != NULLS_ONLY) {
			if(mit.NullsPossibleInPack(dim)) {
				do {
					if(mit[dim] != NULL_VALUE_64)
						mit.ResetCurrent();
					++mit;
				} while(mit.IsValid() && !mit.PackrowStarted());
			} else
				mit.ResetCurrentPack();
		}
		mit.NextPackrow();
	}
}

void RCAttr::EvaluatePack_NotNull(MIUpdatingIterator &mit, int dim)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {			// nulls only
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}
	if(pack_mode[pack] != 0 && loc_no_nulls[pack] != 0) {
		do {
			if(mit[dim] == NULL_VALUE_64  || packs[pack]->IsNull(mit.GetCurInpack(dim)))
				mit.ResetCurrent();
			++mit;
		} while(mit.IsValid() && !mit.PackrowStarted());
	} else {							// pack is trivial - uniform or null only
		if(GetPackOntologicalStatus(pack) == NULLS_ONLY)
			mit.ResetCurrentPack();
		else if(mit.NullsPossibleInPack(dim)) {
			do {
				if(mit[dim] == NULL_VALUE_64)
					mit.ResetCurrent();
				++mit;
			} while(mit.IsValid() && !mit.PackrowStarted());
		}
		mit.NextPackrow();
	}
}

void RCAttr::EvaluatePack_Like(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}
	AttrPackS *p = (AttrPackS*)packs[pack];
	RCBString pattern;
	d.val1.vc->GetValueString(pattern, mit);
	int min_len = 0; // the number of fixed characters
	for(uint i = 0; i < pattern.len; i++)
		if(pattern[i] != '%')
			min_len++;
	do {
		if(mit[dim] == NULL_VALUE_64 || p->IsNull(mit.GetCurInpack(dim)))
			mit.ResetCurrent();
		else {
			int len = p->GetSize(mit.GetCurInpack(dim));
			bool res;
			if(len < min_len)
				res = false;
			else {
				RCBString v(p->GetVal(mit.GetCurInpack(dim)), len);
				res = v.Like(pattern);
			}
			if(d.op == O_NOT_LIKE)
				res = !res;
			if(!res)
				mit.ResetCurrent();
		}
		++mit;
	} while(mit.IsValid() && !mit.PackrowStarted());
}

void RCAttr::EvaluatePack_Like_UTF(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}
	AttrPackS *p = (AttrPackS*)packs[pack];
	RCBString pattern;
	d.val1.vc->GetValueString(pattern, mit);
	int min_len = 0; // the number of fixed characters 
	for(uint i = 0; i < pattern.len; i++)
		if(pattern[i] != '%')
			min_len++;
	do {
		if(mit[dim] == NULL_VALUE_64 || p->IsNull(mit.GetCurInpack(dim)))
			mit.ResetCurrent();
		else {
			int len = p->GetSize(mit.GetCurInpack(dim));
			bool res;
			if(len < min_len)
				res = false;
			else {
				RCBString v(p->GetVal(mit.GetCurInpack(dim)), len);
				int x = d.GetCollation().collation->coll->wildcmp(d.GetCollation().collation, 
					v.val, v.val + v.len, 
					pattern.val, pattern.val + pattern.len, 
					'\\', '_', '%'); 
				res = (x == 0 ? true : false);
				//res = v.Like(pattern);
			}
			if(d.op == O_NOT_LIKE)
				res = !res;
			if(!res)
				mit.ResetCurrent();
		}
		++mit;
	} while(mit.IsValid() && !mit.PackrowStarted());
}

void RCAttr::EvaluatePack_InString(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}
	AttrPackS *p = (AttrPackS*)packs[pack];
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(dynamic_cast<MultiValColumn*>(d.val1.vc) != NULL);
	MultiValColumn *multival_column = static_cast<MultiValColumn*>(d.val1.vc);
	do {
		if(mit[dim] == NULL_VALUE_64 || p->IsNull(mit.GetCurInpack(dim)))
			mit.ResetCurrent();
		else {
			BHTribool res;
			res = multival_column->Contains(mit, RCBString(p->GetVal(mit.GetCurInpack(dim)), p->GetSize(mit.GetCurInpack(dim))));
			if(d.op == O_NOT_IN)
				res = !res;
			if(res != true)
				mit.ResetCurrent();
		}
		if(ConnectionInfoOnTLS.Get().killed())
			throw KilledRCException();
		++mit;
	} while(mit.IsValid() && !mit.PackrowStarted());
}

void RCAttr::EvaluatePack_InString_UTF(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}
	AttrPackS *p = (AttrPackS*)packs[pack];

	int v_len = CollationBufLen(d.GetCollation(), GetActualSize(mit.GetCurPackrow(dim)));
	RCBString v(NULL, v_len, true);

	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(dynamic_cast<MultiValColumn*>(d.val1.vc) != NULL);
	MultiValColumn *multival_column = static_cast<MultiValColumn*>(d.val1.vc);
	do {
		if(mit[dim] == NULL_VALUE_64 || p->IsNull(mit.GetCurInpack(dim)))
			mit.ResetCurrent();
		else {
			BHTribool res;
			RCBString vt(p->GetVal(mit.GetCurInpack(dim)), p->GetSize(mit.GetCurInpack(dim)));
			//ConvertToBinaryForm(vt, v, v_len, charset, false);
			res = multival_column->Contains(mit, vt);
			if(d.op == O_NOT_IN)
				res = !res;
			if(res != true)
				mit.ResetCurrent();
		}
		if(ConnectionInfoOnTLS.Get().killed())
			throw KilledRCException();
		++mit;
	} while(mit.IsValid() && !mit.PackrowStarted());
}

void RCAttr::EvaluatePack_InNum(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}

	AttrPackN* p = (AttrPackN*)packs[pack];
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(dynamic_cast<MultiValColumn*>(d.val1.vc) != NULL);
	MultiValColumn *multival_column = static_cast<MultiValColumn*>(d.val1.vc);
	bool lookup_to_num = ATI::IsStringType(TypeName());
	//RCValueObject val;
	BHTribool res;
	boost::scoped_ptr<RCDataType> value(ValuePrototype(lookup_to_num).Clone());
	if(local_min[pack] == local_max[pack]) {
		if(GetPackOntologicalStatus(pack) == NULLS_ONLY) {
			mit.ResetCurrentPack();
			mit.NextPackrow();
		} else {
			// only nulls and single value
			// pack does not exist
			do {
				if(IsNull(mit[dim]))
					mit.ResetCurrent();
				else {
					// find the first non-null and set the rest basing on it.
					//const RCValueObject& val = GetValue(mit[dim], lookup_to_num);
					// note: res may be UNKNOWN for NOT IN (...null...)
					res = multival_column->Contains(mit, GetValue(mit[dim], *value, lookup_to_num));
					if(d.op == O_NOT_IN)
						res = !res;
					if(res == true) {
						if(loc_no_nulls[pack] != 0)
							EvaluatePack_NotNull(mit, dim);
						else
							mit.NextPackrow();
					} else {
						mit.ResetCurrentPack();
						mit.NextPackrow();
					}
					break;
				}
				++mit;
			} while(mit.IsValid() && !mit.PackrowStarted());
		}
	} else {
		do {
			if(mit[dim] == NULL_VALUE_64 || p->IsNull(mit.GetCurInpack(dim)))
				mit.ResetCurrent();
			else {
				;
				//const RCValueObject& val = GetValue(mit[dim], lookup_to_num);
				// note: res may be UNKNOWN for NOT IN (...null...)
				res = multival_column->Contains(mit, GetValue(mit[dim], *value, lookup_to_num));
				if(d.op == O_NOT_IN)
					res = !res;
				if(res != true)
					mit.ResetCurrent();
			}
			if(ConnectionInfoOnTLS.Get().killed())
				throw KilledRCException();
			++mit;
		} while(mit.IsValid() && !mit.PackrowStarted());
	}
}

void RCAttr::EvaluatePack_BetweenString(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}

	AttrPackS* p = (AttrPackS*)packs[pack];
	RCBString v1, v2;
	d.val1.vc->GetValueString(v1, mit);
	d.val2.vc->GetValueString(v2, mit);
	do {
		int inpack = mit.GetCurInpack(dim);		// row number inside the pack
		if(mit[dim] == NULL_VALUE_64 || p->IsNull(inpack))
			mit.ResetCurrent();
		else {
			bool res;
			RCBString v(p->GetVal(inpack), p->GetSize(inpack));
			res = ( d.sharp && ((v1.IsNull() || v > v1)  && (v2.IsNull() || v < v2) )) ||
				  (!d.sharp && ((v1.IsNull() || v >= v1) && (v2.IsNull() || v <= v2))) ;
			if(d.op == O_NOT_BETWEEN)
				res = !res;
			if(!res)
				mit.ResetCurrent();
		}
		++mit;
	} while(mit.IsValid() && !mit.PackrowStarted());
}

void RCAttr::EvaluatePack_BetweenString_UTF(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}

	AttrPackS* p = (AttrPackS*)packs[pack];
	RCBString v1, v2;
	DTCollation coll = d.GetCollation();
	d.val1.vc->GetValueString(v1, mit);
	d.val2.vc->GetValueString(v2, mit);
	do {
		int inpack = mit.GetCurInpack(dim);		// row number inside the pack
		if(mit[dim] == NULL_VALUE_64 || p->IsNull(inpack))
			mit.ResetCurrent();
		else {
			RCBString v(p->GetVal(inpack), p->GetSize(inpack));
			bool res = ( d.sharp && ((v1.IsNull() || CollationStrCmp(coll, v, v1) > 0) && (v2.IsNull() || CollationStrCmp(coll, v, v2) < 0) )) ||
				  (!d.sharp && ((v1.IsNull() || CollationStrCmp(coll, v, v1) >= 0) && (v2.IsNull() || CollationStrCmp(coll, v, v2) <= 0)));
			if(d.op == O_NOT_BETWEEN)
				res = !res;
			if(!res)
				mit.ResetCurrent();
		}
		++mit;
	} while(mit.IsValid() && !mit.PackrowStarted());
}

void RCAttr::EvaluatePack_BetweenInt(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}

	AttrPackN *p = (AttrPackN*)packs[pack];
	_int64 pv1 = d.val1.vc->GetValueInt64(mit);
	_int64 pv2 = d.val2.vc->GetValueInt64(mit);
	if(pv1 != MINUS_INF_64)
		pv1 = pv1 - local_min[pack];
	else
		pv1 = 0;

	if(pv2 != PLUS_INF_64) // encode from 0-level to 2-level
		pv2 = pv2 - local_min[pack];
	else
		pv2 = local_max[pack] - local_min[pack];
	if(local_min[pack] != local_max[pack]) {
		bool res;
		_int64 v;
		do {
			if(mit[dim] == NULL_VALUE_64 || p->IsNull(mit.GetCurInpack(dim)))
				mit.ResetCurrent();
			else {
				v = p->GetVal64(mit.GetCurInpack(dim));
				res = (pv1 <= v && v <= pv2);
				if(d.op == O_NOT_BETWEEN)
					res = !res;
				if(!res)
					mit.ResetCurrent();
			}
			++mit;
		} while(mit.IsValid() && !mit.PackrowStarted());
	} else {
		// local_min==local_max, and in 2-level encoding both are 0
		if(	((pv1 > 0 || pv2 < 0)	&& d.op == O_BETWEEN) ||
			( pv1 <= 0 && pv2 >= 0	&& d.op == O_NOT_BETWEEN) ) {
			mit.ResetCurrentPack();
			mit.NextPackrow();
		}
		else
			EvaluatePack_NotNull(mit, dim);
	}
}

void RCAttr::EvaluatePack_BetweenReal(MIUpdatingIterator &mit, int dim, Descriptor& d)
{
	int pack = mit.GetCurPackrow(dim);
	if(pack == -1) {
		mit.ResetCurrentPack();
		mit.NextPackrow();
		return;
	}

	AttrPackN *p = (AttrPackN*)packs[pack];
	_int64 pv1 = d.val1.vc->GetValueInt64(mit);
	_int64 pv2 = d.val2.vc->GetValueInt64(mit);
	double dv1 = *((double*)&pv1);
	double dv2 = *((double*)&pv2);
	if(local_min[pack] != local_max[pack]) {
		bool res;
		double v;
		do {
			if(mit[dim] == NULL_VALUE_64 || p->IsNull(mit.GetCurInpack(dim)))
				mit.ResetCurrent();
			else {
				v = p->GetValD(mit.GetCurInpack(dim));
				res = (dv1 <= v && v <= dv2);
				if(d.op == O_NOT_BETWEEN)
					res = !res;
				if(!res)
					mit.ResetCurrent();
			}
			++mit;
		} while(mit.IsValid() && !mit.PackrowStarted());
	} else {
		// local_min==local_max, for float
		double uni_val = *((double*)(local_min + pack));
		if(	((dv1 > uni_val  || dv2 < uni_val) && d.op == O_BETWEEN) ||
			( dv1 <= uni_val && dv2 >= uni_val && d.op == O_NOT_BETWEEN) ) {
			mit.ResetCurrentPack();
			mit.NextPackrow();
		} else
			EvaluatePack_NotNull(mit, dim);
	}
}

/*void RCAttr::UpdateFilterDelayed(_uint64 obj, Filter& f, Descriptor& desc)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(obj >= 0 && obj < f.NoObj());
	if(CheckCondition(obj, desc) == true)
		f.SetDelayed((_int64)obj);
	else
		f.ResetDelayed((_int64)obj);
}*/

/*void RCAttr::UpdateFilter(_uint64 obj, Filter& f, Descriptor& desc)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(obj >= 0 && obj < f.NoObj());
	if(CheckCondition(obj, desc) == true)
		f.Set((_int64)obj);
	else
		f.Reset((_int64)obj);
}*/

/*bool RCAttr::CheckCondition(_uint64 obj, Descriptor& d_in)			// check whether the given object meet the given condition
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("RCAttr::CheckConditio(...)");
#endif
	Descriptor d;
	EncodeCondition(d_in, d, (obj == NULL_VALUE_64) );	// replace the old descriptor by the encoded one
														// the third argument: additional nulls possible?
	LoadPackInfo();						// descriptor is already encoded (normalized)!

	if(d.op == O_FALSE)		return false;
	if(d.op == O_TRUE)		return true;
	if( obj == NULL_VALUE_64 ) {
		if(d.op == O_IS_NULL)
			return true;
		return false;
	}

	int pack=(int)(obj >> 16);
	RSValue packres = RoughCheck(pack,d);

	if(packres==RS_ALL)		return true;
	if(packres==RS_NONE)	return false;

	LockPackForUse(pack, NULL);		// TODO: should use ConnectionInfo for statistics.
	bool res=false;

	try
	{
		if(d.op == O_IS_NULL)
			res = (IsNull(obj)==1);
		else if(d.op == O_NOT_NULL)
			res = (IsNull(obj)==0);
		else if(IsNull(obj))
			res = false;
		else if(d.op == O_LIKE || d.op == O_NOT_LIKE)
		{
			RCBString pattern((char*)(d.val1.val_t));
			res = (GetValueString(obj,true)).Like(pattern);
			if(d.op == O_NOT_LIKE)
				res = !res;
		}
		else if(d.op == O_IN || d.op == O_NOT_IN)
		{
			if(PackType() == PackN && ATI::IsRealType(TypeName())) {
				_int64 v64 = GetValueInt64(obj);
				RCNum objNum( *(double*)(&v64));
				res = d.val1.val_set->Contains(objNum);
			} else if(PackType() == PackN && ATI::IsDateTimeType(TypeName())) {
				_int64 v64 = GetValueInt64(obj);
				RCDateTime x(v64,TypeName());
				res = d.val1.val_set->Contains(x);
			} else if(PackType() == PackN)  {
				RCNum objRCN = 	RCNum( GetValueInt64(obj),Type().GetScale() );
				res = d.val1.val_set->Contains(objRCN);
			}
			else if(PackType() == PackS) {
				RCBString objRCBS  = GetValueString(obj,true);
				res = d.val1.val_set->Contains(objRCBS);
			}
			if(d.op == O_NOT_IN)
				res = !res;
		}
		/////////////////////////// String operations  /////////////////
		else if(PackType() == PackS)
		{
			// note that already converted into BETWEEN or NOT_EQ
			RCBString v = GetValueString(obj,true);
			if(d.op == O_NOT_EQ)
				res = (strcmp(RCBString(d.val1.val_t), v) != 0);
			else if(strcmp(v, d.val1.val_t) == 0 || strcmp(v, d.val2.val_t) == 0)
				res = !d.sharp;
			else if((d.val1.IsNull() || strcmp(v, d.val1.val_t) > 0) &&	(d.val2.IsNull() || strcmp(v,d.val2.val_t) < 0))
				res = true;
			else
				res = false;
			if(d.op == O_NOT_BETWEEN)
				res = !res;
		}
		/////////////////////////// Numerical operations  //////////////
		else if(ATI::IsRealType(TypeName()))
		{
			//note that already converted into BETWEEN or NOT_BETWEEN
			BHASSERT(d.sharp==false, "d.sharp should be false at this point!");
			_int64 v64 = GetValueInt64(obj);
			double v=*(double*)(&v64);
			double v1=*(double*)(&d.val1.val_n);
			double v2=*(double*)(&d.val2.val_n);
			BHASSERT(d.op != O_NOT_EQ, "d.op should not be O_NOT_EQ at this point!");
			if( (v1==MINUS_INF_DBL || v>=v1) &&
				(v2==PLUS_INF_DBL  || v<=v2))	res = true;
			else res = false;

			if(d.op == O_NOT_BETWEEN)
				res = !res;
		}
		else
		{
			// note that already converted into BETWEEN or NOT_EQ
			BHASSERT(d.sharp==false, "d.sharp should be false at this point!");
			_int64 v=GetValueInt64(obj);
			BHASSERT(d.op != O_NOT_EQ, "d.op should not be O_NOT_EQ at this point!");
			if(v>=d.val1.val_n && v<=d.val2.val_n)
				res = true;
			else
				res = false;
			if(d.op == O_NOT_BETWEEN)
				res = !res;
		}
	}
	catch(...)
	{
		UnlockPackFromUse(pack);
		throw;
	}
	UnlockPackFromUse(pack);
	return res;
}
*/
/*
void RCAttr::CheckPackCondition(int pack, Filter &f, Descriptor& d, ConnectionInfo *conn)	// reset filter on all positions where the given condition is not met
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("RCAttr::CheckPackCondition(...)");
#endif
	BHASSERT(d.encoded, "Descriptor is not encoded!");
	if(d.op == O_FALSE || f.IsEmpty(pack)) {
		f.ResetBlock(pack);
		return;
	}
	if(d.op == O_TRUE)
		return;
	LoadPackInfo();
	LockPackForUse(pack, conn);
	///////////////////////////////// IS NULL / IS NOT NULL //////////////////////////
	if(d.op == O_IS_NULL || d.op == O_NOT_NULL) {
		if(d.op == O_NOT_NULL) {
			if(pack_mode[pack] != 0)
				packs[pack]->AndNotNull(f, pack);
			else if(GetPackOntologicalStatus(pack) == NULLS_ONLY)
				f.ResetBlock(pack);
		} else {
			if(pack_mode[pack] != 0)
				packs[pack]->AndIsNull(f, pack);
			else if(GetPackOntologicalStatus(pack) == UNIFORM)
				f.ResetBlock(pack);
		}
	}
	///////////////////////////////// LIKE //////////////////////////
	else if(d.op == O_LIKE || d.op == O_NOT_LIKE) {
		RCBString pattern((char*) (d.val1.val_t));
		int min_len = 0; // the number of fixed characters
		for(uint i = 0; i < pattern.len; i++)
			if(pattern[i] != '%')
				min_len++;
		((AttrPackS*) packs[pack])->CheckLike(f, pack, pattern, min_len, d.lop, (d.op == O_NOT_LIKE));
	}
	///////////////////////////////// IN  //////////////////////////
	else if(d.op == O_IN || d.op == O_NOT_IN) {
		try {
			if(PackType() == PackS) {
				((AttrPackS*) packs[pack])->CheckIn(f, pack, d.val1.val_set, d.lop, (d.op == O_NOT_IN));
			} else {
				bool lookup_to_num = ATI::IsStringType(TypeName());
				if(local_min[pack] == local_max[pack]) {
					RCValueObject objRCVO = GetMinValue(pack);
					bool all_in = d.val1.val_set->Contains(objRCVO);
					if(((all_in && d.op == O_IN) || (!all_in && d.op == O_NOT_IN))) {
						if(loc_no_nulls[pack] != 0)
							((AttrPackN*) packs[pack])->AndNotNull(f, pack);
					} else
						f.ResetBlock(pack);
				} else {
					_int64 obj = 0;
					int pack_loc = pack;
					f.Rewind(pack);
					while(1) {
						obj = f.GetNext1();
						pack_loc = int(obj >> 16);
						if(obj < 0 || pack_loc != pack) {
							f.CommitPack(pack);
							break; // end of pack
						}
						RCValueObject objRCVO = GetValue(obj, lookup_to_num);
						if( IsNull(obj) ||
						    (d.val1.val_set->Contains(objRCVO) == (d.op == O_NOT_IN)) ||
						    (d.op == O_NOT_IN && d.val1.val_set->ContainsNulls()) ) // (match and NOT IN) or (not match and IN)
							f.ResetCurrentDelayed();
					}
				}
			}
		} catch (...) {
			UnlockPackFromUse(pack);
			throw ;
		}
	}
	/////////////////////////// String operations  /////////////////
	else if(PackType() == PackS) {
		// note that already converted into BETWEEN or NOT_BETWEEN
		((AttrPackS*)packs[pack])->CheckBetween(f,pack, d.val1, d.val2, d.lop, d.sharp, &dic, (d.op == O_NOT_BETWEEN));
	}
	/////////////////////////// Numerical operations  //////////////
	else {
		BHASSERT(PackType()==PackN, "Pack type must be numeric!");
		_int64 pv1 = 0, pv2 = 0; // 2-level local values
		if(!ATI::IsRealType(TypeName())) { // && type != RC_VARCHAR && type != RC_STRING)
			if(d.val1 != MINUS_INF_64)
				pv1 = d.val1 - local_min[pack];
			else
				pv1 = 0;

			if(d.val2 != PLUS_INF_64) // encode from 0-level to 2-level
				pv2 = d.val2 - local_min[pack];
			else
				pv2 = local_max[pack] - local_min[pack];
		} else {
			pv1=d.val1;
			pv2=d.val2;
		}
		if(local_min[pack]!=local_max[pack])
			((AttrPackN*)packs[pack])->CheckBetween(f,pack,pv1,pv2, TypeName(), d.lop, false, (d.op == O_NOT_BETWEEN)); // never sharp, because modified by double epsilon
		else {
			if(!ATI::IsRealType(TypeName())) {
				// local_min==local_max, and in 2-level encoding both are 0
				if(d.op == O_BETWEEN) {
					if( pv1> 0 || pv2 < 0 )
						f.ResetBlock(pack);
					else if(loc_no_nulls[pack] != 0)
						packs[pack]->AndNotNull(f, pack);
				} else {
					assert(d.op == O_NOT_BETWEEN);
					if( pv1 <= 0 && pv2 >= 0 )
						f.ResetBlock(pack);
					else if(loc_no_nulls[pack] != 0)
						packs[pack]->AndNotNull(f, pack);
				}
			} else {
				// local_min==local_max, for float
				double uni_val = *((double*)(local_min + pack));
				if(d.op == O_BETWEEN) {
					if( *((double*)&pv1)> uni_val || *((double*)&pv2) < uni_val )
						f.ResetBlock(pack);
					else if(loc_no_nulls[pack] != 0)
						packs[pack]->AndNotNull(f, pack);
				} else {
					assert(d.op == O_NOT_BETWEEN);
					if( *((double*)&pv1) <= uni_val && *((double*)&pv2) >= uni_val )
						f.ResetBlock(pack);
					else if(loc_no_nulls[pack] != 0)
						packs[pack]->AndNotNull(f, pack);
				}
			}
		}
	}
	////////////////////////////// End of operation list ///////////////////////////////
	UnlockPackFromUse(pack);
}
*/

////////////////////////////////////////////////////////////////////////////////////////////////////
/*
_int64 RCAttr::Sum(Filter *f, _int64* quant, int q_pack_start, int q_pack_stop, _int64 quant_factor, _int64 &counter)
{
	#ifdef FUNCTIONS_EXECUTION_TIMES
		FETOperator feto("RCAttr::Sum(...)");
	#endif
	if(f->IsEmpty())
	{
		return 0;
	}
	if(TypeName() == RC_STRING || TypeName() == RC_VARCHAR || ATI::IsDateTimeType(TypeName()) || TypeName() == RC_BYTE || TypeName() == RC_VARBYTE)
	{
		throw NotImplementedRCException("Operator Sum is not implemented for this data type.");
	}

	_int64 sum = 0;
	counter = 0;
	double dsum = 0;
	_int64 quant_offset = (static_cast<_uint64>(q_pack_start) << 16);			// quant is defined on a subset only!
	if(NoObj() == 0 || NoObj() == NoNulls())
		sum = NULL_VALUE_64;
	else
	{
		BHASSERT(f->NoBlocks() == NoPack(),"Error: internal (RCAttr::Sum).");
		LoadPackInfo();
        for(int b = q_pack_start; b <= q_pack_stop; b++)
		{
			bool use_stats = false;
			PackOntologicalStatus current_pack_status = GetPackOntologicalStatus( b );
			if(f->IsEmpty(b) || current_pack_status == NULLS_ONLY || pack_file[b] == PF_NO_OBJ)
			{
//					sum += 0;
				use_stats = true;
			}
			else if(f->IsFull(b))
			{
				use_stats = true;
				_int64 qu_value = 1;
				if(quant)
				{
					_int64 stop_obj = min(65536*_int64(b+1), f->NoObj());
					qu_value = quant[ 65536 * _int64(b) - quant_offset ];
					for(_int64 o = 65536*_int64(b) + 1; o < stop_obj; o++)
						// check the uniformity of the pack
						if(qu_value != quant[ o - quant_offset ])
						{
							use_stats = false;
							break;
						}
				}
				////// check possible overflows: if values are larger than 2^63 / 65536, then do not trust DPNs
				if( use_stats && !ATI::IsRealType( TypeName() ) &&
					( current_pack_status != UNIFORM && (local_max[b]>140000000000000LL || local_min[b]<-140000000000000LL))
				)
					use_stats = false;

				////// use DPNs instead of actual data packs
				if(use_stats)
				{
					if(!ATI::IsRealType(TypeName())) {
						sum += (GetSum(b) * qu_value);
						dsum += (double(GetSum(b)) * qu_value);
					}
					else
						dsum += GetSumD(b);
					counter += (_int64(loc_no_obj[b]) + 1 - loc_no_nulls[b]) * qu_value;
				}
			}
			else if( current_pack_status == UNIFORM)
			{
				use_stats = true;
				_int64 qu_value = -1;
				if(quant)
				{
					_int64 stop_obj = min(65536*_int64(b+1), f->NoObj());
					for(_int64 o = 65536*_int64(b); o < stop_obj; o++)		// check the uniformity of the pack
					{
						if(f->Get(o))
						{
							if(qu_value == -1)
							{
								qu_value = quant[ o - quant_offset ];
							}
							else if(qu_value != quant[ o - quant_offset ])
							{
								use_stats = false;
								break;
							}
						}
					}
					if(qu_value == -1)			// something wrong: f set to 1, but quant[o]=0, or f empty
						qu_value = 0;
				}
				else
					qu_value = 1;

				if(use_stats)
				{
					if(!ATI::IsRealType(TypeName()))
					{
						_int64 titmp = local_min[b];
						titmp *= int(f->NoOnes(b));
						sum += titmp * qu_value;
						dsum += double(titmp) * qu_value;
					}
					else
					{
						_int64 titmp = local_min[b];
						double d = *(double*)&titmp;
						dsum += (d * int(f->NoOnes(b)) * qu_value);
					}
					counter += int(f->NoOnes(b)) * qu_value;
				}
			}
			if( !use_stats )
			{
				LockPackForUse(b, NULL);			// TODO: should use ConnectionInfo for statistics.
				if(pack_mode[b] != 2)
					LoadPack(b);
				if( !ATI::IsRealType(TypeName()) )
				{
					_int64 titmp;
					f->Rewind(b);
					int block = 0;
					ushort n = 0;
					f->GetNext1(block, n);
					while(b == block)
					{
						if( current_pack_status == UNIFORM || packs[b]->IsNull(n)==0 )
						{
							if(quant)
							{
								_uint64 obj = (static_cast<_uint64>(block)<<16) + n;
								if( current_pack_status == NORMAL )
									titmp = GetTable64Loaded( b, n );
								else
									titmp = local_min[b];
								sum += titmp * quant[ obj - quant_offset ];
								dsum += double(titmp) * quant[ obj - quant_offset ];
								counter += quant[ obj - quant_offset ];
							}
							else
							{
								if( current_pack_status == NORMAL )
									titmp = GetTable64Loaded( b, n );
								else
									titmp = local_min[b];
								sum += titmp;
								dsum += double(titmp);
								counter++;
							}
						}
						f->GetNext1(block, n);
					}
				}
				else
				{
					f->Rewind(b);
					int block = 0;
					ushort n = 0;
					_int64 d;
					f->GetNext1(block, n);
					while(b == block)
					{
						if( current_pack_status == UNIFORM || packs[b]->IsNull(n)==0 )
						{
							if(quant)
							{
								_uint64 obj = (static_cast<_uint64>(block)<<16) + n;
								if( current_pack_status == NORMAL )
									d = GetTable64Loaded( b, n );
								else
									d = local_min[b];
								dsum += *(double*)(&d) * quant[ obj - quant_offset ];
								counter += quant[ obj - quant_offset ];
							}
							else
							{
								if( current_pack_status == NORMAL )
									d = GetTable64Loaded( b, n );
								else
									d = local_min[b];
								dsum += *(double*)(&d);
								counter++;
							}
						}
						f->GetNext1(block, n);
					}
				}
				UnlockPackFromUse(b);
			}
		}
		// end for
	}
	dsum *= quant_factor;
	sum  *= quant_factor;
	counter *= quant_factor;
	if(ATI::IsRealType(TypeName()))
		sum = *(_int64*)(&dsum);
	else if(dsum > double(PLUS_INF_64) || dsum < double(MINUS_INF_64))	// int64 overflow
	{
		dsum /= PowOfTen(Type().GetScale());	// dsum was in this case calculated basing on integer representation of decimals
		throw NotImplementedRCException("Overflow while summing numbers");
		throw DataTypeConversionRCException("Overflow: data typed changed", *(_int64*)(&dsum), RC_REAL);
	}
	if(counter==0)
		sum = NULL_VALUE_64;
	return sum;
}

_int64 RCAttr::CountNotNull(Filter *f, _int64* quant, int q_pack_start, int q_pack_stop)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("RCAttr::CountNotNull(...)");
#endif
	if(f==NULL)
	{
		LoadPackInfo();
		return NoObj()-NoNulls();
	}
	_int64 count = 0;
	if(f->IsEmpty() || NoObj() == 0)
	{
		return 0;
	}
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(f->NoBlocks() == NoPack());
	LoadPackInfo();
	int bb;
	ushort nn;
	_int64 quant_offset = (static_cast<_uint64>(q_pack_start) << 16);			// quant is defined on a subset only!
	for(int b = q_pack_start; b <= q_pack_stop; b++)
	{
		if(f->IsEmpty(b) || pack_file[b] == PF_NULLS_ONLY || pack_file[b] == PF_NO_OBJ)
		{
			//					count += 0;
		}
		else if(NoNulls()==0 || loc_no_nulls[b]==0)
		{
			if(quant)
			{
				f->Rewind(b);
				do
				{
					f->GetNext1(bb,nn);
					if(bb!=b) break;
					count += quant[ (_int64(b)<<16)+nn - quant_offset];
				} while(bb==b);
			}
			else
				count += f->NoOnes(b);

		}
		else
		{
			f->Rewind(b);
			LockPackForUse(b, NULL);		// TODO: should use ConnectionInfo for statistics.
			do
			{
				f->GetNext1(bb,nn);
				if(bb!=b) break;
				if(IsNull((_int64(b)<<16)+nn)==false)
				{
					if(quant)
						count += quant[ (_int64(b)<<16)+nn-quant_offset ];
					else
						count++;
				}
			} while(bb==b);
			UnlockPackFromUse(b);
		}
	}// end for
	return count;
}
*/
bool RCAttr::IsDistinct(Filter* f)
{
	if(PhysicalColumn::IsDistinct() == RS_ALL) {		// = is_unique_updated && is_unique
		if(f == NULL)
			return (NoNulls() == 0);		// no nulls at all, and is_unique  => distinct
		LoadPackInfo();
		for(int b = 0; b < NoPack(); b++)
			if(!f->IsEmpty(b) && loc_no_nulls[b] > 0)		// any null in nonempty pack?
				return false;
		return true;
	}
	return false;
}


_uint64 RCAttr::ApproxAnswerSize(Descriptor& d)
{
	BHASSERT(d.encoded, "The descriptor is not encoded!");
	static MIIterator const mit(NULL);
	LoadPackInfo();
	if(d.op == O_NOT_NULL)
		return NoObj() - NoNulls();
	if(d.op == O_IS_NULL)
		return NoNulls();
	if(d.val1.vc && !d.val1.vc->IsConst()) {
		_uint64 no_distinct = ApproxDistinctVals(false);
		if(no_distinct == 0)
			no_distinct = 1;
		if(d.op == O_EQ)
			return	NoObj() / no_distinct;
		if(d.op == O_NOT_EQ)
			return	NoObj() - (NoObj() / no_distinct);
		return (NoObj() - NoNulls()) / 2;		// default
	}
	if(d.op == O_BETWEEN && d.val1.vc->IsConst() && d.val2.vc->IsConst() && PackType()==PackN) {
		double res = 0;
		_int64 val1 = d.val1.vc->GetValueInt64(mit);
		_int64 val2 = d.val2.vc->GetValueInt64(mit);
		if(!ATI::IsRealType(TypeName())) {
			_int64 span1, span2;				// numerical case: approximate number of rows in each pack
			for(int b = 0; b < NoPack(); b++) {
				if(local_min[b] > val2 || local_max[b] < val1 || loc_no_nulls[b] == loc_no_obj[b])
					continue;	// pack irrelevant
				if((local_min[b] < -4611686018427387900LL && local_max[b] > 4611686018427387900LL ))
					span1 = 1;
				else
					span1 = local_max[b] - local_min[b] + 1;
				if(val2 < local_max[b])			// calculate the size of intersection
					span2 = val2;
				else
					span2 = local_max[b];
				if(val1 > local_min[b])
					span2 -= val1;
				else
					span2 -= local_min[b];
				span2 += 1;
				BHASSERT(span1 > 0, "A variable 'span1' should be greater than 0!");
				res += (loc_no_obj[b] - loc_no_nulls[b]) * double(span2) / span1;		// supposing uniform distribution of values
			}
		} else {		// double
			double span1, span2;				// numerical case: approximate number of rows in each pack
			double v_min = *(double*)&val1;
			double v_max = *(double*)&val2;
			for(int b = 0; b < NoPack(); b++) {
				double d_min = *((double*)(local_min + b));
				double d_max = *((double*)(local_max + b));
				if(d_min > v_max || d_max < v_min || loc_no_nulls[b] == loc_no_obj[b])
					continue;	// pack irrelevant
				span1 = d_max - d_min;
				span2 = min(v_max, d_max) - max(v_min, d_min);
				if(span1 == 0)
					res += loc_no_obj[b] - loc_no_nulls[b];
				else if(span2 == 0)					// just one value
					res += 1;
				else
					res += (loc_no_obj[b] - loc_no_nulls[b]) * (span2 / span1);		// supposing uniform distribution of values
			}
		}
		return _int64(res);
	}
	return (NoObj() - NoNulls()) / 2;		// default
}

ushort RCAttr::MaxStringSize(Filter *f)		// maximal byte string length in column
{
	LoadPackInfo();
	ushort max_size = 1;
	if(Type().IsLookup()) {
		_int64 cur_min = PLUS_INF_64;
		_int64 cur_max = MINUS_INF_64;
		for(int b = 0; b < NoPack(); b++)
		{
			if((f && f->IsEmpty(b)) || GetPackOntologicalStatus(b) == NULLS_ONLY)
				continue;
			if(local_min[b] < cur_min)
				cur_min = local_min[b];
			if(local_max[b] > cur_max)
				cur_max = local_max[b];
		}
		if(cur_min != PLUS_INF_64)
			max_size = dic.MaxValueSize(cur_min, cur_max);
	} else {
		for(int b = 0; b < NoPack(); b++)
		{
			if(f && f->IsEmpty(b))
				continue;
			ushort cur_size = GetActualSize(b);
			if(max_size < cur_size)
				max_size = cur_size;
			if(max_size == Type().GetPrecision())
				break;
		}
	}
	return max_size;
}

///////////////////////////////////////////////////////////////////////////////////

bool RCAttr::TryToMerge(Descriptor &d1, Descriptor &d2)
{
	if((d1.op != O_BETWEEN && d1.op != O_NOT_BETWEEN) || (d2.op != O_BETWEEN && d2.op != O_NOT_BETWEEN))
		return false;
	if(PackType() == PackN &&
		d1.val1.vc && d1.val2.vc && d2.val1.vc && d2.val2.vc &&
		d1.val1.vc->IsConst() && d1.val2.vc->IsConst() && d2.val1.vc->IsConst() && d2.val2.vc->IsConst())
	{
		static MIIterator const mit(NULL);
		_int64 d1min = d1.val1.vc->GetValueInt64(mit);
		_int64 d1max = d1.val2.vc->GetValueInt64(mit);
		_int64 d2min = d2.val1.vc->GetValueInt64(mit);
		_int64 d2max = d2.val2.vc->GetValueInt64(mit);
		if(!ATI::IsRealType(TypeName())) {
			if(d1.op == O_BETWEEN && d2.op == O_BETWEEN)
			{
				if(d2min > d1min)
					swap(d1.val1, d2.val1);
				if(d2max < d1max)
					swap(d1.val2, d2.val2);
				if(d1min > d1max)
					d1.op = O_FALSE;	// disjoint?
				return true;
			}
			if(d1.op == O_NOT_BETWEEN && d2.op == O_NOT_BETWEEN)
			{
				if(d1min < d2max && d2min < d1max)
				{
					if(d2min < d1min)
						swap(d1.val1, d2.val1);
					if(d2max > d1max)
						swap(d1.val2, d2.val2);
					return true;
				}
			}
		} else {		// double
			double dv1min = *((double*)&d1min);
			double dv1max = *((double*)&d1max);
			double dv2min = *((double*)&d2min);
			double dv2max = *((double*)&d2max);
			if(d1.op == O_BETWEEN && d2.op == O_BETWEEN)
			{
				if(dv2min > dv1min)
					swap(d1.val1, d2.val1);
				if(dv2max < dv1max)
					swap(d1.val2, d2.val2);
				if(dv1min > dv1max)
					d1.op = O_FALSE;	// disjoint?
				return true;
			}
			if(d1.op == O_NOT_BETWEEN && d2.op == O_NOT_BETWEEN)
			{
				if(dv1min < dv2max && dv2min < dv1max)
				{
					if(dv2min < dv1min)
						swap(d1.val1, d2.val1);
					if(dv2max > dv1max)
						swap(d1.val2, d2.val2);
					return true;
				}
			}
		}
	}
	return false;
}
