/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "RCAttr.h"
#include "RCAttrPack.h"
#include "bintools.h"
#include "compress/BitstreamCompressor.h"
#include "compress/TextCompressor.h"
#include "compress/PartDict.h"
#include "compress/NumCompressor.h"
#include "system/TextUtils.h"
#include "system/IBFile.h"
#include "WinTools.h"
#include "tools.h"
#include "ValueSet.h"

AttrPackS::AttrPackS(RCAttr& attr, int pack_no)
	:	AttrPack(attr, pack_no)
{
	data = 0;
	index = 0;
	lens = 0;
	Construct();
}

void AttrPackS::Prepare(int no_nulls)
{
	ConstructionLock();

	Construct();
	this->no_obj = no_nulls;
	for(int i = 0; i < no_nulls; i++)
		SetNull(i);						// no_nulls set inside
	is_empty = false;
	ConstructionUnlock();
}

void AttrPackS::LoadData(IBFile* fcurfile)
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("AttrPackS::LoadData(...)");
	count_distinct_dp_loads.insert(DPID(table, column, dp));
#endif
	ConstructionLock();

	try
	{
		Construct();
		no_nulls = 0;
		optimal_mode = 0;
		uchar head[13];
		int buf_used = fcurfile->Read(head, 13);
		// WARNING: if the first 4 bytes are NOT going to indicate file size, change CheckPackFileSize in RCAttr !
		uint total_size = previous_size = *((uint *)head);
		optimal_mode = head[4];
		no_obj = *((ushort*)(head+5));
		no_obj++;
		no_nulls = *((ushort*)(head+7));
		data_full_byte_size = *((uint*)(head + 9));

		if(!IsModeValid()) {
			rclog << lock << "Error: Unexpected byte in data pack (AttrPackS)."  << unlock;
			throw DatabaseRCException("Unexpected byte in data pack (AttrPackS).");
		}
		else {
			if(IsModePackEmpty()) {
				no_obj = 0;
				no_nulls = 0;
			}
			else if(IsModeNullsOnly()) {
				no_nulls=no_obj;
				for(uint i = 0; i < no_nulls; i++)
					SetNull(i);
			} else if(IsModeNoCompression()) {
				LoadUncompressed(fcurfile);
			} else {
				// normal compression
				index = 0;
				if(IsModeCompressionApplied())	// 0,1	= Huffmann 16-bit compression of remainings with tree (0 - flat null encoding, 1 - compressed nulls)
				{
					comp_buf_size = total_size - 13;
					dealloc(compressed_buf);
					compressed_buf = 0;
					compressed_buf = (uchar*)alloc((comp_buf_size+1)*sizeof(uchar), BLOCK_COMPRESSED);

					if(!compressed_buf) {
						rclog << lock << "Error: out of memory (" << comp_buf_size + 1 << " bytes failed). (32)" << unlock;
						throw OutOfMemoryRCException();
					}
					buf_used=fcurfile->Read((char*)compressed_buf, comp_buf_size);
				}
				else
				{
					BHASSERT(0, "The process reached to an invalid code path! (LoadData)");
				}

				compressed_up_to_date = true;
				saved_up_to_date = true;
			}
		}
		max_no_obj = no_obj;
		is_empty = false;
#ifdef FUNCTIONS_EXECUTION_TIMES
		NoBytesReadByDPs += total_size;
#endif
	}
	catch (DatabaseRCException& e)
	{
		ConstructionUnlock();
		throw e;
	}

	ConstructionUnlock();
}

void AttrPackS::Destroy()
{
	if(data) {

		int ds = (int) rc_msize(data) / sizeof(char*);
		for (int i = 0; i < ds && i < data_id; i++) {
			dealloc(data[i]);
			data[i] = 0;
		}
		dealloc(data);
		data = 0;
	}

	dealloc(index);
	index = 0;

	dealloc(lens);
	lens = 0;
	is_empty = true;

	dealloc(nulls);
	dealloc(compressed_buf);
	compressed_buf = 0;
	nulls = 0;
}

AttrPackS::~AttrPackS()
{
	DestructionLock();
	Destroy();
}

void AttrPackS::Construct()
{
	data = 0;
	index = 0;
	lens = 0;
	max_no_obj = no_obj = 0;
	compressed_up_to_date = false;
	saved_up_to_date = false;

//	values_psbs = 0;

	if(attr_type == RC_BIN)
		len_mode = sizeof(uint);
	else
		len_mode = sizeof(ushort);
	//cout << "len_mode = " << len_mode << endl;

	last_copied = -1;
	binding = false;
	nulls = 0;
	no_nulls = 0;
	data_full_byte_size = 0;
	compressed_buf = 0;
	data_id = 0;
	optimal_mode = 0;
	comp_buf_size = comp_null_buf_size = comp_len_buf_size = 0;
//	memset(predictor_hist, 0, 32);
	//value_t = 0;
}

void AttrPackS::Expand(int no_obj)
{
	BHASSERT(!is_only_compressed, "The pack is compressed!");
	BHASSERT(this->no_obj + no_obj <= MAX_NO_OBJ, "Expression evaluation failed on MAX_NO_OBJ!");
	max_no_obj = this->no_obj + no_obj;
	index = (uchar**)rc_realloc(index, max_no_obj * sizeof(uchar*), BLOCK_UNCOMPRESSED);
	lens = rc_realloc(lens, len_mode * max_no_obj * sizeof(char), BLOCK_UNCOMPRESSED);
	//lens = realloc(lens, len_mode * max_no_obj);	 // warning C6308: 'realloc' might return null pointer: assigning null pointer to 'lens', which is passed as an argument to 'realloc', will cause the original memory block to be leaked

	memset((char*)lens + (this->no_obj * len_mode), 0, (no_obj * len_mode));

	int cs = data ? (int)rc_msize(data) : 0;
	data = (uchar**)rc_realloc(data, cs + (no_obj * sizeof(uchar*)), BLOCK_UNCOMPRESSED);

	memset((char*)data + cs, 0, no_obj * sizeof(uchar*));
	cs /= sizeof(uchar*);

	for(uint o = this->no_obj; o < max_no_obj; o++)
		index[o] = data[cs++];
}

void AttrPackS::BindValue(bool null, uchar* value, uint size)
{

	if(!binding)
	{
		binding = true;
		last_copied = (int)no_obj - 1;
	}

	if(null)
	{
		SetNull(no_obj);
		index[no_obj] = 0;
	}
	else if(size == 0)
	{
		SetSize(no_obj, 0);
		index[no_obj] = 0;
	}
	else
	{
		data_full_byte_size += size;
		SetSize(no_obj, size);
		index[no_obj] = value;
	}
	no_obj++;
	compressed_up_to_date = false;
}

void AttrPackS::CopyBinded()
{
	BHASSERT(!is_only_compressed, "The pack is compressed!");
	if(binding)
	{
		int size_sum = 0;
		for(uint i = last_copied + 1; i < no_obj; i++)
			size_sum += GetSize(i);
		if(size_sum > 0)
		{
			data[data_id] = (uchar*)alloc(size_sum, BLOCK_UNCOMPRESSED);

			size_sum = 0;
			for(uint i = last_copied + 1; i < no_obj; i++)
			{
				int size = GetSize(i);
				if(!IsNull(i) && size != 0)
				{
					uchar* value = (uchar*&)index[i];
					memcpy(data[data_id] + size_sum, value, size);
					index[i] = data[data_id] + size_sum;
					size_sum += size;
				}
				else
					index[i] = 0;
			}
		}
		else
			data[data_id] = 0;
		data_id++;
	}
	last_copied = no_obj -1;
	binding = false;
}


void AttrPackS::SetNull(int ono)
{
	BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!is_only_compressed);
	if(nulls == 0)
	{
		nulls = (uint*)alloc(8192, BLOCK_UNCOMPRESSED);
		memset(nulls, 0, 8192);
	}
	int mask = (uint)(1) << (ono%32);
	if((nulls[ono>>5] & mask) == 0)
	{
		nulls[ono>>5] |= mask;		// set the i-th bit of the table
		no_nulls++;
	}
	else
	{
		BHASSERT(0, "Expression evaluation failed!");
	}
}

void AttrPackS::SetSize(int ono, uint size)
{
	BHASSERT(!is_only_compressed, "The pack is compressed!");
	if(len_mode == sizeof(ushort))
		((ushort*)lens)[ono] = (ushort)size;
	else
		((uint*)lens)[ono] = (uint)size;
}

int AttrPackS::GetVal_T(int ono, char* value_t)
{
	if(IsNull(ono)) return 0;
	uint size = GetSize(ono);
	Convert2Hex(index[ono], size, value_t, size*2, false);
	return size*2;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////

void AttrPackS::Uncompress()
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	FETOperator feto("AttrPackS::Uncompress()");
	count_distinct_dp_decompressions.insert(DPID(table, column, dp));
#endif
	if(IsModeNoCompression())
		return;

	is_only_compressed = false;
	if(optimal_mode == 0 && ATI::IsBinType(attr_type)) {
		rclog << lock << "Error: Compression format no longer supported." << unlock;
		return;
	}

	AllocBuffers();
	char** tmp_index = (char**) alloc(no_obj * sizeof(char*), BLOCK_TEMPORARY);
	////////////////////////////////////////////////////////
	int i; //,j,obj_317,no_of_rep,rle_bits;
	uint *cur_buf = (uint*) compressed_buf;
	char *cur_pos = (char*) *data;
	int i_bit = 0; // buffer position
	char *last_not_null[32];
	int last_not_null_size[32];
	for(i = 0; i < 32; i++) {
		last_not_null[i] = NULL;
		last_not_null_size[i] = 0;
	}

	///////////////////////////////////////////////////////
	// decode nulls
	char (*FREE_PLACE)(reinterpret_cast<char*> (-1));

	uint null_buf_size = 0;
	if(no_nulls > 0) {
		null_buf_size = (*(ushort*) cur_buf);
		if(!IsModeNullsCompressed()) // flat null encoding
			memcpy(nulls, (char*) cur_buf + 2, null_buf_size);
		else {
			BitstreamCompressor bsc;
			bsc.Decompress((char*) nulls, null_buf_size, (char*) cur_buf + 2, no_obj, no_nulls);
			// For tests:
			uint nulls_counted = 0;
			for(i = 0; i < 2048; i++)
				nulls_counted += CalculateBinSum(nulls[i]);
			if(no_nulls != nulls_counted)
				rclog << lock << "Error: AttrPackT::Uncompress uncompressed wrong number of nulls." << unlock;
		}
		cur_buf = (uint*) ((char*) cur_buf + null_buf_size + 2);
		for(i = 0; i < (int) no_obj; i++) {
			if(IsNull(i))
				tmp_index[i] = NULL; // i.e. nulls
			else
				tmp_index[i] = FREE_PLACE; // special value: an object awaiting decoding
		}
	} else
		for(i = 0; i < (int) no_obj; i++)
			tmp_index[i] = FREE_PLACE;

	///////////////////////////////////////////////////////
	//	<null_buf_size><nulls><lens><char_lib_size><huffmann_size><huffmann><rle_bits><obj>...<obj>
	if(optimal_mode == 0) {
		rclog << lock << "Error: Compression format no longer supported." << unlock;
	} else {
		comp_len_buf_size = *cur_buf;
		if((_uint64) * (cur_buf + 1) != 0) {
			NumCompressor<uint> nc;
			uint* cn_ptr = (uint*) alloc((1 << 16) * sizeof(uint), BLOCK_TEMPORARY);

			nc.Decompress(cn_ptr, (char*) (cur_buf + 2), comp_len_buf_size - 8, no_obj - no_nulls, *(uint*) (cur_buf + 1));

			int oid = 0;
			for(uint o = 0; o < no_obj; o++)
				if(!IsNull(int(o)))
					SetSize(o, (uint) cn_ptr[oid++]);

			dealloc(cn_ptr);
		} else {
			for(uint o = 0; o < no_obj; o++)
				if(!IsNull(int(o)))
					SetSize(o, 0);
		}

		cur_buf = (uint*) ((char*) (cur_buf) + comp_len_buf_size);
		TextCompressor tc;
		int dlen = *(int*) cur_buf;
		cur_buf += 1;
		int zlo = 0;
		for(uint obj = 0; obj < no_obj; obj++)
			if(!IsNull(obj) && GetSize(obj) == 0)
				zlo++;
		int objs = no_obj - no_nulls - zlo;

		ushort* tmp_len = 0;
		if(objs) {
			tmp_len = (ushort*) alloc(objs * sizeof(ushort), BLOCK_TEMPORARY);

			for(uint tmp_id = 0, id = 0; id < no_obj; id++)
				if(!IsNull(id) && GetSize(id) != 0)
					tmp_len[tmp_id++] = GetSize(id);
			tc.Decompress((char*) *data, data_full_byte_size, (char*) cur_buf, dlen, tmp_index, tmp_len, objs);

			dealloc(tmp_len);
		}
	}

	for(uint tmp_id = 0, id = 0; id < no_obj; id++) {
		if(!IsNull(id) && GetSize(id) != 0)
			index[id] = (uchar*) tmp_index[tmp_id++];
		else {
			SetSize(id, 0);
			index[id] = 0;
		}
		if(optimal_mode == 0)
			tmp_id++;
	}
	dealloc(tmp_index);

#ifdef FUNCTIONS_EXECUTION_TIMES
	SizeOfUncompressedDP += (1 + rc_msize(*data) + rc_msize(nulls) + rc_msize(lens) + rc_msize(index));
#endif
}

void AttrPackS::Compress()
{
#ifdef FUNCTIONS_EXECUTION_TIMES
	std::stringstream s1;
	s1 << "aS[" << column << "].Compress(...)";
	FETOperator fet1(s1.str());
#endif

	//SystemInfo si;
	//uint nopf_start = SystemInfo::NoPageFaults();
	//uint nopf_end = 0;
	if(no_obj==0)
	{
		SetModePackEmpty();
		return;		// nothing to compress
	}
	if(no_obj==no_nulls)
	{
		SetModeNullsOnly();
		return;		// nothing to compress
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////
	// Compress nulls:

	comp_len_buf_size = comp_null_buf_size = comp_buf_size = 0;
	uchar *comp_null_buf = 0;
	if(no_nulls > 0) {
		if(ShouldCompress()) {
			comp_null_buf = (uchar*)nulls;
			comp_null_buf_size = ((no_obj+7)/8);
			ResetModeNullsCompressed();
		} else {
			comp_null_buf_size = ((no_obj+7)/8);
			comp_null_buf = (uchar*)alloc((comp_null_buf_size + 2)*sizeof(uchar), BLOCK_TEMPORARY);

			uint cnbl = comp_null_buf_size+1;
			comp_null_buf[cnbl]=0xBA;			// just checking - buffer overrun
			BitstreamCompressor bsc;
			if(CPRS_ERR_BUF == bsc.Compress((char*)comp_null_buf, comp_null_buf_size, (char*)nulls, no_obj, no_nulls))
			{
				if(comp_null_buf[cnbl]!=0xBA)
				{
					rclog << lock << "ERROR: buffer overrun by BitstreamCompressor (T f)." << unlock;
					BHASSERT(0, "ERROR: buffer overrun by BitstreamCompressor (T f)!");
				}

				dealloc(comp_null_buf);
				comp_null_buf = (uchar*)nulls;
				comp_null_buf_size = ((no_obj+7)/8);
				ResetModeNullsCompressed();
			} else {
				SetModeNullsCompressed();
				if(comp_null_buf[cnbl]!=0xBA)
				{
					rclog << lock << "ERROR: buffer overrun by BitstreamCompressor (T)." << unlock;
					BHASSERT(0, "ERROR: buffer overrun by BitstreamCompressor (T f)!");
				}
			}
		}
	}

	uint *comp_len_buf = 0;

	NumCompressor<uint> nc(inserting_mode == 1 && no_obj < 0x10000);
	uint* nc_buffer = (uint*)alloc((1<<16)*sizeof(uint), BLOCK_TEMPORARY);

	int onn = 0;
	uint maxv = 0;
	uint cv = 0;
	for(uint o = 0; o < no_obj; o++)
		if(!IsNull(o))
		{
			cv = GetSize(o);
			*(nc_buffer + onn++) = cv;
			if(cv > maxv)
				maxv = cv;
		}
		if(maxv != 0)
		{
			comp_len_buf_size = onn * sizeof(uint) + 28;
			//comp_len_buf = new uint [comp_len_buf_size/4];
			comp_len_buf = (uint*)alloc(comp_len_buf_size/4 * sizeof(uint), BLOCK_TEMPORARY);
			uint tmp_comp_len_buf_size = comp_len_buf_size - 8;
			nc.Compress((char*)(comp_len_buf + 2), tmp_comp_len_buf_size, nc_buffer, onn, maxv);
			//cout << "tmp = " << tmp_comp_len_buf_size << ", " << " size = " << comp_len_buf_size << endl;
			comp_len_buf_size = tmp_comp_len_buf_size + 8;
		}
		else
		{
			comp_len_buf_size = 8;
			//comp_len_buf = new uint [2];
			comp_len_buf = (uint*)alloc(sizeof(uint) * 2, BLOCK_TEMPORARY);
		}

		dealloc(nc_buffer);

		*comp_len_buf = comp_len_buf_size;
		*(comp_len_buf + 1) = maxv;

		TextCompressor tc;
		int zlo = 0;
		for(uint obj = 0; obj < no_obj; obj++)
			if(!IsNull(obj) && GetSize(obj) == 0)
				zlo++;
		int dlen = data_full_byte_size + 10;
		//cout << data_full_byte_size << endl;

		char* comp_buf = (char*)alloc(dlen * sizeof(char), BLOCK_TEMPORARY);

		if(data_full_byte_size)
		{
			int objs = (no_obj - no_nulls) - zlo;

			char** tmp_index = (char**)alloc(objs*sizeof(char*), BLOCK_TEMPORARY);
			ushort* tmp_len = (ushort*)alloc(objs*sizeof(ushort), BLOCK_TEMPORARY);

			//_uint64 ss = 0;
			int nid = 0;
			for(int id = 0; id < (int)no_obj; id++)
				if(!IsNull(id) && GetSize(id) != 0)
				{
					tmp_index[nid] = (char*)index[id];
					tmp_len[nid++] = GetSize(id);
					//ss += GetSize(id);
				}

			tc.Compress(comp_buf, dlen, tmp_index, tmp_len, objs, inserting_mode == 1 && no_obj < 0x10000 ? 0 : TextCompressor::VER);

			dealloc(tmp_index);
			dealloc(tmp_len);
		}
		else
		{
			dlen = 0;
		}

		comp_buf_size = (comp_null_buf_size > 0 ? 2 + comp_null_buf_size : 0) + comp_len_buf_size + 4 + dlen;

		uchar* new_compressed_buf = (uchar*)alloc(comp_buf_size*sizeof(uchar), BLOCK_COMPRESSED);

		if(!new_compressed_buf)
			rclog << lock << "Error: out of memory (" << comp_buf_size << " bytes failed). (37)" << unlock;
		uchar* p = new_compressed_buf;

		////////////////////////////////Nulls
		if(no_nulls>0)
		{
			*((ushort*)p) = (ushort)comp_null_buf_size;
			p+=2;
			memcpy(p, comp_null_buf, comp_null_buf_size);
			p+=comp_null_buf_size;
		}

		////////////////////////////////	Lens

		if(comp_len_buf_size)
			memcpy(p, comp_len_buf, comp_len_buf_size);

		p += comp_len_buf_size;

		*((int*)p) = dlen;
		p+=sizeof(int);
		if(dlen)
			memcpy(p, comp_buf, dlen);

		dealloc(comp_buf);
		dealloc(compressed_buf);

		compressed_buf = new_compressed_buf;
		dealloc(comp_len_buf);
		if(comp_null_buf != (uchar*)nulls)
		{
			dealloc(comp_null_buf);
		}
		//optimal_mode |= (1 << 1);
		SetModeDataCompressed();
		compressed_up_to_date = true;
}

void AttrPackS::StayCompressed()
{
	if(!compressed_up_to_date)
		saved_up_to_date = false;
	if(compressed_buf == 0 || !compressed_up_to_date)
		Compress();

	if(data)
	{
		int ds = (int)rc_msize(data) / sizeof(char*);
		for(int i = 0; i < ds; i++)
			dealloc(data[i]);
		dealloc(data);
		data = 0;
	}

	dealloc(nulls);
	dealloc(index);
	dealloc(lens);
	index = 0;
	lens = 0;
	nulls = 0;
	//cout << "Attr " << attr->Name() << ", pack " << pack_no << " compressed." << endl;
	is_only_compressed = true;
}

int AttrPackS::Save(IBFile* fcurfile)
{
	if(compressed_buf==NULL || !compressed_up_to_date)
	{
		Compress();
	}
	uint total_size=0;
	uchar head[13];

	head[4]=optimal_mode;
	if(!IsModePackVirtual())
	{
		*((ushort*)(head+5))=(ushort)(no_obj-1);
		*((ushort*)(head+7))=(ushort)(no_nulls);
	}
	else
		*((uint *)(head+5))=0;

	*((uint*)(head+9))=data_full_byte_size;
	//EnterCriticalSection(&rc_save_pack_cs);
	if(IsModeCompressionApplied())
	{
		total_size = 13 + comp_buf_size ;
		*((uint*)head)=total_size;
		fcurfile->Write((char*)head,13);
		fcurfile->Write((char*)compressed_buf, comp_buf_size);
	}
	else if(IsModePackVirtual())		// optimal_mode==255, 254
	{
		total_size=13;
		*((uint*)head)=total_size;
		fcurfile->Write((char*)head,13);
	} else if(IsModeNoCompression()) {
		SaveUncompressed(head, fcurfile);
	} else {
		assert("unexpected PackS mode");
		rclog << lock << "unexpected PackS mode " << optimal_mode << " << unlock";
	}
	//LeaveCriticalSection(&rc_save_pack_cs);
	previous_size = TotalSaveSize();
	saved_up_to_date = true;
	return 0;
}

uint AttrPackS::TotalSaveSize()
{
	uint tot = 13;
	if(this->IsModeCompressionApplied() )
		tot += comp_buf_size;
	else if(IsModeNoCompression()) {
		uint string_size = 0;
		for(int i=0; i< no_obj; i++) {
			if(!IsNull(i))
				string_size += GetSize(i);
		}
		tot+= (data ? string_size: 0) + (nulls ? (no_obj+7)/8 : 0) +
				(lens ? len_mode*no_obj : 0);
	}

	return tot;
}

void AttrPackS::SaveUncompressed(uchar* head, IBFile* fcurfile)
{

	uint total_size=0;
	uint string_size = 0;
	for(int i=0; i< no_obj; i++) {
		if(!IsNull(i))
			string_size += GetSize(i);
	}
	total_size = TotalSaveSize();
	*((uint*)head)=total_size;
	fcurfile->Write((char*)head,13);
	if(data) {
		char* tmpbuf = (char*)alloc(string_size * sizeof(char), BLOCK_TEMPORARY);
		char* tmpptr = tmpbuf ;
		for(int i=0; i< no_obj; i++) {
			if(!IsNull(i)) {
				memcpy(tmpptr, GetVal(i), GetSize(i));
				tmpptr+= GetSize(i);
			}
		}
		fcurfile->Write((char*)tmpbuf,string_size);
		dealloc(tmpbuf);
	}

	if(nulls)
		fcurfile->Write((char*)nulls,(no_obj+7)/8);
	if(lens)
		fcurfile->Write((char*)lens,len_mode*no_obj);
}

void AttrPackS::LoadUncompressed(IBFile* fcurfile)
{
	AllocBuffers();
	if(data)
		fcurfile->Read((char*)*data,data_full_byte_size);
	if(nulls)
		fcurfile->Read((char*)nulls,(no_obj+7)/8);
	if(lens)
		fcurfile->Read((char*)lens,len_mode*no_obj);
	int sumlen = 0;
	for(int i=0; i< no_obj; i++){
		if(!IsNull(i) && GetSize(i) != 0) {
			index[i] = ((uchar*) (*data)) + sumlen;
			sumlen += GetSize(i);
		} else {
			SetSize(i, 0);
			index[i] = 0;
		}
	}
}

void AttrPackS::AllocBuffers()
{
	if(data == NULL) {
		data = (uchar**) alloc(sizeof(uchar*), BLOCK_UNCOMPRESSED);
		if(data_full_byte_size) {
			*data = (uchar*) alloc(data_full_byte_size, BLOCK_UNCOMPRESSED);
			if(!*data)
				rclog << lock << "Error: out of memory (" << data_full_byte_size << " bytes failed). (40)" << unlock;
		} else
			*data = 0;
		data_id = 1;
	}
	assert(!lens && !index);
	lens = alloc(len_mode * no_obj * sizeof(char), BLOCK_UNCOMPRESSED);
	index = (uchar**)alloc(no_obj * sizeof(uchar*), BLOCK_UNCOMPRESSED);
	if(no_nulls > 0) {
		if(!nulls)
			nulls = (uint*) alloc(2048 * sizeof(uint), BLOCK_UNCOMPRESSED);
		if(no_obj < 65536)
			memset(nulls, 0, 8192);
	}
}
