/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <fstream>

#include "RSI_PackPack.h"

using namespace std;

RSIndex_PP::RSIndex_PP()				// create an empty index
{
	t=NULL;
	u=NULL;
	no_obj1=0;
	no_obj2=0;
	no_pack1=0;
	no_pack2=0;
}

RSIndex_PP::~RSIndex_PP()
{
	delete t;
	delete u;
}

void RSIndex_PP::Clear()
{
	delete t;
	t=NULL;
	delete u;
	u=NULL;
	no_obj1=0;
	no_obj2=0;
	no_pack1=0;
	no_pack2=0;
}

void RSIndex_PP::Create(_int64 _no_obj1,_int64 _no_obj2)
{
	delete t;
	delete u;
	no_obj1	= _no_obj1;
	no_obj2	= _no_obj2;
	no_pack1= bh::common::NoObj2NoPacks(no_obj1);
	no_pack2= bh::common::NoObj2NoPacks(no_obj2);
	t = new BinMatrix(no_pack1, no_pack2, true);
	u = new BinMatrix(no_pack1, no_pack2);
}

void RSIndex_PP::Verify(_int64 _no_obj1,_int64 _no_obj2)
{
	if(_no_obj1>no_obj1)
	{
		for(int pack=0;pack<no_pack2;pack++)			// invalidate the last column
			SetUpToDate(no_pack1-1,pack,false);
	}
	if(_no_obj2>no_obj2)
	{
		for(int pack=0;pack<no_pack1;pack++)			// invalidate the last row
			SetUpToDate(pack,no_pack2-1,false);
	}
}

int RSIndex_PP::Load(IBFile *frs_index, int unused)
{
	// File format:
	// Header:	<no_obj1><no_obj2>						- 8+8 = 16 bytes
	//			<table_t>								- (no_pack2/32 + 1)*no_pack1 integers
	delete t;
	delete u;
	t = 0;
	u = 0;
	unsigned char header[17];
	int bytes;
	bytes = frs_index->Read(header,17);
	if(bytes<16)
	{
		no_obj1	= 0;
		no_obj2	= 0;
		return 1;
	}
	if(*header != current_verion)
	{
		assert(0);
		return 1;
	}
	no_obj1	= *(_int64*)(header+sizeof(uchar));
	no_obj2	= *(_int64*)(header+8+sizeof(uchar));
	no_pack1= bh::common::NoObj2NoPacks(no_obj1);
	no_pack2= bh::common::NoObj2NoPacks(no_obj2);
	t = new BinMatrix(no_pack1, no_pack2);
	u = new BinMatrix(no_pack1, no_pack2);
	bytes = frs_index->Read(t->Buf(),t->BufSize());
	if((uint)bytes != t->BufSize() || frs_index->Read(u->Buf(), u->BufSize()) != u->BufSize())
	{
		no_obj1	= 0;
		no_obj2	= 0;
		return 1;
	}
	return 0;
}

int RSIndex_PP::Save(IBFile *frs_index, int unused)
{
	unsigned char header[17];
	int bytes;
	*header = current_verion;
	*(_int64*)(header+sizeof(char))		= no_obj1;
	*(_int64*)(header+8+sizeof(char))	= no_obj2;
	bytes = frs_index->WriteExact(header,17);

	if(frs_index->WriteExact(t->Buf(), t->BufSize()) != t->BufSize())
		return 1;
	if(frs_index->WriteExact(u->Buf(), u->BufSize()) != u->BufSize())
		return 1;
	return 0;
}

int RSIndex_PP::LoadLastOnly(IBFile *frs_index, int unused)
{
	return Load(frs_index, unused);
}

void RSIndex_PP::Update(_int64 _new_no_obj1, _int64 _new_no_obj2)
{
	no_obj1	= _new_no_obj1;
	no_obj2	= _new_no_obj2;
	no_pack1= bh::common::NoObj2NoPacks(no_obj1);
	no_pack2= bh::common::NoObj2NoPacks(no_obj2);
	if( no_pack1 == 0 || no_pack2 == 0 )
		return;
	if(t)	t->Resize(no_pack1, no_pack2, true);
	else	t = new BinMatrix(no_pack1, no_pack2, true);
	if(u)	u->Resize(no_pack1, no_pack2);
	else	u = new BinMatrix(no_pack1, no_pack2);
}

double RSIndex_PP::Density(int pack2)
{
	if(_int64(no_pack1)*no_pack2==0) return 0;
	_int64 res=0;
	if(pack2==-1)
	{
		return t->Density();
	}
	for(int i=0;i<no_pack1;i++)
		if(GetValue(i,pack2))
			res++;
	if(res==no_pack1) return 1;
	return double(res)/no_pack1;
}

bool RSIndex_PP::AllUpdated(BinMatrix &not_touched)
{
	return u->SumIsFull(not_touched);
}

bool RSIndex_PP::AllUpdated()
{
	return u->IsFull();
}

#ifdef __WIN__
#undef close
#endif

void RSIndex_PP::SaveAsText(char *fname)
{
	string p = infobright_data_dir + fname;
	ofstream out( p.c_str(), ios_base::out | ios_base::app );
	out << endl;
	out << "[ " << no_pack1 << " " << no_pack2 << " ]" << endl;
	for( int i=0; i<no_pack1; i++ ) {
		for( int j=0; j<no_pack2; j++ )
			out << ( u->Get( i, j ) ? ( t->Get( i, j ) ? "1" : "0" ) : "?" );
		out << endl;
	}
	out.close();
}
