/* Copyright (C)  2005-2009 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */


#include "SorterWrapper.h"
#include "MIIterator.h"

SorterWrapper::SorterWrapper(MultiIndex &_mind, _int64 _limit)
{
	no_of_rows = _mind.NoTuples();
	limit = _limit;
	s = NULL;
	buf_size = 0;
	cur_val = NULL;
	input_buf = NULL;
}

SorterWrapper::~SorterWrapper()
{
	delete s;
	delete [] input_buf;
}

void SorterWrapper::AddSortedColumn(VirtualColumn *col, int sort_order, bool in_output)
{
	input_cols.push_back(SorterColumnDescription(col, sort_order, in_output));
}

void SorterWrapper::InitSorter()
{
	int total_bytes = 0;
	int key_bytes = 0;

	// TODO: define virtual encoding (by tuple numbers for each dim)

	// Pass 1: identify columns to order by, exclude unnecessary ones
	vector<int> order_by;
	for(int i = 0; i < input_cols.size(); i++)
		order_by.push_back(-1);			// -1 means "no column"

	for(int i = 0; i < input_cols.size(); i++) {
		int ord = input_cols[i].sort_order;
		if(ord < 0)
			ord = -ord;
		if(ord != 0) {
			assert(order_by[ord - 1] == -1);
			order_by[ord - 1] = i;
		}
	}

	for(int i = 0; i < input_cols.size(); i++) // check whether all sorting columns are really needed
		if(order_by[i] != -1) {
			assert(i == 0 || order_by[i - 1] != -1);	// assure continuity
			if(input_cols[order_by[i]].col->IsDistinct()) {
				// do not sort by any other column:
				for(int j = i + 1; j < input_cols.size(); j++)
					if(order_by[j] != -1) {
						input_cols[order_by[j]].sort_order = 0;
						order_by[j] = -1;
				}
				break;
			}
	}

	// Pass 2: determine encodings, provide offsets

	for(int i = 0; i < input_cols.size(); i++) {
		int flags = 0;
		if(input_cols[i].sort_order != 0)
			flags |= ColumnBinEncoder::ENCODER_MONOTONIC;
		if(input_cols[i].sort_order < 0)
			flags |= ColumnBinEncoder::ENCODER_DESCENDING;
		if(input_cols[i].in_output)
			flags |= ColumnBinEncoder::ENCODER_DECODABLE;
		scol.push_back(ColumnBinEncoder(flags));
		scol[i].PrepareEncoder(input_cols[i].col);
	}

	int j = 0;
	while(j < input_cols.size() && order_by[j] != -1) {
		scol[order_by[j]].SetPrimaryOffset(key_bytes);
		key_bytes += scol[order_by[j]].GetPrimarySize();
		total_bytes += scol[order_by[j]].GetPrimarySize();
		j++;
	}

	for(int i = 0; i < scol.size(); i++) {
		if(scol[i].IsEnabled()) {
			if(input_cols[i].sort_order == 0) {
				scol[i].SetPrimaryOffset(total_bytes);
				total_bytes += scol[i].GetPrimarySize();
			}
			else if(scol[i].GetSecondarySize() > 0) {		// secondary place needed for sorted columns?
				scol[i].SetSecondaryOffset(total_bytes);
				total_bytes += scol[i].GetSecondarySize();
			}
		}
	}

	// Everything prepared - create sorter
	if(total_bytes > 0) {			// the sorter is nontrivial
		s = Sorter3::CreateSorter(no_of_rows, key_bytes, total_bytes, limit);
		input_buf = new unsigned char [total_bytes];
		rccontrol.lock(s->conn->GetThreadID()) << "Sorter initialized for " << no_of_rows << " rows, " << key_bytes << "+" << total_bytes - key_bytes << " bytes each." << unlock;
	}
}

bool SorterWrapper::PutValues(MIIterator &mit)
{
	if(s == NULL)
		return false;		// trivial sorter (constant values)
	for(int i = 0; i < scol.size(); i++)
		if(scol[i].IsEnabled() && scol[i].IsNontrivial()) {	// constant num. columns are trivial
			if(mit.PackrowStarted())
				scol[i].LockSourcePacks(mit);
			scol[i].Encode(input_buf, mit);
		}
	return s->PutValue(input_buf);
}

bool SorterWrapper::FetchNextRow()
{
	if(s == NULL) {
		no_of_rows--;
		return (no_of_rows >= 0);		// trivial sorter (constant values, cur_val is always NULL)
	}
	cur_val = s->GetNextValue();
	return (cur_val != NULL);
}
