/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

/////////////////////////////////////////////////////////////////////////////////////////////////////
// This is a part of TempTable implementation concerned with the query execution low-level mechanisms
/////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "TempTable.h"
#include "MIOrderedIterator.h"
#include "PackGuardian.h"
#include "GroupByWrapper.h"
#include "system/RCSystem.h"
#include "system/ConnectionInfo.h"
#include "core/tools.h"

using namespace std;

#define SIZE_2_POW_47	140737488355326LL

//////////////////////////////////////////////////////////////////////////////////////////////////////////

_int64 TempTable::MultiDimensionalGroupByScan(GroupByWrapper &gbw, _int64 upper_approx_of_groups, _int64 limit, _int64 offset)
{
	MEASURE_FET("TempTable::MultiDimensionalGroupByScan(...)");
	bool first_pass = true;
	_int64 no_groups = 0;
	_int64 cur_tuple = 0;		// tuples are numbered according to m_index->Next() order
	_int64 total_no_groups = 0;
	_int64 total_output_rows = 0;
	// Some statistics for display:
	_int64 packrows_found = 0;			// all packrows, except these completely omitted (as aggregated before)
	_int64 packrows_omitted = 0;		// packrows completely omitted by KNs (both grouping values and aggr. values)
	_int64 packrows_part_omitted = 0;	// packrows partially omitted (some columns/aggregations)

	// Determine dimensions to be iterated
	bool no_dims_found = true;
	DimensionVector dims(filter.m_index);
	gbw.FillDimsUsed(dims);
	for(int i = 0; i < filter.m_index->NoDimensions(); i++)
		if(dims[i]) {
			no_dims_found = false;
			break;
		}
	if(no_dims_found)
		dims[0] = true;							// at least one dimension is needed

	vector<PackOrderer> po(filter.m_index->NoDimensions());
	map<int,vector<PackOrderer::OrderingInfo> >::const_iterator oi_it;
	for(oi_it = oi.begin(); oi_it!= oi.end(); oi_it++)
		PackOrderer::ChoosePackOrderer(po[(*oi_it).first],(*oi_it).second);

	MIIterator mit(filter.m_index, dims);
	//MIOrderedIterator mit(filter.m_index, dims, po);
	_int64 factor = mit.Factor();
	if( mit.NoTuples() == NULL_VALUE_64 || mit.NoTuples() >= SIZE_2_POW_47 ) {	// 2^47, a limit for filter below
		throw OutOfMemoryRCException("Aggregation is too large.");
	}
	::Filter tuple_left(mit.NoTuples());
	tuple_left.Set();
	gbw.SetDistinctTuples(tuple_left.NoObj());
	do {
		if(rccontrol.isOn())  {
			if(upper_approx_of_groups == 1)
				rccontrol.lock(m_conn.GetThreadID()) << "Aggregating: " << tuple_left.NoOnes() << " tuples left." << unlock;
			else
				rccontrol.lock(m_conn.GetThreadID()) << "Aggregating: " << tuple_left.NoOnes() << " tuples left, " << total_no_groups << " gr. found so far" << unlock;
		}
		cur_tuple = 0;
		no_groups = 0;					// locally created in this pass
		gbw.ClearDistinctBuffers();		// reset buffers for a new contents
		gbw.AddAllGroupingConstants(mit);
		mit.Rewind();	// aggregated rows will be massively omitted packrow by packrow
		bool require_locking_ag = true;
		bool require_locking_gr = true;
		_int64 uniform_pos = NULL_VALUE_64;			// if the packrow is uniform, its group id. is stored here
		bool aggregations_not_changeable = false;
		while(mit.IsValid()) { ///////////////////////// First stage - some distincts may be delayed
			if(mit.PackrowStarted()) {
				if(m_conn.killed())
					throw KilledRCException();

				/////////////// All packrow-level operations /////////
				gbw.ResetPackrow();
				bool skip_packrow = false;
				bool packrow_done = false;
				bool part_omitted = false;
				aggregations_not_changeable = false;
				_int64 packrow_length = mit.GetPackSizeLeft();
				// Check whether the packrow contain any not aggregated rows
				if(tuple_left.IsEmptyBetween(cur_tuple, cur_tuple + packrow_length - 1))
					skip_packrow = true;
				else {
					_int64 rows_in_pack = tuple_left.NoOnesBetween(cur_tuple, cur_tuple + packrow_length - 1);
					packrows_found++;					// for statistics
					skip_packrow = AggregateRough(gbw, mit, packrow_done, part_omitted,
													aggregations_not_changeable, no_groups,
													uniform_pos, rows_in_pack, factor);
					if(total_no_groups + no_groups == upper_approx_of_groups) { // no more groups!
						gbw.SetNoMoreGroups();
						if(skip_packrow)			// no aggr. changeable and no new groups possible?
							packrow_done = true;
						if(gbw.NoGroupingAttr() == gbw.NoAttr()) {	// just DISTINCT without grouping
							tuple_left.Reset();	// no more rows needed, just produce output
							break;
						}
					}
					if(skip_packrow)
						packrows_omitted++;
					else if(part_omitted)
						packrows_part_omitted++;
					if(packrow_done) { // This packrow will not be needed any more
						tuple_left.CommitPack(uint(cur_tuple >> 16));		// commit the last pack to prevent mixing delayed and non-delayed filter access
						tuple_left.ResetBetween(cur_tuple, cur_tuple + packrow_length - 1);
					}
				}
				if(skip_packrow) {
					mit.NextPackrow();
					cur_tuple += packrow_length;
					continue;
				}
				require_locking_ag  = true;					// a new packrow, so locking will be needed
				require_locking_gr	= (uniform_pos == NULL_VALUE_64);	// do not lock if the grouping row is uniform
			}

			/////////////// All row-level operations ////////////
			if(tuple_left.Get(cur_tuple)) {
				if(require_locking_gr) {
					for(int gr_a = 0; gr_a < gbw.NoGroupingAttr(); gr_a++)
						gbw.LockMultidimensionalPack(gr_a, mit);	// note: ColumnNotOmitted checked inside
					require_locking_gr = false;
				}

				_int64 pos = 0;
				bool existed = true;
				if(uniform_pos != NULL_VALUE_64)		// either uniform because of KNs, or = 0, because there is no grouping columns
					pos = uniform_pos; // existed == true, as above
				else {
					for(int gr_a = 0; gr_a < gbw.NoGroupingAttr(); gr_a++)
						if(gbw.ColumnNotOmitted(gr_a))
							gbw.PutGroupingValue(gr_a, mit);
					existed = gbw.FindCurrentRow(pos);
				}

				if(pos != NULL_VALUE_64) {						// Any place left? If not, just omit the tuple.
					tuple_left.ResetDelayed(cur_tuple);			// delayed for optimization purposes - must be committed at the end
					if(!existed) {
						aggregations_not_changeable = false;
						no_groups++;								// successfully added
						if(total_no_groups + no_groups == upper_approx_of_groups) { // no more groups!
							gbw.SetNoMoreGroups();
							if(gbw.NoGroupingAttr() == gbw.NoAttr()) {	// just DISTINCT without grouping
								tuple_left.Reset();	// no more rows needed, just produce output
								break;
							}
						}
					}
					if(!aggregations_not_changeable) {
						// Lock packs if needed
						if(require_locking_ag) {
							for(int gr_a = gbw.NoGroupingAttr(); gr_a < gbw.NoAttr(); gr_a++)
								gbw.LockMultidimensionalPack(gr_a, mit);	// note: ColumnNotOmitted checked inside
							require_locking_ag = false;
						}

						// Prepare packs for aggregated columns
						for(int gr_a = gbw.NoGroupingAttr(); gr_a < gbw.NoAttr(); gr_a++)
							if(gbw.ColumnNotOmitted(gr_a)) {
								bool value_successfully_aggregated = gbw.PutAggregatedValue(gr_a, pos, mit, factor);
								if(!value_successfully_aggregated)
									gbw.DistinctlyOmitted(gr_a, cur_tuple);
							}
					}
				}
			}
			cur_tuple++;
			++mit;
		}
		tuple_left.Commit();					// committing delayed resets
		MultiDimensionalDistinctScan(gbw, dims);		// if not needed, no effect

		////////////////////////////////////////////////////////////////////////////
		// Now it is time to prepare output values
		if(first_pass) {
			first_pass = false;
			_int64 upper_groups = no_groups + tuple_left.NoOnes();		// upper approximation: the current size + all other possible rows
			this->CalculatePageSize(upper_groups);
			if(upper_groups > upper_approx_of_groups)
				upper_groups = upper_approx_of_groups;					// another upper limitation: not more than theoretical number of combinations

			for(uint i = 0; i < attrs.size(); i++)
				attrs[i]->CreateBuffer(upper_groups);					// note: may be more than needed
		}
		for(int occupied_row = 0; occupied_row < gbw.GetNoOfOccupiedRows(); occupied_row++) {		
			// copy GroupTable into TempTable, row by row
			if(total_output_rows >= offset && total_output_rows < offset + limit)
				AggregateFillOutput(gbw, gbw.GetOccupiedRow(occupied_row), total_no_groups);
			total_output_rows++;
		}
		if(total_output_rows > offset + limit)
			break;
		if(!tuple_left.IsEmpty())
			gbw.ClearUsed();				// prepare for the next pass, if needed
	} while(!tuple_left.IsEmpty());			// do the next pass, if anything left
	if(rccontrol.isOn())
		rccontrol.lock(m_conn.GetThreadID()) << "Aggregated (" << total_no_groups
									<< " gr). Omitted packrows: " << packrows_omitted << " + "
									<< packrows_part_omitted << " partially, out of " << packrows_found << " total." << unlock;
	return total_no_groups;
}

void TempTable::MultiDimensionalDistinctScan(GroupByWrapper& gbw, DimensionVector &dims)
{
	MEASURE_FET("TempTable::MultiDimensionalDistinctScan(GroupByWrapper& gbw)");
	while(gbw.AnyOmittedByDistinct()) {	/////////// any distincts omitted? => another pass needed
		///// Some displays
		_int64 max_size_for_display = 0;
		for(int i = gbw.NoGroupingAttr(); i < gbw.NoAttr(); i++)
			if(gbw.distinct_watch.OmittedFilter(i) && gbw.distinct_watch.OmittedFilter(i)->NoOnes() > max_size_for_display)
				max_size_for_display = gbw.distinct_watch.OmittedFilter(i)->NoOnes();
		rccontrol.lock(m_conn.GetThreadID()) << "Next distinct pass: " << max_size_for_display << " rows left" << unlock;

		gbw.RewindDistinctBuffers();			// reset buffers for a new contents, rewind cache
		MIIterator mit(filter.m_index, dims);	// NOTE: to maintain distinct cache compatibility, rows must be visited in the same order!
		for(int distinct_attr = gbw.NoGroupingAttr(); distinct_attr < gbw.NoAttr(); distinct_attr++) {
			::Filter *omit_filter = gbw.distinct_watch.OmittedFilter(distinct_attr);
			if(omit_filter && !omit_filter->IsEmpty()) {
				mit.Rewind();
				_int64 cur_tuple = 0;
				_int64 uniform_pos = NULL_VALUE_64;
				bool require_locking = true;
				while(mit.IsValid()) {
					if(mit.PackrowStarted()) {
						if(m_conn.killed())
							throw KilledRCException();
						/////////////// All packrow-level operations /////////
						omit_filter->Commit();
						gbw.ResetPackrow();
						bool skip_packrow = false;
						bool packrow_done = false;
						bool part_omitted = false;
						_int64 packrow_length = mit.GetPackSizeLeft();
						// Check whether the packrow contain any not aggregated rows
						if(omit_filter->IsEmptyBetween(cur_tuple, cur_tuple + packrow_length - 1))
							skip_packrow = true;
						else {
							_int64 rows_in_pack = omit_filter->NoOnesBetween(cur_tuple, cur_tuple + packrow_length - 1);
							_int64 dummy_no_groups = 0;
							bool agg_not_changeable = false;
							AggregateRough(gbw, mit, packrow_done, part_omitted, agg_not_changeable, dummy_no_groups,
														  uniform_pos, rows_in_pack, 1, distinct_attr);
							if(packrow_done) // This packrow will not be needed any more
								omit_filter->ResetBetween(cur_tuple, cur_tuple + packrow_length - 1);
						}
						if(skip_packrow) {
							mit.NextPackrow();
							cur_tuple += packrow_length;
							continue;
						}
						require_locking  = true;					// a new packrow, so locking will be needed
					}

					/////////////// All row-level operations ////////////
					if(omit_filter->Get(cur_tuple)) {
						bool value_successfully_aggregated = false;
						if(gbw.CacheValid(distinct_attr)) {
							value_successfully_aggregated = gbw.PutCachedValue(distinct_attr);
						}
						else {
							bool pack_omitted = false;
							// Locking etc.
							if(require_locking) {
								gbw.LockMultidimensionalPack(distinct_attr, mit);
								if(uniform_pos != PLUS_INF_64)
									for(int gr_a = 0; gr_a < gbw.NoGroupingAttr(); gr_a++)
										gbw.LockMultidimensionalPack(gr_a, mit);
								require_locking = false;
							}

							_int64 pos = 0;
							bool existed = true;
							if(uniform_pos != PLUS_INF_64)
								pos = uniform_pos; // existed == true, as above
							else { // Construct the grouping vector
								for(int gr_a = 0; gr_a < gbw.NoGroupingAttr(); gr_a++) {
									if(gbw.ColumnNotOmitted(gr_a))
										gbw.PutGroupingValue(gr_a, mit);
								}
								existed = gbw.FindCurrentRow(pos);
							}
							BHASSERT_WITH_NO_PERFORMANCE_IMPACT(existed && pos!=NULL_VALUE_64);
							value_successfully_aggregated = gbw.PutAggregatedValue(distinct_attr, pos, mit);
						}
						if(value_successfully_aggregated)
							omit_filter->ResetDelayed(cur_tuple);
						gbw.distinct_watch.gd_cache[distinct_attr].NextRead();
					}
					cur_tuple++;
					++mit;
				}
				omit_filter->Commit();	// committing delayed resets
			}
		}
		gbw.UpdateDistinctCaches();					// take into account values already counted
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////

_int64 TempTable::Aggregate(bool just_distinct, _int64 limit, _int64 offset)
{
	MEASURE_FET("TempTable::Aggregate_new(...)");

	////////////////////////////////////////////
	bool aggregation_found		= false;
	bool group_by_found			= false;

	GroupByWrapper gbw(attrs.size(), just_distinct, m_conn);
	_uint64 upper_approx_of_groups = 1;		// will remain 1 if there is no grouping columns (aggreg. only)
	
	vector<bool> utf_columns;
	vector<ColOperation> utf_orig_modes;
	for(uint i = 0; i < attrs.size(); i++) {
		utf_columns.push_back(false);
		utf_orig_modes.push_back(LISTING);
	}

	for(uint i = 0; i < attrs.size(); i++) {			// first pass: find all grouping attributes
		Attr& cur_a = *attrs[i];
		if(cur_a.mode == DELAYED)						// delayed column (e.g. complex exp. on aggregations)
			continue;
		if(( just_distinct && cur_a.alias) || cur_a.mode == GROUP_BY) {
			if(cur_a.mode == GROUP_BY)
				group_by_found = true;
			bool already_added = false;
			for(uint j = 0; j < i; j++) {
				if(*attrs[j] == cur_a) {
					already_added = true;
					gbw.DefineAsEquivalent(i,j);
					break;
				}
			}
			if(already_added == false) {
				int new_attr_number = gbw.NoGroupingAttr();
				gbw.AddGroupingColumn(new_attr_number, i, *attrs[i]);	// attrs[i] is needed

				// in case of UTF column we need to add non binary version
				// to be used for output
				if(RequiresUTFConversions(attrs[i]->GetCollation())) {
					utf_orig_modes[i] = attrs[i]->mode;
					attrs[i]->mode = LISTING;
					utf_columns[i] = true;
				}
				// approximate a number of groups
				if(upper_approx_of_groups < filter.m_index->NoTuples()) {
					_uint64 dist_vals = (_uint64)gbw.ApproxDistinctVals(new_attr_number, filter.m_index);
					if(dist_vals > filter.m_index->NoTuples() || upper_approx_of_groups * dist_vals > filter.m_index->NoTuples())
						upper_approx_of_groups = filter.m_index->NoTuples();
					else
						upper_approx_of_groups *= dist_vals;
				}
			}
		}
	}

	for(uint i = 0; i < attrs.size(); i++)	{		// second pass: find all aggregated attributes
		if(attrs[i]->mode == DELAYED)				// delayed column (e.g. complex exp. on aggregations)
			continue;
		if( (!just_distinct && attrs[i]->mode != GROUP_BY) ||	// aggregation
			( just_distinct && attrs[i]->alias == NULL ) ||
			utf_columns[i] ) {		// special case: hidden column for DISTINCT
			Attr& cur_a = *attrs[i];
			bool already_added = false;
			for(uint j = 0; j < i; j++) {
				if(*attrs[j] == cur_a) {
					already_added = true;
					gbw.DefineAsEquivalent(i,j);
					break;
				}
			}
			if(already_added)
				continue;
			aggregation_found = true;
			_int64 max_no_of_distinct = filter.m_index->NoTuples();
			_int64 min_v = MINUS_INF_64;
			_int64 max_v = PLUS_INF_64;
			ushort max_size = cur_a.Type().GetInternalSize();

			if(cur_a.term.vc) {
				max_size = cur_a.term.vc->MaxStringSize();
				if(cur_a.distinct && cur_a.term.vc->IsDistinct() && cur_a.mode != LISTING)
					cur_a.distinct = false;			// "distinct" not needed, as values are distinct anyway
				else if(cur_a.distinct) {
					max_no_of_distinct = cur_a.term.vc->GetApproxDistVals(false);	// no nulls included
					min_v = cur_a.term.vc->RoughMin();
					max_v = cur_a.term.vc->RoughMax();
					if(rccontrol.isOn())
						rccontrol.lock(m_conn.GetThreadID()) << "Adding dist. column, min = " << min_v << ",  max = " << max_v << ",  dist = " << max_no_of_distinct << unlock;
				}
			}
			gbw.AddAggregatedColumn(i, cur_a, max_no_of_distinct, min_v, max_v, max_size);

			//setup pack order for each dim
			if(cur_a.term.vc && cur_a.term.vc->GetDim() != -1)
				oi[cur_a.term.vc->GetDim()].push_back(PackOrderer::OrderingInfo(cur_a.term.vc, cur_a.mode));
		}
	}
	for(uint i = 0; i < attrs.size(); i++) {
		if(utf_columns[i])
			attrs[i]->mode = utf_orig_modes[i];
	}
	if((just_distinct || group_by_found) && filter.m_index->ZeroTuples())
		return 0;

	if(limit != -1 && upper_approx_of_groups > offset + limit)
		upper_approx_of_groups = offset + limit;
	////////////////////////////////////////////////////////////////////////////

	gbw.Initialize(upper_approx_of_groups);
	_int64 cur_output_tuple = 0;
	if((gbw.IsCountOnly() || gbw.IsCountDistinctOnly()) && (offset > 0 || limit == 0))
		return 0;			// one, already omitted row

	// TODO: do all these special cases here in one loop, and left the unresolved aggregations for normal run
	// Special cases: SELECT COUNT(*) FROM ...,    SELECT 1, 2, MIN(a), SUM(b) WHERE false
	bool all_done_in_one_row = false;
	_int64 row = 0;
	if(gbw.IsCountOnly() || filter.m_index->ZeroTuples()) {
		DimensionVector dims(filter.m_index);
		dims.SetAll();
		MIIterator mit(filter.m_index, dims);
		gbw.AddAllGroupingConstants(mit);
		gbw.AddAllAggregatedConstants(mit);
		gbw.FindCurrentRow(row);		// needed to initialize grouping buffer
		gbw.AddAllCountStar(row, mit, filter.m_index->NoTuples());
		all_done_in_one_row = true;
	} // Special case 2, if applicable: SELECT COUNT(DISTINCT col) FROM .....;
	else if(gbw.IsCountDistinctOnly()) {
		_int64 count_distinct = attrs[0]->term.vc->GetExactDistVals();	// multiindex checked inside
		if(count_distinct != NULL_VALUE_64) {
			_int64 row = 0;
			gbw.FindCurrentRow(row);		// needed to initialize grouping buffer
			gbw.PutAggregatedValueForCount(0, row, count_distinct);
			all_done_in_one_row = true;
		}
	}
	if(all_done_in_one_row) {
		for(uint i = 0; i < attrs.size(); i++) {		// left as uninitialized (NULL or 0)
			attrs[i]->page_size = 1;
			attrs[i]->CreateBuffer(1);
		}
		AggregateFillOutput(gbw, row, cur_output_tuple);
		return 1;
	}
	BHASSERT(!filter.m_index->IsCountOnly(), "Aggregation internal error (count(*)).");

	if(limit == -1)
		limit = upper_approx_of_groups;
	cur_output_tuple = MultiDimensionalGroupByScan(gbw, upper_approx_of_groups, limit, offset);

	delete filter.m_index;			// cleanup (i.e. regarded as materialized, one-dimensional)
	filter.m_index = NULL;
	return cur_output_tuple;
}


void TempTable::AggregateFillOutput(GroupByWrapper &gbw, _int64 gt_pos, _int64 &cur_output_tuple)
{
	MEASURE_FET("TempTable::AggregateFillOutput(...)");
	for(uint i = 0; i < attrs.size(); i++) {
		TempTable::Attr *a = attrs[i];								// change to pointer - for speed
		int gt_column = gbw.AttrMapping(i);
		if(gt_column == -1)							// delayed column (e.g. complex exp. on aggregations)
			continue;
		if( ATI::IsStringType(a->TypeName()) ) {
			 (*(AttrBuffer<RCBString> *)a->buffer).Set(cur_output_tuple, gbw.GetValueT( gt_column, gt_pos ) );
		} else {
			_int64 v = gbw.GetValue64( gt_column, gt_pos );
			switch(a->TypeName()) {
				case RC_INT:
				case RC_MEDIUMINT :
					if(v == NULL_VALUE_64)
						(*(AttrBuffer<int> *)a->buffer).Set(cur_output_tuple, NULL_VALUE_32);
					else
						(*(AttrBuffer<int> *)a->buffer).Set(cur_output_tuple, (int)v );
					break;
				case RC_BYTEINT:
					if(v == NULL_VALUE_64)
						(*(AttrBuffer<char> *)a->buffer).Set(cur_output_tuple, (char)NULL_VALUE_C);
					else
						(*(AttrBuffer<char> *)a->buffer).Set(cur_output_tuple, (char)v );
					break;
				case RC_SMALLINT:
					if(v == NULL_VALUE_64)
						(*(AttrBuffer<short> *)a->buffer).Set(cur_output_tuple, (short)NULL_VALUE_SH);
					else
						(*(AttrBuffer<short> *)a->buffer).Set(cur_output_tuple, (short)v );
					break;
				default:											// all 64-bit values
					(*(AttrBuffer<_int64> *)a->buffer).Set(cur_output_tuple, v);
					break;
			}
		}
	}
	/* ClearUsed() instead
	gbw.ClearRow(gt_pos);			// prepare for a new pass
	gbw.InitRow(gt_pos);
	*/
	cur_output_tuple++;
}



bool TempTable::AggregateRough(GroupByWrapper &gbw, MIIterator &mit,
								bool &packrow_done, bool &part_omitted, bool &aggregations_not_changeable,
								_int64 &no_groups, _int64 &uniform_pos, _int64 rows_in_pack, _int64 factor,
								int just_one_aggr)
{
	MEASURE_FET("TempTable::AggregateRough(...)");
	//// Return situations:
	//// a) ignore the packrow and check it in future (next grouping pass) =>  return true;
	//// b) ignore the packrow forever =>  packrow_done = true; return true;
	//// c) ignore just some columns =>    gbw.OmitColumnForPackrow(i); return false;
	////
	//// If just_one_aggr > -1, then it is just for one aggregation column, plus all grouping columns (next disinct pass)
	packrow_done = false;
	part_omitted = false;
	aggregations_not_changeable = false;
	bool grouping_packrow_uniform = true;
	for(int i = 0; i < gbw.NoGroupingAttr(); i++) {		// first test: whether the grouping values are not all uniform
		if(gbw.AddPackIfUniform(i, mit)) {
			gbw.OmitColumnForPackrow(i);				// rowpack may be partially uniform; grouping value already in buffer
			part_omitted = true;
		}
		else
			grouping_packrow_uniform = false;
	}
	uniform_pos = NULL_VALUE_64;					// not changed => no uniform packrow
	if(grouping_packrow_uniform) {					// the whole packrow is defined by uniform values
		bool existed = gbw.FindCurrentRow(uniform_pos);	// the row is fully prepared
		if(uniform_pos != NULL_VALUE_64) {	// Successfully found? The whole packrow goes to one group.
			if( !existed )
				no_groups++;							// successfully added
			packrow_done = true;
			for(int i = gbw.NoGroupingAttr(); i < gbw.NoAttr(); i++)
				if(just_one_aggr == -1 || just_one_aggr == i) {
					if(gbw.AggregatePackInOneGroup(i, mit, uniform_pos, rows_in_pack, factor)) {
						gbw.OmitColumnForPackrow(i);		// the column is marked as already done
						part_omitted = true;
					}
					else
						packrow_done = false;
				}
			if(packrow_done)
				return true;	// pack done, get the next one
		} else
			return true;		// no space to add uniform values in this pass, omit it for now
	}
	assert(!packrow_done);

	// the next step: check if anything may be changed (when no new groups can be added))
	bool all_done = false;

	if(gbw.IsFull()) {
		all_done = false;
		bool any_gr_column_left = false;		// no gr. column left => all uniform, do not ignore them
		for(int i = 0; i < gbw.NoGroupingAttr(); i++)
			if( gbw.ColumnNotOmitted(i) ) {
				any_gr_column_left = true;
				if(!gbw.AttrMayBeUpdatedByPack(i, mit)) {	// grouping values out of scope?
					all_done = true;
					break;
				}
			}
		if(all_done && any_gr_column_left)
			return true;	// packrow is done for now (to be checked in the future), get the next one

		all_done = true;
		if( gbw.NoGroupingAttr() == gbw.NoAttr() )		// no aggregations, i.e. select distinct ... - cannot omit
			all_done = false;
	}

	aggregations_not_changeable = true;		// note that it is for current groups only, the flag will be ignored if any new group is found
	for(int i = gbw.NoGroupingAttr(); i < gbw.NoAttr(); i++)
		if(just_one_aggr == -1 || just_one_aggr == i) {
			if(gbw.ColumnNotOmitted(i)) {
				if(gbw.PackWillNotUpdateAggregation(i, mit)) {
					if(gbw.IsFull()) {
						gbw.OmitColumnForPackrow(i);
						part_omitted = true;
					}
				} else
					aggregations_not_changeable = false;
			}
			if(gbw.ColumnNotOmitted(i))
				all_done = false;	// i.e. there is something to be done in case of existing grouping values
		}

	// note that we may need to open the packrow to localize groups for future scans, except:
	if( all_done && gbw.IsFull() && gbw.NoMoreGroups()) {	// no aggregation may be changed by this packrow?
		packrow_done = true;
		return true;	// packrow is excluded from the search, get the next one
	}
	return all_done;
}

