/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

/////////////////////////////////////////////////////////////////////////////////////////////////////
// This is a part of TempTable implementation concerned with the query execution low-level mechanisms
/////////////////////////////////////////////////////////////////////////////////////////////////////
#include <boost/shared_ptr.hpp>
#include <boost/assign.hpp>

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "TempTable.h"
#include "PackGuardian.h"
#include "system/TextUtils.h"
#include "types/DataConverter.h"
#include "system/RCSystem.h"
#include "system/ConnectionInfo.h"
#include "shared/common/DataFormat.h"
#include "shared/exporter/DataExporter.h"
#include "types/ValueParserForText.h"
#include "system/IOParameters.h"
#include "vc/VirtualColumn.h"
#include "SorterWrapper.h"
#include "tools.h"
#include "common/bhassert.h"

using namespace std;
using namespace boost;
using namespace boost::assign;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool TempTable::OrderByAndMaterialize(vector<SortDescriptor> &ord, _uint64 limit, _uint64 offset)	// Sort MultiIndex using some (existing) attributes in some tables
{	// "limit=10; offset=20" means that the first 10 positions of sorted table will contain objects 21...30.
	MEASURE_FET("TempTable::OrderBy(...)");
	if((int)ord.size() == 0 || filter.m_index->NoTuples() < 2)
		return false;

	// Prepare sorter
	SorterWrapper sorted_table(*(filter.m_index), limit + offset);
	int sort_order = 0;
	for(int j = 0; j < attrs.size(); j++) 
		if(attrs[j]->alias != NULL) {
			VirtualColumn* vc = attrs[j]->term.vc;
			assert(vc);
			sort_order = 0;
			for(int i = 0; i < ord.size(); i++)
				if(ord[i].vc == vc) {
					sort_order = (ord[i].dir == 0 ? (i + 1) : -(i + 1));
					ord[i].vc = NULL;		// annotate this entry as already added
				}
			sorted_table.AddSortedColumn(vc, sort_order, true);
	}
	for(int i = 0; i < ord.size(); i++) {		// find all columns not added yet (i.e. not visible in output)
		if(ord[i].vc != NULL) {
			sort_order = (ord[i].dir == 0 ? (i + 1) : -(i + 1));
			sorted_table.AddSortedColumn(ord[i].vc, sort_order, false);
		}
	}
	ord.clear();
	sorted_table.InitSorter();

	// Put data
	MIIterator it(filter.m_index);
	_int64 local_row = 0;
	bool continue_now = true;
	while(it.IsValid() && continue_now) {
		if(m_conn.killed()) throw KilledRCException();
		continue_now = sorted_table.PutValues(it);		// return false if a limit is already reached (min. values only)
		++it;

		local_row++;
		if(local_row % 100000000 == 0)
			rccontrol.lock(m_conn.GetThreadID()) << "Preparing values to sort (" << int(local_row / double(filter.m_index->NoTuples()) * 100) << "% done)." << unlock;
	}

	// Create output
	no_obj = limit;
	for(uint i = 0; i < NoAttrs(); i++)
		if(attrs[i]->alias != NULL)
			attrs[i]->CreateBuffer(no_obj, &m_conn);

	_int64 global_row = 0;
	local_row = 0;
	RCBString val_s;
	_int64 val64;
	bool null_value;
	while(sorted_table.FetchNextRow()) {
		int col = 0;
		if(global_row >= offset) {
			if(m_conn.killed()) throw KilledRCException();
			for(attrs_t::iterator attr = attrs.begin(), end = attrs.end(); attr != end; ++attr) 
				if((*attr)->alias != NULL) {
					switch((*attr)->TypeName()) {
						case RC_STRING:
						case RC_VARCHAR:
						case RC_BIN:
						case RC_BYTE:
						case RC_VARBYTE:
							val_s = sorted_table.GetValueT(col);
							(*attr)->SetValueString(local_row, val_s);
							break;
						default:
							val64 = sorted_table.GetValue64(col, null_value);	// works also for constants
							if(null_value)
								(*attr)->SetValueInt64(local_row, NULL_VALUE_64);
							else
								(*attr)->SetValueInt64(local_row, val64);
					}
					col++;
			}
			local_row++;
			if(local_row % 100000000 == 0)
				rccontrol.lock(m_conn.GetThreadID()) << "Retrieving sorted rows (" << int(local_row / double(limit - offset) * 100) << "% done)." << unlock;
		}
		global_row++;
		if(global_row >= limit + offset)
			break;
	}
	rccontrol.lock(m_conn.GetThreadID()) << "Sorted rows retrieved." << unlock;
	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void TempTable::FillMaterializedBuffers(_uint64 local_limit, _uint64 local_offset) 
{
	_int64 val_d = 0;
	RCBString vals;
	int size = 0;
	uchar *p = NULL;
	JustATable * curr_table = NULL;
	no_obj = local_limit;
	int page_size = CalculatePageSize();
	if(!filter.m_index->ZeroTuples())
	{
		//////// Column statistics ////////////////////////
		if(m_conn.DisplayAttrStats()) {
			for(uint j = 0; j < NoAttrs(); j++) {
				if(attrs[j]->term.vc)
					attrs[j]->term.vc->DisplayAttrStats();
			}
			m_conn.SetDisplayAttrStats(false); // already displayed
		}
		///////////////////////////////////////////////////
		/* Virtual Column SELECT * FROM t; */
		for(uint i = 0; i < NoAttrs(); i++)
			attrs[i]->CreateBuffer(no_obj, &m_conn);

		MIIterator it(filter.m_index);
		bool hasIntrestingColumns(false);
		for(attrs_t::const_iterator attr = attrs.begin(), end = attrs.end(); attr != end; ++attr) { /* materialize dependent tables */
			if(((*attr)->mode == LISTING) && (*attr)->term.vc && (*attr)->alias) { // constant value, the buffer is already filled in
				hasIntrestingColumns = true;
				break;
			}
		}
		_int64 row = 0;
		while(hasIntrestingColumns && it.IsValid() && row < no_obj + local_offset) { /* go thru all rows */
			bool outerIteratorUpdated = false;
			MIIterator pageStart(it);
			_int64 start_row = row;
			_int64 page_end = (((row - local_offset) / page_size) + 1) * page_size + local_offset;
			// where the current TempTable buffer ends, in terms of multiindex rows (integer division)
			if(page_end > no_obj + local_offset)
				page_end = no_obj + local_offset;
			for(attrs_t::iterator attr = attrs.begin(), end = attrs.end(); attr != end; ++attr) { /* fill buffers in attrs */
				VirtualColumn* vc = (*attr)->term.vc;
				if(((*attr)->mode == LISTING) && vc && (*attr)->alias) { // constant value, the buffer is already filled in
					MIIterator i(pageStart);
					int locdim = vc->GetDim();
					_int64 r = start_row;
					//						_int64 reasonable_portion = r + 6500000;	// merge smaller packrows into portions
					_int64 output_row = 0;		/* actual row number in output buffers */
					do { /* go thru all rows (fitting into page size) in given column in given packrow */
						if(i.PackrowStarted())
							vc->LockSourcePacks(i);
						output_row = r - local_offset;
						if(output_row >= 0)
							switch((*attr)->TypeName()) {
								case RC_STRING:
								case RC_VARCHAR:
									vc->GetValueString(vals, i, false);
									(*attr)->SetValueString(output_row, vals);
									break;
								case RC_BIN:
								case RC_BYTE:
								case RC_VARBYTE:
									if( !vc->IsNull(i) )
										vc->GetValueString(vals, i, false);
									else
										vals = RCBString();
									(*attr)->SetValueString(output_row, vals );
									break;
								default:
									(*attr)->SetValueInt64(output_row, vc->GetValueInt64(i));
						}
						++i; /* local multi-index-iterator */
						++r; /* local output row index */
					} while(i.IsValid()
						//									&& (!i.PackrowStarted() || r < reasonable_portion)
						&& r < page_end);
					if(!outerIteratorUpdated) {
						it.swap(i); /* update global iterator - once */
						row = r;
						outerIteratorUpdated = true;
					}
				}
			}
		}
	}

}

_uint64 TempTable::Export(const IOParameters& iop, fields_t const& fields)
{
	const char *out_path = iop.Path();
	EDF edf = iop.GetEDF();
	uchar delimiter = iop.Delimiter();
	uchar string_qualifier = iop.StringQualifier();

	string nulls_str = iop.NullsStr();

	MEASURE_FET("TempTable::Export(...)");
	bool non_listing = false;

	for(uint i = 0; i < NoAttrs(); i++)
		if (attrs[i]->mode != LISTING) {
			non_listing = true;
			break;
		}

	if((!materialized && (group_by || non_listing || mode.distinct || order_by.size() > 0)) ||
		(materialized && (order_by.size() > 0 || mode.top) && no_obj))
		Materialize();

	shared_ptr<LargeBuffer> rcbuffer = shared_ptr<LargeBuffer>(new LargeBuffer());

	if(!rcbuffer->BufOpen(iop, OVERWRITE))
		throw FileRCException("Unable to open file or named pipe.");

	_uint64 nor = 0;

	try
	{
		auto_ptr<DataExporter> rcde = DataFormat::GetDataFormat(edf)->CreateDataExporter(delimiter, string_qualifier, nulls_str);
		rcde->Init(rcbuffer, GetATIs(), fields);

		if(materialized)
			nor = ExportFromMaterialized(*rcde, fields);
		else
			nor = ExportFromNonMaterialized(*rcde, fields);

	}
	catch (...)
	{
		rcbuffer->FlushAndClose();
		throw;
	}
	rcbuffer->FlushAndClose();
	return nor;
}

std::vector<ATI> TempTable::GetATIs()
{
	vector<ATI> deas;
	for(uint i = 0; i < NoAttrs(); i++) {
		if(!IsDisplayAttr(i))
			continue;
		deas += ATI(attrs[i]->TypeName(), !attrs[i]->Type().GetNullsMode(), attrs[i]->Type().GetPrecision(), attrs[i]->Type().GetScale(), false, false, attrs[i]->Type().GetCollation());
	}
	return deas;
}

_uint64 TempTable::ExportFromMaterialized(DataExporter& rcde, fields_t const& fields)
{
	_uint64 nor = 0;
	TempTable::RecordIterator iter = this->begin(&m_conn, JustATable::RcTableLockPolicy::attr_uses_t() );
	for(; !(iter == this->end(&m_conn)); ++ iter )
	{
		fields_t::const_iterator it = fields.begin();
		for(uint o = 0; o < NoDisplaybleAttrs(); o++, ++ it )
		{
			BHASSERT( (it != fields.end()), "mysql fileds type array incomplete" );
			RCItem item = (*iter)[o];

			if(item.IsNull())
				rcde.PutNull();
			else if(ATI::IsTxtType(item.Type()))
				rcde.PutText(RCValueObject(item).ToRCString());
			else if(ATI::IsBinType(item.Type()))
				rcde.PutBin(RCValueObject(item).ToRCString());
			else if(ATI::IsNumericType(item.Type())) {
				if(item.Type() == RC_BYTEINT)
					rcde.PutNumeric(*(char*)item.Buf());
				else if(item.Type() == RC_SMALLINT)
					rcde.PutNumeric(*(short*)item.Buf());
				else if(item.Type() == RC_INT || item.Type() == RC_MEDIUMINT)
					rcde.PutNumeric(*(int*)item.Buf());
				else
					rcde.PutNumeric(*(_int64*)item.Buf());
			} else if(ATI::IsDateTimeType(item.Type()))
				rcde.PutDateTime(*(_int64*)item.Buf());
		}
		rcde.PutRowEnd();
		nor++;
	}
	return nor;
}

_uint64 TempTable::ExportFromNonMaterialized(DataExporter& rcde, fields_t const& fields)
{
	_uint64 offset = 0;
	if(mode.top) {
		offset = mode.param1 >= 0 ? mode.param1 : 0;
		if(offset > no_obj)
			offset = no_obj;
		if(offset + mode.param2 > no_obj)
			mode.param2 = no_obj - offset;
		if(mode.param2 < 0)
			mode.param2 = 0;
		no_obj = mode.param2;
	}
	if(!no_obj)
		return 0;
	MIIterator mit(filter.m_index);

	// skip 'offset' rows
	for(_uint64 i = 0; i < offset; i++) {
		if(!mit.IsValid())
			BHERROR("In TempTable::ExportNonAggregated(), MultiIndex::NoTuples() gives incorrect number");
		++mit;
	}
	bool first_row = true;
	_uint64 obj = 0;
	while(mit.IsValid() && obj < no_obj) {
		fields_t::const_iterator it = fields.begin();
		for(uint j = 0; j < NoAttrs(); j++, ++ it ) {
			BHASSERT( (it != fields.end()), "mysql fileds type array incomplete" );
			if(IsDisplayAttr(j)) {
				VirtualColumn *vc = attrs[j]->term.vc;
				if(first_row || mit.PackrowStarted()) {
					vc->LockSourcePacks(mit);
				}
				AttributeType t( vc->TypeName() );
				if(vc->IsNull(mit))
					rcde.PutNull();
				else if(ATI::IsTxtType(t)) {
					RCBString val;
					vc->GetValueString(val, mit);
					if(*it == MYSQL_TYPE_DATE) {
						RCDateTime dt;
						ValueParserForText::ParseDateTime(val,dt,RC_DATE);
						rcde.PutDateTime(dt.GetInt64());
					} else if ((*it == MYSQL_TYPE_DATETIME) || (*it == MYSQL_TYPE_TIMESTAMP)) {
						RCDateTime dt;
						ValueParserForText::ParseDateTime(val,dt,RC_DATETIME);
						rcde.PutDateTime(dt.GetInt64());
					} else {
						rcde.PutText(val);
					}
				} else if(ATI::IsBinType(t)) {
					RCBString val;
					vc->GetValueString(val, mit);
					rcde.PutBin(val);
				} else if(ATI::IsDateTimeType(t))
					rcde.PutDateTime(vc->GetValueInt64(mit));
				else if(vc->Type().IsNumeric())
					rcde.PutNumeric(vc->GetValueInt64(mit));
				else BHERROR("Incorrect attribute type when exporting");
			}
		}
		rcde.PutRowEnd();
		++mit;
		obj++;
		first_row = false;
	}
	return no_obj;
}

void TempTable::VerifyAttrsSizes()	// verifies attr[i].field_size basing on the current multiindex contents
{
	for(uint i = 0; i < attrs.size(); i++)
		if(ATI::IsStringType(attrs[i]->TypeName()))
			attrs[i]->OverrideStringSize(attrs[i]->term.vc->MaxStringSize());
}

