/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef __TOOLS_H
#define __TOOLS_H

#include <set>
#include <cstdio>

#include "system/RCSystem.h"
#include "system/ib_system.h"

/* Release (close) a FILE* on return from a function (i.e. on destruction of this object) */
class CloseFILE
{
	FILE* f;
public:
	CloseFILE(FILE* _f)		{ f = _f; }
	~CloseFILE()				{ fclose(f); }
};

//T - class type
//A - parameter type
template<class T, typename A>
class MemberFunctionExecutor
{
public:
	MemberFunctionExecutor(T& obj, void (T::*on_start)(int), void (T::*on_end)(int), A argument)
		: argument(argument), obj(&obj)
	{
		this->on_start = on_start;
		this->on_end = on_end;
		((this->obj)->*on_start)(argument);
	}

	~MemberFunctionExecutor()
	{
		(obj->*on_end)(argument);
	}

private:
	A argument;
	T* obj;
	void (T::*on_start)(A);
	void (T::*on_end)(A);
};

template<typename T>
class TemporalValueReplacement
{
public:
	TemporalValueReplacement(T& to_replace, T temp_value)
		: orig_value(to_replace), orig_value_ptr(&to_replace)
	{
		*orig_value_ptr = temp_value;
	}

	~TemporalValueReplacement()
	{
		*orig_value_ptr = orig_value;
	}
private:
	T orig_value;
	T* orig_value_ptr;
};


#include <boost/function/function0.hpp>
class FunctionExecutor
{
typedef boost::function0<void> F;
public:
    FunctionExecutor(F call_in_constructor, F call_in_deconstructor)
        : call_in_constructor(call_in_constructor), call_in_deconstructor(call_in_deconstructor)
    {
        call_in_constructor();
    }

    virtual ~FunctionExecutor()
    {
        call_in_deconstructor();
    }

private:
	boost::function0<void> call_in_constructor;
	boost::function0<void> call_in_deconstructor;
};

#ifdef FUNCTIONS_EXECUTION_TIMES

class FunctionsExecutionTimes
{
	struct Entry {
		_uint64		call;	// no. of calls
		_uint64		time;	// total time of calls
		Entry(_uint64 t)	{ call = 1; time = t; }
		Entry()				{}
	};

public:
	FunctionsExecutionTimes()
	{
		times = new std::map<std::string, Entry>;
		flush = false;
        /******************************************
		pthread_mutexattr_t mattr;
		pthread_mutexattr_init(&mattr);
		pthread_mutexattr_settype(&mattr, PTHREAD_MUTEX_RECURSIVE);
		pthread_mutex_init(&cs, &mattr);
        ******************************************/
		InstallFlushSignal();
	}
	~FunctionsExecutionTimes()
	{
		times->clear();
		delete times;
		//pthread_mutex_destroy(&cs);
	}

	void Add(std::string function_name, _uint64 time)
	{
		Lock();

		if(flush) {
			Print();
			Clear();
			flush = false;
		}

		if(times->find(function_name) != times->end())
		{
			Entry& entry = (*times)[function_name];
			entry.time += time ;
			entry.call ++;
		}
		else
		{
			//times.insert(function_name);
			Entry entry(time);
			times->insert(make_pair(function_name, entry));
		}
		UnLock();
	}

	void PrintToRcdev()
	{
		Lock();
		Print();
		UnLock();
	}

	// order flushing; it will be executed on the next Add()
	void Flush()					{ flush = true; }

	static void FlushFET(int signum);	// flush global object 'fet'
	static void InstallFlushSignal();

private:
	void Lock()		{ fet_mutex.Lock(); /*pthread_mutex_lock(&cs);*/ }
	void UnLock()	{ fet_mutex.Unlock(); /*pthread_mutex_unlock(&cs);*/ }

	void Clear()	{ times->clear(); }

	void Print()
	{
		rcdev << lock << "***** Function Execution Times **********************************************************" << unlock;
		std::map<std::string, Entry>::iterator iter = times->begin();
		char str[512] = {0};
		for(;iter != times->end(); iter++) {
			sprintf(str, "%-60s \t%6llu \t%8.2fs", iter->first.c_str(),
					iter->second.call, iter->second.time/1000000.0);
			rcdev << lock << str << unlock;
		}
		rcdev << lock << "Number of distinct Data Packs loads: " << (uint)count_distinct_dp_loads.size() << unlock;
		rcdev << lock << "Number of distinct Data Packs decompressions: " << (uint)count_distinct_dp_decompressions.size() << unlock;
		rcdev << lock << "Size of DP read from disk: " << ((double)NoBytesReadByDPs / 1024 / 1024) << "MB" << unlock;
		rcdev << lock << "Size of uncompressed DPs: " << ((double)SizeOfUncompressedDP / 1024 / 1024) << "MB" << unlock;
		rcdev << lock << "******************************************************************************************" << unlock;
	}

	std::map<std::string, Entry>* times;
    IBMutex fet_mutex;
	//pthread_mutex_t	cs;
	bool 			flush;		// should we flush on the next Add()?
};

class FETOperator
{
public:
	FETOperator(const std::string& identyfier)
		:id(identyfier)
	{
		gettimeofday(&start_time, NULL);

	}

	~FETOperator()
	{
		struct timeval t2;
		_uint64 sec, usec;
		sec = usec = 0;
		gettimeofday(&t2, NULL);
        sec += (t2.tv_sec - start_time.tv_sec);
        if (t2.tv_usec < start_time.tv_usec) {
                sec--;
                t2.tv_usec += 1000000l;
        };

        usec += (t2.tv_usec - start_time.tv_usec);
        if (usec >= 1000000l)
        {
                usec -=  1000000l;
                sec++;
        };
        usec += sec * 1000000;
		fet->Add(id, usec);
	}

private:
	std::string id;
	struct timeval start_time, t2;
};


inline void FunctionsExecutionTimes::FlushFET(int signum)
{
	fet->Flush();
}

inline void FunctionsExecutionTimes::InstallFlushSignal()
{
	// install FlushFET as a signal handler for SIGUSR1
	void* ret = (void*)signal(16, FunctionsExecutionTimes::FlushFET);
}

//#ifdef FUNCTIONS_EXECUTION_TIMES
//#define _FET(name,str)		FETOperator name(str)
//#else
//#define _FET(name,str)
//#endif

class DPID
{
public:
	DPID(ushort tid, ushort cid, uint did)
		:	tid(tid), cid(cid), did(did)
	{
	}

	bool operator<(const DPID& dpid) const
	{
		return (
				(tid < dpid.tid) ||
				(tid == dpid.tid && cid < dpid.cid) ||
				(tid == dpid.tid && cid == dpid.cid && did < dpid.did)
				);
	}
private:
	ushort	tid;
	ushort	cid;
	uint	did;
};

typedef std::set<DPID> LoadedDataPackCounter;

#define MEASURE_FET(X) FETOperator feto(X)

#else
	#define MEASURE_FET(X)
#endif

//#define PROFILE_LOCK_WAITING
#ifdef PROFILE_LOCK_WAITING

class LockProfiler
{
	public:
		LockProfiler() {
			//pthread_mutex_init(&m,NULL);
		}
		~LockProfiler() {
			map<string,_uint64> :: iterator it;
			for(it = lock_waiting.begin(); it!= lock_waiting.end(); it++) {
				cerr << it->first << "\t" << it->second << endl;
			}
			lock_waiting.clear();
		}
		void Record(string where, _uint64 usec) {
			//pthread_mutex_lock(&m);
            IBGuard guard(lock_profiler_mutex);
			lock_waiting[where]+= usec;
			//pthread_mutex_unlock(&m);
		}

	private:
        IBMutex lock_profiler_mutex;
		//pthread_mutex_t m;
		map<string,_uint64> lock_waiting;


};

extern LockProfiler lock_profiler;

#define STRINGIFY(x) #x
#define TOSTRING(x) STRINGIFY(x)

#define pthread_mutex_lock(m) \
{ \
struct timeval start_time, t2;	\
_uint64 sec, usec; 		\
gettimeofday(&start_time, NULL); 		\
pthread_mutex_lock(m);	\
gettimeofday(&t2, NULL); \
sec = usec = 0;			\
sec = (t2.tv_sec - start_time.tv_sec); \
if (t2.tv_usec < start_time.tv_usec) {  \
        sec--; \
        t2.tv_usec += 1000000l; \
}; \
usec = (t2.tv_usec - start_time.tv_usec); \
if (usec >= 1000000l) \
{ \
        usec -=  1000000l; \
        sec++; \
}; \
usec += sec * 1000000; \
lock_profiler.Record(string(__FILE__) + TOSTRING(__LINE__) , usec);\
}


#define pthread_cond_wait(c,m) \
({ \
int r; \
struct timeval start_time, t2;	\
_uint64 sec, usec; 		\
gettimeofday(&start_time, NULL); 		\
r = pthread_cond_wait(c,m);	\
gettimeofday(&t2, NULL); \
sec = usec = 0;			\
sec = (t2.tv_sec - start_time.tv_sec); \
if (t2.tv_usec < start_time.tv_usec) {  \
        sec--; \
        t2.tv_usec += 1000000l; \
}; \
usec = (t2.tv_usec - start_time.tv_usec); \
if (usec >= 1000000l) \
{ \
        usec -=  1000000l; \
        sec++; \
}; \
usec += sec * 1000000; \
lock_profiler.Record(string(__FILE__) + TOSTRING(__LINE__) , usec);\
r;\
})

#endif


/* Use objects of this class to enter critical section and automatically leave it when returning from a function.
   To do this, declare a local object: "(...) LockCS _cs(&critical_section); (...)".
   Its destructor will be invoked automatically. */
/******
 * replaced by IBGuard
class LockCS
{
	pthread_mutex_t *cs;
public:
	LockCS(pthread_mutex_t* cs_)		{ cs = cs_; pthread_mutex_lock(cs); }
	~LockCS()							{ pthread_mutex_unlock(cs); }
};
********/

template<typename SourceType>
void store_bytes(char*& destination, const SourceType& source)
{
	memcpy(destination, (unsigned char*)&source, sizeof(SourceType));
	destination += sizeof(SourceType);
}

template<typename DestinationType>
void unstore_bytes(DestinationType& destination, char*& source)
{
	memcpy((unsigned char*)&destination, source, sizeof(DestinationType));
	source += sizeof(DestinationType);
}

#endif
