/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include "shared/core/RCEngine.h"
#include "QueryTranslator.h"

using namespace std;

namespace na {

/*
**********  Class: QueryTranslator  **********
*/

QueryTranslator::Result QueryTranslator::TranslateQuery(SELECT_LEX* selects_list, SELECT_LEX* last_distinct, BHQuery* query)
{
	if(last_distinct || selects_list->next_select()) return FAILURE;		// UNION is present
	return TranslateSelect(selects_list, query);
}

QueryTranslator::Result QueryTranslator::TranslateSelect(SELECT_LEX* select, BHQuery* query)
{
	TABLE_LIST *tables = select->leaf_tables, *tab = tables;
	if(tables->next_leaf) return FAILURE;			// more than 1 table is present
	if(tab->derived) return FAILURE;				// FROM subquery ('derived' points to the SELECT_LEX_UNIT of subquery)

	// FROM clause

	TableDescriptor* from = new TableDescriptor;
	query->SetFrom(from);							// from now on, 'query' is responsible for deallocation of 'from'

	ASSERT_MYSQL_STRING(tab->table->s->path);
	string path = tab->table->s->path.str;
	from->bhtable = rceng->GetTable(path, tab->table->s);
	if (!from->bhtable) {
		BHERROR("'from->bhtable' should not be null");
		return FAILURE;
	}
	from->mytable = tab->table;

	// SELECT clause

	List<Item>* fields = &select->join->fields_list;
	List_iterator_fast<Item> li(*fields);
	Item* item;
	BHItem* bhitem;
	Expression* expr;

	if (!fields) {
		BHERROR("'fields' should not be null");
		return FAILURE;
	}
	current_query = query;

	while ((item = li++))
		if(TranslateItem(item, bhitem) == SUCCESS) {
			expr = dynamic_cast<Expression*>(bhitem);
			if (expr)
				query->AddSelect(expr);
			else {
				BHERROR("bad cast");
				return FAILURE;
			}
		}
		else return FAILURE;

	return SUCCESS;
}

QueryTranslator::Result QueryTranslator::TranslateItem(Item* item, BHItem*& bhitem)
{
	if(!item) return FAILURE;

	// item_int_with_ref?
	Item_int_with_ref * int_ref = dynamic_cast<Item_int_with_ref*>(item);
	if (int_ref)
		return TranslateItem(int_ref->real_item(), bhitem);

	try {
		switch(item->type())
		{
		case Item::FIELD_ITEM:		return ItemField(item,bhitem);								// field name
		case Item::INT_ITEM:		return ItemInt(item,bhitem);								// integer constant
		case Item::FUNC_ITEM:		return ItemFunc(item,bhitem);								// operators, ...
		case Item::REF_ITEM:		return TranslateItem(*((Item_ref*)item)->ref, bhitem);		// reference - skip it
		default:					return FAILURE;
		}
	} catch(...) { return FAILURE; }
}

QueryTranslator::Result QueryTranslator::ItemField(Item* item, BHItem*& bhitem)
{
	Item_field* it = (Item_field*) item;
	// hm, I'm not sure which field is good for us; they seem to be always equal
	BHASSERT(it->field == it->result_field, "is it always true? 'it->field == it->result_field'");
	if (!current_query) {
		BHERROR("'current_query' should not be null");
		return FAILURE;
	}

	TableDescriptor* tab = current_query->FindTable(it->field->table);
	if (!tab) {
		BHERROR("'tab' should not be null");
		return FAILURE;
	}
	TABLE* mytab = tab->mytable;
	const char* field_name = it->field->field_name;

	int col = it->field->field_index;		// maybe it's sometimes wrong?
	//int col = 0;
	//for (; mytab->field[col]; col++)
	//	if (mytab->field[col]->field_name == field_name)
	//		break;
	//if (!mytab->field[col]) return FAILURE;

	bhitem = new ExprVar(tab,col);
	return SUCCESS;
}

QueryTranslator::Result QueryTranslator::ItemInt(Item* item, BHItem*& bhitem)
{
	if(item->unsigned_flag) return FAILURE;		// big unsigned integer - we can't hold it
	bhitem = new ExprConst(new VInteger(item->val_int()));
	return SUCCESS;
}

QueryTranslator::Result QueryTranslator::ItemFunc(Item* item, BHItem*& bhitem)
{
	// 'item' is Item_func_neg (unary minus)?
	Item_func_neg *item_neg = dynamic_cast<Item_func_neg*>(item);
	if ((item_neg == 0) || (item_neg->arg_count != 1))	   // not a minus or the argument count too large
		return FAILURE;

	BHItem* arg = NULL;
	if((TranslateItem(item_neg->arguments()[0], arg) == FAILURE) ||	!arg->IsExpression())
		return FAILURE;

	bhitem = new ExprArithNeg((Expression*)arg);
	return SUCCESS;
}

}	// namespace na

