/* Copyright (C) 2000 MySQL AB, portions Copyright (C) 2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <iostream>

#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <boost/scoped_ptr.hpp>

#define MYSQL_SERVER  1

#ifndef __BH_COMMUNITY__
#	include "enterprise/local.h"
#else /* not __BH_COMMUNITY__ */
#	include "community/local.h"
#endif /* __BH_COMMUNITY__ */
#include "BrighthouseHandler.h"
#include "mysql/plugin.h"

#include "system/RCException.h"
#include "core/compilation_tools.h"
#include "common/bhassert.h"
#include "core/tools.h"
#include "core/TempTable.h"
#include "vc/VirtualColumn.h"

using namespace std;

handler *rcbase_create_handler(handlerton *hton, TABLE_SHARE *table, MEM_ROOT *mem_root);
int rcbase_panic_func(handlerton *hton, enum ha_panic_function flag);
int rcbase_close_connection(handlerton *hton, THD* thd);
int rcbase_commit(handlerton *hton, THD *thd, mysql_bool all);
int rcbase_rollback(handlerton *hton, THD *thd, mysql_bool all);
int rcbase_init_func(void *p);

handlerton *rcbase_hton;

/* Variables for rcbase share methods */
HASH rcbase_open_tables; // Hash used to track open tables
//pthread_mutex_t rcbase_mutex; // This is the mutex we use to init the hash
IBMutex rcbase_mutex;

/*
 Function we use in the creation of our hash to get key.
 */
uchar* rcbase_get_key(RCBASE_SHARE *share, size_t *length, my_bool not_used __attribute__((unused)))
{
	*length=share->table_name_length;
	return (uchar*) share->table_name;
}

/*
 If frm_error() is called then we will use this to to find out what file extentions
 exist for the storage engine. This is also used by the default rename_table and
 delete_table method in handler.cc.
 */
const char *ha_rcbase_exts[] = { ".bht", NullS };

my_bool infobright_bootstrap = 0;

#ifdef __WIN__
#define __PRETTY_FUNCTION__   __FUNCTION__
#endif

static int rcbase_done_func(void *p)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	int error= 0;

	if(rcbase_open_tables.records)
		error= 1;
	hash_free(&rcbase_open_tables);
	//pthread_mutex_destroy(&rcbase_mutex);

	DBUG_RETURN(0);
}

int rcbase_panic_func(handlerton *hton, enum ha_panic_function flag)
{
	if(flag==HA_PANIC_CLOSE)
		delete rceng;
	return 0;
}

int rcbase_close_connection(handlerton *hton, THD* thd)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	int ret = 1;
	try {
		rcbase_rollback(hton, thd, true);
		table_lock_manager.ReleaseLocks(*thd);
		rceng->ClearThreadParams(*thd);
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(ret);
}

int rcbase_commit(handlerton *hton, THD *thd, mysql_bool all)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	int ret = 1;
	try {
		// GA: double check that thd->no_errors is the same as query_error
		if(thd->no_errors != 0 || thd->killed)
			rceng->Rollback(thd, all);
		else
			rceng->Commit(thd, all);
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(ret);
}

int rcbase_rollback(handlerton *hton, THD *thd, mysql_bool all)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	int ret = 1;
	try {
		rceng->Rollback(thd, all);
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(ret);
}

/*
 Example of simple lock controls. The "share" it creates is structure we will
 pass to each rcbase handler. Do you have to have one of these? Well, you have
 pieces that are used for locking, and they are needed to function.
 */
static RCBASE_SHARE *get_share(const char *table_name, TABLE *table)
{
	RCBASE_SHARE *share;
	uint length;
	char *tmp_name;

    rcbase_mutex.Lock();
	length=(uint) strlen(table_name);

	if(!(share=(RCBASE_SHARE*) hash_search(&rcbase_open_tables, (uchar*) table_name, length))) {
		if(!(share=(RCBASE_SHARE *) my_multi_malloc(MYF(MY_WME | MY_ZEROFILL), &share, sizeof(*share), &tmp_name,
				length+1, NullS))) {
            rcbase_mutex.Unlock();
			return NULL;
		}

		share->use_count=0;
		share->table_name_length=length;
		share->table_name=tmp_name;
		strmov(share->table_name, table_name);
		if(my_hash_insert(&rcbase_open_tables, (uchar*) share))
			goto error;
		thr_lock_init(&share->lock);
	}
	share->use_count++;
    rcbase_mutex.Unlock();

	return share;

error:
	my_free((uchar*) share, MYF(0));
	return NULL;
}

/*
 Free lock controls. We call this whenever we close a table. If the table had
 the last reference to the share then we free memory associated with it.
 */
static int free_share(RCBASE_SHARE *share)
{
	//pthread_mutex_lock(&rcbase_mutex);
    rcbase_mutex.Lock();
	if(!--share->use_count) {
		hash_delete(&rcbase_open_tables, (uchar*) share);
		thr_lock_delete(&share->lock);
		//pthread_mutex_destroy(&share->mutex);
		my_free((uchar*) share, MYF(0));
	}
    rcbase_mutex.Unlock();
	//pthread_mutex_unlock(&rcbase_mutex);

	return 0;
}

handler* rcbase_create_handler(handlerton *hton, TABLE_SHARE *table, MEM_ROOT *mem_root)
{
	handler* ret = 0;
	try	{
		ret = new (mem_root) BrighthouseHandler(hton, table);
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}
	return ret;
}

my_bool rcbase_query_caching_of_table_permitted(THD* thd, char* full_name, uint full_name_len, ulonglong *unused)
{
	if (!(thd->options & (OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN)))
		return ((my_bool)TRUE);
	return ((my_bool)FALSE);
}

BrighthouseHandler::BrighthouseHandler(handlerton *hton, TABLE_SHARE *table_arg)
	: handler(hton, table_arg), m_conn(NULL), m_table_path(NULL),
		_table( NULL ), _tableIter(), _tableIterEnd(), m_query(NULL), m_cq(NULL),
		m_rctable(NULL)
{
	err_msg[0] = '\0';
	ref_length = sizeof(uint64);
}

BrighthouseHandler::~BrighthouseHandler()
{
	delete [] m_table_path;
}

const char **BrighthouseHandler::bas_ext() const
{
	return ha_rcbase_exts;
}

namespace {

JustATable::RcTableLockPolicy::attr_uses_t get_attr_uses( st_table* table ) {
	int col_id = 0;
	JustATable::RcTableLockPolicy::attr_uses_t attrUses;
	for ( Field **field = table->field; *field; ++ field, ++ col_id ) {
		if ( bitmap_is_set( table->read_set, col_id ) ) { //this time no '+1' is needed
			attrUses.push_back( col_id );
		}
	}
	return ( attrUses );
}

}

/*
 Used for opening tables. The name will be the name of the file.
 A table is opened when it needs to be opened. For instance
 when a request comes in for a select on the table (tables are not
 open and closed for each request, they are cached).

 Called from handler.cc by handler::ha_open(). The server opens all tables by
 calling ha_open() which then calls the handler specific open().
 */
int BrighthouseHandler::open(const char *name, int mode, uint test_if_locked)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	int ret = 1;
	rceng->GetThreadParams(*table->in_use);
	table_lock_manager.RegisterTable(name);

	if(!(share = get_share(name, table)))
		DBUG_RETURN(ret);
	thr_lock_data_init(&share->lock, &m_lock, NULL);

	m_table_path = new char[strlen(name) + 1];
	strcpy(m_table_path, name);
	try {
		rceng->OpenTable(name, table_share);
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	info(HA_STATUS_CONST);

	DBUG_RETURN(ret);
}

/*
 Closes a table. We call the free_share() function to free any resources
 that we have allocated in the "shared" structure.

 Called from sql_base.cc, sql_select.cc, and table.cc.
 In sql_select.cc it is only used to close up temporary tables or during
 the process where a temporary table is converted over to being a
 myisam table.
 For sql_base.cc look at close_data_tables().
 */
int BrighthouseHandler::close(void)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(free_share(share));
}

/*
 Positions an index cursor to the index specified in the handle. Fetches the
 row if available. If the key value is null, begin at the first key of the
 index.
 */
int BrighthouseHandler::index_read(uchar * buf, const uchar * key, uint key_len __attribute__((unused)), enum ha_rkey_function find_flag
                           __attribute__((unused)))
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

/*
 Used to read forward through the index.
 */
int BrighthouseHandler::index_next(uchar * buf)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

/*
 Used to read backwards through the index.
 */
int BrighthouseHandler::index_prev(uchar * buf)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

/*
 index_first() asks for the first key in the index.

 Called from opt_range.cc, opt_sum.cc, sql_handler.cc,
 and sql_select.cc.
 */
int BrighthouseHandler::index_first(uchar * buf)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}

/*
 index_last() asks for the last key in the index.

 Called from opt_range.cc, opt_sum.cc, sql_handler.cc,
 and sql_select.cc.
 */
int BrighthouseHandler::index_last(uchar * buf)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(HA_ERR_WRONG_COMMAND);
}


/*
 rnd_init() is called when the system wants the storage engine to do a table
 scan.
 See the example in the introduction at the top of this file to see when
 rnd_init() is called.

 Called from filesort.cc, records.cc, sql_handler.cc, sql_select.cc, sql_table.cc,
 and sql_update.cc.
 */
int BrighthouseHandler::rnd_init(mysql_bool scan)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	//FIXME there is no exception safety here - try-catch was added here to block exceptions
	//			and not letting them crash the server
	int ret = 1;
	try {
		if(m_query && !m_result && _table->NoObj() != 0) {
			m_cq->Result(m_tmp_table); //it is ALWAYS -2 though....
			m_result = true;

			try {
				FunctionExecutor fe(
					boost::bind(&Query::LockPackInfoForUse, m_query),
					boost::bind(&Query::UnlockPackInfoFromUse, m_query)
				);

				TempTable* push_down_result = m_query->Preexecute(*m_cq, false);
				if(!push_down_result || push_down_result->NoTables() != 1)
					throw InternalRCException("Query execution returned no result object");

				Filter* filter(push_down_result->GetMultiIndexP()->GetFilter(0));
				if(!filter)
					_filterPtr = FilterAutoPtr(new Filter(push_down_result->GetMultiIndexP()->OrigSize(0)));
				else
					_filterPtr = FilterAutoPtr(new Filter(*filter));

				_table = push_down_result->GetTableP(0);
				_tableIter = _table->begin(m_conn, *filter, get_attr_uses(table));
				_tableIterEnd = _table->end(m_conn);

			} catch (RCException const &e) {
				rccontrol << lock << "Error in push-down execution, push-down execution aborted: " << e.what() << unlock;
				rclog << lock << "Error in push-down execution, push-down execution aborted: " << e.what() << unlock;
			}
			delete m_query;
			delete m_cq;
			m_query = 0;
			m_cq = 0;
		} else {
			if(scan && _filterPtr.get()) {
				_tableIter = _table->begin(m_conn, *_filterPtr, get_attr_uses(table));
				_tableIterEnd = _table->end(m_conn);
			} else {
				rceng->GetTableIterator(m_table_path, _tableIter, _tableIterEnd, _table,
								get_attr_uses(table), table->in_use, m_conn);
				_filterPtr.reset();
			}
		}
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(ret);
}

int BrighthouseHandler::rnd_end()
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	reset();
	table->sort.found_records = 0;

	DBUG_RETURN(0);
}

/*
 This is called for each row of the table scan. When you run out of records
 you should return HA_ERR_END_OF_FILE. Fill buff up with the row information.
 The Field structure for the table is the key to getting data into buf
 in a manner that will allow the server to understand it.

 Called from filesort.cc, records.cc, sql_handler.cc, sql_select.cc, sql_table.cc,
 and sql_update.cc.
 */
int BrighthouseHandler::rnd_next(uchar *buf)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	int ret = HA_ERR_END_OF_FILE;
	try {
		table->status = 0;
		if(fill_row(buf) == HA_ERR_END_OF_FILE) {
			table->status = STATUS_NOT_FOUND;
			DBUG_RETURN(ret);
		}
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(ret);
}

/*
 position() is called after each call to rnd_next() if the data needs
 to be ordered. You can do something like the following to store
 the position:
 my_store_ptr(ref, ref_length, current_position);

 The server uses ref to store data. ref_length in the above case is
 the size needed to store current_position. ref is just a byte array
 that the server will maintain. If you are using offsets to mark rows, then
 current_position should be the offset. If it is a primary key like in
 BDB, then it needs to be a primary key.

 Called from filesort.cc, sql_select.cc, sql_delete.cc and sql_update.cc.
 */
void BrighthouseHandler::position(const uchar *record)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	my_store_ptr(ref, ref_length, current_position);

	DBUG_VOID_RETURN;
}

/*
 This is like rnd_next, but you are given a position to use
 to determine the row. The position will be of the type that you stored in
 ref. You can use ha_get_ptr(pos,ref_length) to retrieve whatever key
 or position you saved when position() was called.
 Called from filesort.cc records.cc sql_insert.cc sql_select.cc sql_update.cc.
 */
int BrighthouseHandler::rnd_pos(uchar * buf, uchar *pos)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	int ret = HA_ERR_END_OF_FILE;
	try {
		uint64 position = my_get_ptr(pos,ref_length);
		_tableIter = _table->find( m_conn, position, _filterPtr.get(), get_attr_uses( table ) );
		table->status = 0;
		if(fill_row(buf) == HA_ERR_END_OF_FILE) {
			table->status = STATUS_NOT_FOUND;
			DBUG_RETURN(ret);
		}
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(ret);
}

/*
 extra() is called whenever the server wishes to send a hint to
 the storage engine. The myisam engine implements the most hints.
 ha_innodb.cc has the most exhaustive list of these hints.
 */
int BrighthouseHandler::extra(enum ha_extra_function operation)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	/* This preemptive delete might cause problems here.
	 * Other place where it can be put is BrighthouseHandler::external_lock().
	 */
	if ( operation == HA_EXTRA_NO_CACHE ) {
		delete m_cq;
		m_cq = NULL;
		delete m_query;
		m_query = 0;

	}
	DBUG_RETURN(0);
}

/*
 First you should go read the section "locking functions for mysql" in
 lock.cc to understand this.
 This create a lock on the table. If you are implementing a storage engine
 that can handle transacations look at ha_berkely.cc to see how you will
 want to goo about doing this. Otherwise you should consider calling flock()
 here.

 Called from lock.cc by lock_external() and unlock_external(). Also called
 from sql_table.cc by copy_data_between_tables().
 */
int BrighthouseHandler::external_lock(THD* thd, int lock_type)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	try {
		m_conn = rceng->GetThreadParams(*thd);
		if(lock_type == F_UNLCK) { // RELEASE LOCK
			m_rctable = NULL;
			table_lock_manager.ReleaseLockIfNeeded(table_lock);

			// Rollback the transaction if operation finished with error
			//if(thd->query_error || thd->killed)
			if(thd->killed)
				rceng->Rollback(thd, true);

		} else {
			if(lock_type == F_RDLCK) { // READ LOCK
				if((table_lock = table_lock_manager.AcquireReadLock(*thd, m_table_path)).expired())
					DBUG_RETURN(1);
			} else {
				if((table_lock = table_lock_manager.AcquireWriteLock(*thd, m_table_path)).expired())
					DBUG_RETURN(1);
				trans_register_ha(thd, true, rcbase_hton);
				trans_register_ha(thd, false, rcbase_hton);
			}
		}

	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(0);
}

/*
 The idea with handler::store_lock() is the following:

 The statement decided which locks we should need for the table
 for updates/deletes/inserts we get WRITE locks, for SELECT... we get
 read locks.

 Before adding the lock into the table lock handler (see thr_lock.c)
 mysqld calls store lock with the requested locks.  Store lock can now
 modify a write lock to a read lock (or some other lock), ignore the
 lock (if we don't want to use MySQL table locks at all) or add locks
 for many tables (like we do when we are using a MERGE handler).

 Berkeley DB for example  changes all WRITE locks to TL_WRITE_ALLOW_WRITE
 (which signals that we are doing WRITES, but we are still allowing other
 reader's and writer's.

 When releasing locks, store_lock() are also called. In this case one
 usually doesn't have to do anything.

 In some exceptional cases MySQL may send a request for a TL_IGNORE;
 This means that we are requesting the same lock as last time and this
 should also be ignored. (This may happen when someone does a flush
 table when we have opened a part of the tables, in which case mysqld
 closes and reopens the tables and tries to get the same locks at last
 time).  In the future we will probably try to remove this.

 Called from lock.cc by get_lock_data().
 */
THR_LOCK_DATA **BrighthouseHandler::store_lock(THD *thd, THR_LOCK_DATA **to, enum thr_lock_type lock_type)
{
	if(lock_type != TL_IGNORE && m_lock.type == TL_UNLOCK)
		m_lock.type = lock_type;

	*to++= &m_lock;
	return to;
}

/*
 Ask rcbase handler about permission to cache table during query registration.
 If current thread is in non-autocommit, we don't permit any mysql query caching.
 */
my_bool BrighthouseHandler::register_query_cache_table(THD *thd, char *table_key,
   uint key_length,
   qc_engine_callback *call_back,
   ulonglong *engine_data )
{
	*call_back = rcbase_query_caching_of_table_permitted;
	return rcbase_query_caching_of_table_permitted(thd, table_key,	key_length,	0);
}

/*
 Used to delete a table. By the time delete_table() has been called all
 opened references to this table will have been closed (and your globally
 shared references released. The variable name will just be the name of
 the table. You will need to remove any files you have created at this point.

 If you do not implement this, the default delete_table() is called from
 handler.cc and it will delete all files with the file extentions returned
 by bas_ext().

 Called from handler.cc by delete_table and  ha_create_table(). Only used
 during create if the table_flag HA_DROP_BEFORE_CREATE was specified for
 the storage engine.
 */
int BrighthouseHandler::delete_table(const char *name)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	int ret = 1;
	try {
		rceng->DropTable(name);
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(ret);
}


/*
 Given a starting key, and an ending key estimate the number of rows that
 will exist between the two. end_key may be empty which in case determine
 if start_key matches any rows.

 Called from opt_range.cc by check_quick_keys().
 */
ha_rows BrighthouseHandler::records_in_range(uint inx, key_range *min_key, key_range *max_key)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);
	DBUG_RETURN(10); // low number to force index usage
}

/*
 create() is called to create a database. The variable name will have the name
 of the table. When create() is called you do not need to worry about opening
 the table. Also, the FRM file will have already been created so adjusting
 create_info will not do you any good. You can overwrite the frm file at this
 point if you wish to change the table definition, but there are no methods
 currently provided for doing that.

 Called from handle.cc by ha_create_table().
 */
int BrighthouseHandler::create(const char *name, TABLE *table_arg, HA_CREATE_INFO *create_info)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	BHErrorCode bhec = BHERROR_UNKNOWN;
	try {
		bhec = rceng->CreateTable(name, table_arg);
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}
	if(bhec != BHERROR_SUCCESS) {
		BHError bhe(bhec);
		my_message(bhe.ErrorCode(), bhe.Message().c_str(), MYF(0));
		DBUG_RETURN(TRUE);
	}
	m_table_path = new char[strlen(name)+1];
	strcpy(m_table_path, name);
	info(HA_STATUS_CONST);

	DBUG_RETURN(0);
}

int BrighthouseHandler::fill_row(uchar* buf) {
	if ( _tableIter == _tableIterEnd )
		return HA_ERR_END_OF_FILE;

	int is_null = 0;
	int col_id = 0;

	for(Field **field = table->field; *field; field++) {
		if(bitmap_is_set(table->read_set, col_id)) { //this time no '+1' is needed
			RCItem objItem = (*_tableIter)[col_id];
			RCEngine::Convert(is_null, *field, (*field)->ptr, &objItem);
			if(is_null == 0)
				(*field)->set_notnull();
			else {
				::memset( (*field)->ptr, 0, 2 );
				(*field)->set_null();
			}
		} else {
			::memset( (*field)->ptr, 0, 2 );
			(*field)->set_null();
		}

		col_id++;
	}

	current_position = (*_tableIter).GetRowId();

	//memset(buf, 0, table->s->null_bytes);
	++ _tableIter;

	return 0;
}

char* BrighthouseHandler::update_table_comment(const char * comment)
{
	//FIXME there is no exception safety here - try-catch was added here to block exceptions
	//			and not letting them crash the server
	char* ret = const_cast<char*>(comment);
	try {
		uint length = (uint) strlen(comment);
		char* str= NULL;
		uint extra_len = 0;

		if(length > 64000 - 3) {
			return ((char*)comment); // string too long
		}

		//  get size & ratio
		int64 sum_c=0, sum_u=0;
		AttrInfo* attr_info = 0;
		uint no_attr;
		if(table->in_use && table_lock_manager.HasAnyLockForTable(*table->in_use, m_table_path)) {
			rceng->GetCurrentTableStateAttributes(m_table_path, table_share, attr_info, no_attr);
		} else
			rceng->GetTableAttributes(m_table_path, table_share, attr_info, no_attr);
		for(uint j=0; j < no_attr; j++) {
			sum_c += attr_info[j].comp_size;
			sum_u += attr_info[j].uncomp_size;
		}
		delete [] attr_info;
		double ratio = (sum_c > 0 ? sum_u / double(sum_c) : 0);

		char buf[256];
		//int count = _snprintf(buf, 256, "Overall compression ratio: %.3f", ratio);
		int count = sprintf(buf, "Overall compression ratio: %.3f", ratio);
		extra_len += count;

		str = (char*)my_malloc(length + extra_len + 3, MYF(0));
		if(str) {
			char* pos = str + length;
			if(length) {
				memcpy(str, comment, length);
				*pos++ = ';';
				*pos++ = ' ';
			}

			//memcpy_s(pos, extra_len, buf, extra_len);
			memcpy(pos, buf, extra_len);
			pos[extra_len] = 0;
		}
		ret = str;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	return ret;
	//	return (char*) comment;
}

mysql_bool BrighthouseHandler::get_error_message(int error, String* buf)
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	DBUG_PRINT("enter", ("error: %d", error));
	if(error == HA_RCBASE_ERROR_TABLE_NOT_OPEN && err_msg) {
		buf->append(err_msg, (uint32)strlen(err_msg));
		err_msg[0] = '\0';
	}
	DBUG_PRINT("exit", ("message: %s", buf->ptr()));

	DBUG_RETURN(FALSE);
}

const COND *BrighthouseHandler::cond_push(const COND *a_cond)
{
	//FIXME there is no exception safety here - try-catch was added here to block exceptions
	//			and not letting them crash the server
	COND const* ret = a_cond;
	try {
		//if(_table)
			//return a_cond; //the iterator is already constructed, so why bother?

		if(!m_query)
			rceng->GetTableIterator(m_table_path, _tableIter, _tableIterEnd, _table,
							get_attr_uses(table), table->in_use, m_conn);

		COND* cond = const_cast<COND*>(a_cond);

		map<string, int> table_alias2index;
		map<string, TABLE*> table_alias2ptr;

		if(!m_query) {
			BHASSERT(m_conn!=NULL, "should be 'm_conn!=NULL'");
			m_query = new Query(m_conn);
			m_cq = new CompiledQuery;
			m_result = false;

			m_query->Add((RCTable*)_table);
			TabID t_out;
			m_cq->TableAlias(t_out, TabID(0)); //we apply it to the only table in this query
			m_cq->TmpTable(m_tmp_table, t_out);

			table_alias2index[table->alias] = t_out.n;
			table_alias2ptr[table->alias] = table;

			int col_no = 0;
			AttrID col, at, vc;
			TabID tab(m_tmp_table);
			for(Field **field = table->field; *field; field++) {
				AttrID at;
				if(bitmap_is_set(table->read_set, col_no) ) {
					col.n = col_no++;
					m_cq->CreateVirtualColumn(vc.n, m_tmp_table, t_out, col);
					m_cq->AddColumn(at, m_tmp_table, CQTerm(vc.n), LISTING, (*field)->field_name, false);
//				} else {
//					m_cq->AddColumn(at, m_tmp_table, CQTerm(), LISTING, "", false);
				}
			}
		}

		if(m_result)
			return a_cond; // if m_result there is already a result command in compilation

		CompiledQuery *tmp_cq = new CompiledQuery(*m_cq);

		if(!m_query->BuildFilter(cond, tmp_cq, &table_alias2index, &table_alias2ptr, m_tmp_table,
				WHERE_FILTER, false)) {
//			m_cq->TakeNoVirtColumns(tmp_cq);
			delete m_query;
			m_query = NULL;
			delete tmp_cq;
			return a_cond;
		} else {
			delete m_cq;
			m_cq = tmp_cq;
		}
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	return ret;
}

void BrighthouseHandler::cond_pop()
{
	return;
}

int BrighthouseHandler::reset()
{
	DBUG_ENTER(__PRETTY_FUNCTION__);

	//FIXME there is no exception safety here - try-catch was added here to block exceptions
	//			and not letting them crash the server
	int ret = 1;
	try {
		_tableIter = JustATable::RecordIterator();
		_tableIterEnd = JustATable::RecordIterator();
		_table = NULL;
		m_rctable = NULL;
		_filterPtr.reset();
		delete m_query;
		delete m_cq;
		m_query = 0;
		m_cq = 0;
		m_result = false;
		ret = 0;
	} catch (std::exception& e) {
		my_message(BHERROR_UNKNOWN, e.what(), MYF(0));
		rclog << lock << "Error: " << e.what() << unlock;
	} catch (...) {
		my_message(BHERROR_UNKNOWN, "An unknown system exception error caught.", MYF(0));
		rclog << lock << "Error: An unknown system exception error caught." << unlock;
	}

	DBUG_RETURN(0);
}

struct st_mysql_storage_engine brighthouse_storage_engine= { MYSQL_HANDLERTON_INTERFACE_VERSION };

mysql_declare_plugin(brighthouse)
{
	MYSQL_STORAGE_ENGINE_PLUGIN,
	&brighthouse_storage_engine,
	"BRIGHTHOUSE",
	"InfoBright",
	"Brighthouse storage engine",
	PLUGIN_LICENSE_PROPRIETARY,
	rcbase_init_func, /* Plugin Init */
	rcbase_done_func, /* Plugin Deinit */
	0x0001 /* 0.1 */,
	NULL, /* status variables                */
	NULL, /* system variables                */
	NULL /* config options                  */
}
mysql_declare_plugin_end;

