/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _SYSTEM_IOPARAMS_H_
#define _SYSTEM_IOPARAMS_H_

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////  IOParameters:   /////////////////////////////////////////////////////
//
// A class representing output description
//

#include "common/CommonDefinitions.h"
#include "system/IBFileSystem.h"

#ifndef __GNUC__
#define PATH_MAX  _MAX_PATH
#endif

class IOParameters
{
public:
	IOParameters()
	{
		Init();
	}

	IOParameters(const IOParameters& io_params)
		:	curr_output_mode(io_params.curr_output_mode), timeout(io_params.timeout), delimiter(io_params.delimiter),
			pipe_mode_tmp(io_params.pipe_mode_tmp),
			pipe_mode(io_params.pipe_mode),
			string_qualifier(io_params.string_qualifier),
			escape_character(io_params.escape_character),
			sign(1), minute(0),
			base_path(io_params.base_path),
			table_name(io_params.table_name), null_str(io_params.null_str),
			cur_len(io_params.cur_len)
	{
		strcpy(output_path, io_params.output_path);
	}

	IOParameters(std::string base_path, std::string table_name)
		: base_path(base_path), table_name(table_name)
	{
		Init();
	}

    IOParameters& operator= (const IOParameters& rhs)
    {
        Copy (rhs);
        return *this;
    }

	void SetParameter(Parameter param, int value)
	{
		switch(param)
		{
		case DELIMITER			: delimiter = (char)value; break;
		case STRING_QUALIFIER	: string_qualifier = (char)value; break;
		case TIMEOUT			: timeout = value; break;
		case PIPEMODE			: pipe_mode = value; break;
		}
	}

	void SetTimeZone(short _sign, short _minute) { sign = _sign; minute = _minute; }
	void GetTimeZone(short& _sign, short& _minute) const { _sign = sign; _minute = minute; }

	void SetEscapeCharacter(char value)
	{
		this->escape_character = value;
	}

	void SetNullsStr(std::string& null_str)
	{
		this->null_str = null_str;
	}

	void SetOutput(int _mode, const char *fname, int _pipe_mode_tmp = -1)		// mode: 0 - standard console-style output with header
		// fname: NULL for console
		// These settings will work until the next change
	{
		curr_output_mode=_mode;
		if(fname)
			strcpy(output_path, fname);
		else
			output_path[0]='\0';
		pipe_mode_tmp = _pipe_mode_tmp != -1 ? _pipe_mode_tmp : pipe_mode;
	}

	//int Mode() const { return curr_output_mode; }
	EDF 		GetEDF() const;
	const char*	Path() const			{ return output_path; }
	int			PipeMode() const		{ return pipe_mode_tmp; }
	int			Timeout() const			{ return timeout; }
	std::string	NullsStr() const		{ return null_str; }
	char		Delimiter() const		{ return delimiter; }
	char		StringQualifier() const { return string_qualifier; }
	char		EscapeCharacter() const { return escape_character; }
	std::string	BasePath() 				{ return base_path; }
	std::string	TableName() const		{ return table_name; }
	std::string	TablePath() const
	{
		return MakeDirectoryPathOsSpecific(base_path) += TableName();
	}

	/* Output format:
		0. An ushort(2 byte) value specifying len of output bufer len
		1. A pointer to a null-terminated string that specifies the database's path
		2. A pointer to a null-terminated string that specifies the name of the table
		3. A pointer to a null-terminated string that specifies output/input file/pipe
		4. 1 character that specifies output/input mode
		5. 1 character that specifies delimiter
		6. 1 character that specifies string qualifier
		6. 1 character that specifies escape character
		7. 1 character that specifies pipe's mode
		8. 4 byte integer that specifies timeout parameter. This should be in seconds
		9. 2 byte integer that specifies sign of time zone difference client/server
		10. 2 byte integer that specifies time zone difference in minutes
		11. NULL-terminated string - name of an object use for interprocess synchronization by RSI_MANAGER
	*/
	bool PutParameters(char* dest, int dest_len)
	{
		int left = dest_len;
		size_t msg_len = 0;
		msg_len = (base_path.length() + 1) + (table_name.length() + 1) + (strlen(output_path) + 1) + null_str.length() + 1;
		msg_len += 1 + 1 + 1 + 1 + 1 + 4 + 2 + 2;

		if((int)msg_len > dest_len)
			return false;
		cur_len = 2;

		strcpy(dest + cur_len, base_path.c_str());
		cur_len += (ushort)base_path.length() + 1;

		strcpy(dest + cur_len, table_name.c_str());
		cur_len += (ushort)table_name.length() + 1;

		strcpy(dest + cur_len, output_path);
		cur_len += (ushort)strlen(output_path) + 1;

		*(dest + cur_len) = (char)curr_output_mode;
		cur_len++;

		*(dest + cur_len) = (char)delimiter;
		cur_len++;

		*(dest + cur_len) = (char)string_qualifier;
		cur_len++;

		*(dest + cur_len) = (char)escape_character;
		cur_len++;

		*(dest + cur_len) = (char)pipe_mode_tmp;
		cur_len++;

		*(int*)(dest + cur_len) = timeout;
		cur_len += 4;

		*(short*)(dest + cur_len) = sign;
		cur_len += 2;

		*(short*)(dest + cur_len) = minute;
		cur_len += 2;

		strcpy((dest + cur_len), null_str.c_str());
		cur_len += (ushort)null_str.length() + 1;

		*(ushort*)dest = cur_len - 2;

		return true;
	}

	void GetParameters(char* src)
	{
		ushort read_len = *(ushort*)src;
		cur_len = 0;
		src += 2;
		base_path = src;
		cur_len += (ushort)base_path.length() + 1;

		table_name = src + cur_len;
		cur_len += (ushort)table_name.length() + 1;

		strcpy(output_path, src + cur_len);
		cur_len += (ushort)strlen(output_path) + 1;

		curr_output_mode = *(src + cur_len);
		cur_len++;

		delimiter = *(src + cur_len);
		cur_len++;

		string_qualifier = *(src + cur_len);
		cur_len++;

		escape_character = *(src + cur_len);
		cur_len++;

		pipe_mode_tmp = *(src + cur_len);
		cur_len++;

		timeout = *(int*)(src + cur_len);
		cur_len += 4;

		sign = *(short*)(src + cur_len);
		cur_len += 2;

		minute = *(short*)(src + cur_len);
		cur_len += 2;

		null_str = src + cur_len;
		cur_len += (ushort)null_str.length() + 1;

	}

	ushort GetParamsBufferLen()
	{
		return cur_len;
	}

	static ushort GetParamsBufferLen(char* src)
	{
		return *(ushort*)src;
	}

private:
	int curr_output_mode;		// I/O file format - see RCTable::SaveTable parameters
	char output_path[1024];		// input or output file name, set by "interface..."
	int timeout;
	char delimiter;
	int pipe_mode_tmp;
	int pipe_mode;				// 0 - client, <> 0 - server
	char string_qualifier;
	char escape_character;

	// TimeZone
	short sign;
	short minute;

	std::string base_path;
	std::string table_name;

	std::string null_str;

	ushort cur_len;

private:
	void Init()
	{
		output_path[0]='\0';
		curr_output_mode=0;
		delimiter = DEFAULT_DELIMITER;
		string_qualifier = DEFAULT_STRING_QUALIFIER;
		timeout = DEFAULT_PIPE_TIMEOUT;
		pipe_mode_tmp = pipe_mode = DEFAULT_PIPE_MODE;
		cur_len = 0;
		escape_character = 0;
		sign = 1;
		minute = 0;
	}

    void Copy (const IOParameters& rhs)
    {
	    curr_output_mode = rhs.curr_output_mode;
	    memcpy (output_path, rhs.output_path, sizeof(output_path));
	    timeout = rhs.timeout;
	    delimiter = rhs.delimiter;
	    string_qualifier = rhs.string_qualifier;
	    escape_character = rhs. escape_character;
	    pipe_mode_tmp = rhs.pipe_mode_tmp;
	    pipe_mode = rhs.pipe_mode;
	    base_path = rhs.base_path;
	    table_name = rhs.table_name;
	    null_str = rhs.null_str;
	    cur_len = rhs.cur_len;
	    sign = rhs.sign;
	    minute = rhs.minute;
    }

public:
//	static int IsPipe(char* name)
//	{
//		if (strncasecmp(name, "\\\\", 2)!=0) return false;
//		const char *rest = strstr(name+2, "\\");
//		if ((!rest) || (rest == name+2)) return false;   // backslash not found or found at the beginning of name+2
//		if (strncasecmp(rest, "\\pipe\\", 6)!=0) return false;
//		return true;
//	}

	//static EDF GetEDF(int edf);
};

#endif //_SYSTEM_IOPARAMS_H_

