/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#include <sstream>

#include "common/bhassert.h"
#include "system/IBFile.h"
#include "util/SystemInformation.h"

#ifdef __GNUC__

int getmeminfo(_meminfo &meminfo)
{
	memset(&meminfo, 0, sizeof(meminfo));

#if defined(linux) || defined(_linux) || \
	defined(__linux) || defined(__linux__)

	// Pull the information from the kernel
	struct sysinfo system_info;
	if( 0 != sysinfo(&system_info) )
	{
		// Bail, problem occurred.
		// TODO: Message
		return -1;
	}

	// Grab the physical and swap memory totals and unused
	// segments.
	//
	// Important Note: As of kernel version 2.3.48, the mem_unit
	// must be factored in.
	meminfo.ram_total = system_info.totalram * system_info.mem_unit;
	meminfo.ram_free = system_info.freeram * system_info.mem_unit;
	meminfo.swap_total = system_info.totalswap * system_info.mem_unit;
	meminfo.swap_free = system_info.freeswap * system_info.mem_unit;

	// Calculate the remaining two fields.
	meminfo.ram_use = meminfo.ram_total - meminfo.ram_free;
	meminfo.swap_use = meminfo.swap_total - meminfo.swap_free;

	// Done, success.
	return 0;
#else
	return -1;
#endif
}

_uint64 GetProcessVirtualMemory(pid_t process_id) throw(RCException)
{
	try {
		std::stringstream filename_builder;
		filename_builder << "/proc/" << process_id << "/stat";

		IBFile process_file;
		process_file.OpenReadOnly(filename_builder.str().c_str());

		const int buffer_size = 1024;
		char buffer[buffer_size];

		int read_size = process_file.Read(buffer, buffer_size);
		BHASSERT(read_size > 0 &&  read_size < buffer_size, "proc parser sanity check");

		return ExtractProcessVirtualMemory(std::string(buffer, read_size));
	}
	catch(std::exception e) {
		throw RCException(std::string("Unable to retrieve process VM usage - ") + e.what());
	}
}

_uint64 ExtractProcessVirtualMemory(std::string statistics) throw(RCException)
{
	// The closing bracket ')' before the R (running flag) is the last static
	// point of reference in the stats line.  The vmem size field is 21 fields
	// to the right of that ')'. Refer to the 'proc' man pages more information. 
	std::stringstream tokens(statistics.substr(statistics.find_last_of(')') + 1)); 
	std::string token;

	for (size_t position = 21; position != 0; --position ) {
		if( !(tokens >> token) ) {
			throw RCException("Fragmented system process record encountered");
		}
	}
	return strtoull(token.c_str(), NULL, 10);
}

#endif

/////////////////////////////////////////////////////////////////////////

#ifdef __WIN__
#include <windows.h>
#include <psapi.h>
#endif

size_t VirtualMemoryUsage()
{
#ifdef _MSC_VER
    PROCESS_MEMORY_COUNTERS_EX pmc;
    GetProcessMemoryInfo(GetCurrentProcess(), (PROCESS_MEMORY_COUNTERS*)&pmc, sizeof(pmc));
	return pmc.PrivateUsage;
#else
	return GetProcessVirtualMemory(getpid());
#endif
}

unsigned long long PhysicalMemoryAvailable()
{
#ifdef _MSC_VER
	MEMORYSTATUSEX statex;
	statex.dwLength = sizeof (statex);
	GlobalMemoryStatusEx(&statex);
	return statex.ullAvailPhys;
#else
#ifdef SOLARIS
	unsigned long long pageSize, freePages;
	
	freePages = sysconf( _SC_AVPHYS_PAGES);
	pageSize = sysconf(_SC_PAGESIZE);
	return (freePages * pageSize);
#else
	_meminfo meminfo;
	getmeminfo(meminfo);
	return meminfo.ram_free;
#endif
#endif
}


