/* Copyright (C)  2005-2008 Infobright Inc.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.0 as
published by the Free  Software Foundation.

This program is distributed in the hope that  it will be useful, but
WITHOUT ANY WARRANTY; without even  the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License version 2.0 for more details.

You should have received a  copy of the GNU General Public License
version 2.0  along with this  program; if not, write to the Free
Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
02111-1307 USA  */

#ifndef _BHDURATION_H_
	#define _BHDURATION_H_

#include <sys/time.h>

/*!
 	\brief A over-simplified class for tracking the duration between 
	timing marks.  This is not intended for tight loops and short 
	periods, it is for rough timings on spans of a half second or
	more.  If you need something more accurate, implement from the
	system clock.  
*/
class bhDuration
{
	protected:
		// Copy protection
		bhDuration(const bhDuration &);
		const bhDuration & operator=(const bhDuration &);

		//! Time of the last mark
		struct timeval _mark;
	
	public:
		/*!
 			\brief Simple constructor.  

			This constructor also initializes the object to the current
			time by way of the mark() call.

			\sa mark()
		 */
		bhDuration() 
		{
			mark();
		}
	
		/*!
 			\brief Updates the current time mark.

			This method simply sets the mark member to the current time.

			\sa interval()
 		 */
		void mark()
		{
			gettimeofday(&_mark, 0);
		}

		/*!
 		 	\brief Get the elapsed time.

			Fetches and returns the elapsed time since the previous mark
			and updates the mark to the current time.
		
			\return The elapsed time since the last mark.

			\sa mark()
		 */
		double interval()
		{
			struct timeval now;

			// Grab the current time
			gettimeofday(&now, 0);

			// Calculate the time since the mark
			double seconds = (now.tv_usec - _mark.tv_usec) * 0.000001
				+ now.tv_sec - _mark.tv_sec;

			// Reset the mark
			mark();

			// Done...
			return seconds;
		}
};

#endif // _BHDURATION_H_
