/* Copyright (C)  2005-2008 Infobright Inc.

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License version 2.0 as
	 published by the Free  Software Foundation.

	 This program is distributed in the hope that  it will be useful, but
	 WITHOUT ANY WARRANTY; without even  the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	 General Public License version 2.0 for more details.

	 You should have received a  copy of the GNU General Public License
	 version 2.0  along with this  program; if not, write to the Free
	 Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
	 02111-1307 USA  */


#ifndef CONSTCOLUMN_H_INCLUDED
#define CONSTCOLUMN_H_INCLUDED

#include "core/MIUpdatingIterator.h"
#include "core/PackGuardian.h"
#include "VirtualColumn.h"

/*! \brief A column defined by an expression (including a subquery) or encapsulating a PhysicalColumn
 * ConstColumn is associated with an MultiIndex object and cannot exist without it.
 * Values contained in ConstColumn object may not exist physically, they can be computed on demand.
 *
 */
class ConstColumn : public VirtualColumn
{
public:

	/*! \brief Create a column that represent const value.
	 *
	 * \param val - value for const column.
	 * \param ct - type of const column.
	 */
	ConstColumn(ValueOrNull const& val, ColumnType const& ct) : VirtualColumn(ct, NULL), value(val) {
		dim = -1;
	}
	ConstColumn(const RCValueObject& v, const ColumnType& c) : VirtualColumn(c, NULL), value() {
		dim = -1;
		if(c.IsString()) {
			value = ValueOrNull(v.ToRCString());
		} else if(c.IsNumeric() && !c.IsDateTime()) {
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT(v.GetValueType() == NUMERIC_TYPE);
			// TODO: if it is non-numeric a proper conversion should be done here
			value = ValueOrNull(static_cast<RCNum&>(v));
		} else {
			BHASSERT_WITH_NO_PERFORMANCE_IMPACT(v.GetValueType() == DATE_TIME_TYPE);
			// TODO: if it is non-date-time a proper conversion should be done here
			value = ValueOrNull(static_cast<RCDateTime&>(v));
		}
	}
	ConstColumn(ConstColumn const& cc) : VirtualColumn(cc.ct, cc.mind), value(cc.value) {};
	~ConstColumn() {}

	virtual bool IsConst() const { return true; }
	virtual bool RoughNullsOnly(const MIIterator &mit) { return value.IsNull(); }

	const MysqlExpression::bhfields_cache_t& GetBHItems() const { static MysqlExpression::bhfields_cache_t const dummy; return dummy; }
	char *ToString(char p_buf[], size_t buf_ct) const;
protected:
	virtual const char* DoGetValueStringPtr(MIIterator const&) { return value.IsNull() ? "NULL" : value.sp; }
	virtual BHReturnCode DoGetValueRCNum(MIIterator const&, RCNum& rc);
	virtual BHReturnCode DoGetValueDateTime(MIIterator const&, RCDateTime& dt, AttributeType t = RC_UNKNOWN);
	virtual _int64 DoGetValueInt64(const MIIterator &mit) { return value.IsNull() ? NULL_VALUE_64 : value.Get64(); }
	virtual int DoGetValueInt32(const MIIterator &mit) { return value.IsNull() ? NULL_VALUE_32 : (int)value.Get64(); }
	virtual bool DoIsNull(const MIIterator &mit) { return value.IsNull(); }
	virtual void DoGetValueString(RCBString& s, const MIIterator &mit, bool bin_as_hex = false) { value.GetString(s); }
	virtual ValueSet* DoGetSet(MIIterator const& mit) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT( !"Invalid call for this type of column." );
		return NULL;
	}
	virtual double DoGetValueDouble(const MIIterator& mit);
	virtual void DoGetValueBin(const MIIterator& mit, int& size, char* buf);
	virtual RCValueObject DoGetValue(const MIIterator& mit, bool lookup_to_num);
	virtual _int64 DoGetMinInt64(const MIIterator &mit) { return value.IsNull() ? NULL_VALUE_64 : value.Get64(); }
	virtual _int64 DoGetMaxInt64(const MIIterator &mit) { return value.IsNull() ? NULL_VALUE_64 : value.Get64(); }
	virtual _int64 DoRoughMin()	{ return value.IsNull() ? NULL_VALUE_64 : value.Get64(); }
	virtual _int64 DoRoughMax() { return value.IsNull() ? NULL_VALUE_64 : value.Get64(); }
	virtual RCBString DoGetMaxString(const MIIterator &mit);
	virtual RCBString DoGetMinString(const MIIterator &mit);
	virtual _int64 DoGetNoNulls(const MIIterator &mit) { return value.IsNull() ? mit.GetPackSizeLeft() : 0; }
	virtual bool DoNullsPossible() { return value.IsNull(); }
	virtual _int64 DoGetSum(const MIIterator &mit);
	virtual bool DoIsDistinct() { return false; }
	virtual _int64 DoGetApproxDistVals(bool incl_nulls);
	virtual _int64 GetExactDistVals();
	virtual ushort DoMaxStringSize();		// maximal byte string length in column
	virtual PackOntologicalStatus DoGetPackOntologicalStatus(const MIIterator& mit);
	virtual RSValue DoRoughCheck(const MIIterator& it, Descriptor& d);
	virtual void DoEvaluatePack(MIUpdatingIterator& mit, Descriptor& desc);
	ValueOrNull value;
};

#endif /* not CONSTCOLUMN_H_INCLUDED */

