/* Copyright (C)  2005-2008 Infobright Inc.

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License version 2.0 as
	 published by the Free  Software Foundation.

	 This program is distributed in the hope that  it will be useful, but
	 WITHOUT ANY WARRANTY; without even  the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	 General Public License version 2.0 for more details.

	 You should have received a  copy of the GNU General Public License
	 version 2.0  along with this  program; if not, write to the Free
	 Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
	 02111-1307 USA  */


#ifndef CONSTEXPRESSIONCOLUMN_H_INCLUDED
#define CONSTEXPRESSIONCOLUMN_H_INCLUDED

#include "core/MIUpdatingIterator.h"
#include "core/PackGuardian.h"
#include "ExpressionColumn.h"

/*! \brief A column defined by an expression (including a subquery) or encapsulating a PhysicalColumn
 * ConstExpressionColumn is associated with an MultiIndex object and cannot exist without it.
 * Values contained in ConstExpressionColumn object may not exist physically, they can be computed on demand.
 *
 */
class ConstExpressionColumn : public ExpressionColumn
{
public:

	/*! \brief Create a column that represent const value.
	 *
	 * \param val - value for const column.
	 * \param ct - type of const column.
	 */
	ConstExpressionColumn(MysqlExpression* expr, TempTable* temp_table, int temp_table_alias, MultiIndex* mind) :
			ExpressionColumn(expr, temp_table, temp_table_alias, mind){

		//status = VC_CONST;
		dim = -1;
		if(params.size() == 0)
			last_val = expr->Evaluate();
	}

	ConstExpressionColumn(MysqlExpression* expr, ColumnType forced_ct, TempTable* temp_table, int temp_table_alias, MultiIndex* mind) :
			ExpressionColumn(expr, temp_table, temp_table_alias, mind){
		// special case when naked column is a parameter
		//status = VC_CONST; Used for naked attributes as parameters
		dim = -1;
//		if(params.size() == 0)
//			last_val = expr->Evaluate();
		ct = forced_ct;
	}

	ConstExpressionColumn(ConstExpressionColumn const& cc) : ExpressionColumn(NULL, NULL, NULL_VALUE_32, NULL) {
		last_val = cc.last_val;
		ct = cc.ct;
		first_eval = cc.first_eval;
		if (!expr) assert(0 && "verify if copying expression ConstExpressionColumn is OK");}

	~ConstExpressionColumn() {}

	virtual bool IsConst() const { return true; }
	virtual bool RoughNullsOnly(const MIIterator &mit) { return last_val.IsNull(); }
	virtual void RequestEval();

protected:
	virtual const char* DoGetValueStringPtr(MIIterator const&) { return last_val.IsNull() ? "NULL" : last_val.sp; }
	virtual BHReturnCode DoGetValueRCNum(MIIterator const&, RCNum& rc);
	virtual BHReturnCode DoGetValueDateTime(MIIterator const&, RCDateTime& dt, AttributeType t = RC_UNKNOWN);
	virtual _int64 DoGetValueInt64(const MIIterator &mit) { return last_val.IsNull() ? NULL_VALUE_64 : last_val.Get64(); }
	virtual int DoGetValueInt32(const MIIterator &mit) { return last_val.IsNull() ? NULL_VALUE_32 : (int)last_val.Get64(); }
	virtual bool DoIsNull(const MIIterator &mit) { return last_val.IsNull(); }
	virtual void DoGetValueString(RCBString& s, const MIIterator &mit, bool bin_as_hex = false) { last_val.GetString(s); }
	virtual ValueSet* DoGetSet(MIIterator const& mit) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT( !"Invalid call for this type of column." );
		return NULL;
	}

	virtual double DoGetValueDouble(const MIIterator& mit);
	virtual void DoGetValueBin(const MIIterator& mit, int& size, char* buf);
	virtual RCValueObject DoGetValue(const MIIterator& mit, bool lookup_to_num);
	virtual _int64 DoGetMinInt64(const MIIterator &mit) { return last_val.IsNull() ? NULL_VALUE_64 : last_val.Get64(); }
	virtual _int64 DoGetMaxInt64(const MIIterator &mit) { return last_val.IsNull() ? NULL_VALUE_64 : last_val.Get64(); }
	virtual _int64 DoRoughMin()	{ return last_val.IsNull() ? NULL_VALUE_64 : last_val.Get64(); }
	virtual _int64 DoRoughMax() { return last_val.IsNull() ? NULL_VALUE_64 : last_val.Get64(); }
	virtual RCBString DoGetMaxString(const MIIterator &mit);
	virtual RCBString DoGetMinString(const MIIterator &mit);
	virtual _int64 DoGetNoNulls(const MIIterator &mit) { return last_val.IsNull() ? mit.GetPackSizeLeft() : 0; }
	virtual bool DoNullsPossible() { return last_val.IsNull(); }
	virtual _int64 DoGetSum(const MIIterator &mit);
	virtual bool DoIsDistinct() { return (mind->TooManyTuples() || mind->NoTuples() > 1) ? false : (!last_val.IsNull()); }
	virtual _int64 DoGetApproxDistVals(bool incl_nulls);
	virtual ushort DoMaxStringSize();		// maximal byte string length in column
	virtual PackOntologicalStatus DoGetPackOntologicalStatus(const MIIterator& mit);
	virtual RSValue DoRoughCheck(const MIIterator& it, Descriptor& d);
	virtual void DoEvaluatePack(MIUpdatingIterator& mit, Descriptor& desc);
};

#endif /* not CONSTEXPRESSIONCOLUMN_H_INCLUDED */

