/* Copyright (C)  2005-2008 Infobright Inc.

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License version 2.0 as
	 published by the Free  Software Foundation.

	 This program is distributed in the hope that  it will be useful, but
	 WITHOUT ANY WARRANTY; without even  the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	 General Public License version 2.0 for more details.

	 You should have received a  copy of the GNU General Public License
	 version 2.0  along with this  program; if not, write to the Free
	 Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
	 02111-1307 USA  */


#ifndef SINGLECOLUMN_H_INCLUDED
#define SINGLECOLUMN_H_INCLUDED

#include "core/MIUpdatingIterator.h"
#include "core/PackGuardian.h"
#include "core/PhysicalColumn.h"
#include "VirtualColumn.h"

/*! \brief A column defined by an expression (including a subquery) or encapsulating a PhysicalColumn
 * SingleColumn is associated with an MultiIndex object and cannot exist without it.
 * Values contained in SingleColumn object may not exist physically, they can be computed on demand.
 *
 */
class SingleColumn : public VirtualColumn
{
public:

	/*! \brief Create a virtual column for a given column
	 *
	 * \param col the column to be "wrapped" by VirtualColumn
	 * \param _mind the multiindex to which the SingleColumn is attached.
	 * \param alias the alias number of the source table
	 * \param col_no number of the column in the table, negative if col of TempTable
	 * \param source_table pointer to the table holding \endcode col
	 * \param dim dimension number in the multiindex holding the column.
	 * \e note: not sure if the pointer to the MultiIndex is necessary
	 */
	SingleColumn(PhysicalColumn* col, MultiIndex* mind, int alias, int col_no, JustATable* source_table, int dim);
	~SingleColumn();
	virtual _uint64 ApproxAnswerSize(Descriptor& d) { return col->ApproxAnswerSize(d); }
	PhysicalColumn* GetPhysical() const { return col; }
	bool IsConst() const {return false;}
	virtual bool RoughNullsOnly(const MIIterator &mit) 
	{ return col->GetPackOntologicalStatus(mit.GetCurPackrow(dim)) == NULLS_ONLY; }
	virtual char *ToString(char p_buf[], size_t buf_ct) const;
	virtual void TranslateSourceColumns(map<PhysicalColumn *, PhysicalColumn *> &);

protected:

	virtual _int64 DoGetValueInt64(const MIIterator &mit) {
		return col->GetValueInt64(mit[dim]);
	}
	virtual int DoGetValueInt32(const MIIterator &mit) {
		return col->GetValueInt32(mit[dim]);
	}
	virtual bool DoIsNull(const MIIterator &mit) {
		return col->IsNull(mit[dim]);
	}
	virtual void DoGetValueString(RCBString& s, const MIIterator &mit, bool bin_as_hex = false) {
		col->GetValueString(mit[dim], s, bin_as_hex);
	}

	virtual double DoGetValueDouble(const MIIterator& mit);
	virtual void DoGetValueBin(const MIIterator& mit, int& size, char* buf);
	virtual RCValueObject DoGetValue(const MIIterator& mit, bool lookup_to_num);
	virtual _int64 DoGetMinInt64(const MIIterator &mit);
	virtual _int64 DoGetMaxInt64(const MIIterator &mit);
	virtual _int64 DoGetMinInt64Exact(const MIIterator &mit);
	virtual _int64 DoGetMaxInt64Exact(const MIIterator &mit);
	virtual _int64 DoRoughMin()
		{	return col->RoughMin(mind->GetFilter(dim));	}
	virtual _int64 DoRoughMax()
	{	return col->RoughMax(mind->GetFilter(dim)); }
	virtual double RoughSelectivity();

//	virtual RCBString DoGetMaxString(const MIIterator &mit);
//	virtual RCBString DoGetMinString(const MIIterator &mit);
	virtual _int64 DoGetNoNulls(const MIIterator &mit) {
		_int64 res;
		if(mit.GetCurPackrow(dim) >= 0) {
			res = col->GetNoNulls(mit.GetCurPackrow(dim));
			if(res == 0 || mit.WholePack(dim))
				return res;
		}
		return NULL_VALUE_64;
	}
	virtual bool DoNullsPossible() {
		if(mind->NullsExist(dim))
			return true;
		if(col->GetNoNulls(-1) > 0)
			return true;
		return false;
	}
	virtual _int64 DoGetSum(const MIIterator &mit);

	virtual bool DoIsDistinct() {
		return (col->IsDistinct(mind->GetFilter(dim)) && mind->CanBeDistinct(dim));
	}

	virtual _int64 DoGetApproxDistVals(bool incl_nulls);
	virtual _int64 GetExactDistVals();
	virtual ushort DoMaxStringSize();		// maximal byte string length in column
	virtual PackOntologicalStatus DoGetPackOntologicalStatus(const MIIterator& mit);
	virtual RSValue DoRoughCheck(const MIIterator& it, Descriptor& d);
	RSValue RoughCheckBetween(const MIIterator& mit, _int64 min, _int64 max);
	virtual void DoEvaluatePack(MIUpdatingIterator& mit, Descriptor& desc);
	virtual bool TryToMerge(Descriptor &d1,Descriptor &d2);
	virtual RCBString DecodeValue_S(_int64 code)	{ return col->DecodeValue_S(code); }
	virtual vector<_int64> GetListOfDistinctValues(MIIterator const& mit);
	virtual void DisplayAttrStats();

	const MysqlExpression::bhfields_cache_t& GetBHItems() const {static MysqlExpression::bhfields_cache_t const dummy; return dummy;}

	PhysicalColumn* col; 			//!= NULL if SingleColumn encapsulates a single column only (no expression)
	//	this an easily accessible copy var_map[0].tab->GetColumn(var_map[0].col_ndx)
};

#endif /* not SINGLECOLUMN_H_INCLUDED */

