/* Copyright (C)  2005-2008 Infobright Inc.

	 This program is free software; you can redistribute it and/or modify
	 it under the terms of the GNU General Public License version 2.0 as
	 published by the Free  Software Foundation.

	 This program is distributed in the hope that  it will be useful, but
	 WITHOUT ANY WARRANTY; without even  the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	 General Public License version 2.0 for more details.

	 You should have received a  copy of the GNU General Public License
	 version 2.0  along with this  program; if not, write to the Free
	 Software Foundation,  Inc., 59 Temple Place, Suite 330, Boston, MA
	 02111-1307 USA  */


#ifndef VIRTUALCOLUMN_H_INCLUDED
#define VIRTUALCOLUMN_H_INCLUDED

#include <vector>

#include "core/Column.h"
#include "core/MultiIndex.h"
#include "core/PackGuardian.h"
#include "core/MysqlExpression.h"
#include "core/JustATable.h"

class MIIterator;

/*! \brief A column defined by an expression (including a subquery) or encapsulating a PhysicalColumn
 * VirtualColumn is associated with an MultiIndex object and cannot exist without it.
 * Values contained in VirtualColumn object may not exist physically, they can be computed on demand.
 *
 */


class VirtualColumn : public Column DEBUG_STATEMENT( DEBUG_COMA public DestructionDebugger<VirtualColumn> )
{
public:
	typedef std::set<int> dimensions_t;
	struct VarMap {
		VarMap(VarID v, JustATable* t, int d)
			:	var(v), col_ndx( v.col < 0 ? -v.col - 1 : v.col), dim(d), tab(t->shared_from_this())
		{}
		VarID			var;		//variable identified by table alias and column number
		int 			col_ndx;	//column index (not negative)
		int				dim;		//dimension in multiindex corresponding to the alias

		JustATablePtr	GetTabPtr() const {	return tab.lock(); }

		bool operator < ( VarMap const& v ) const {
			return ( var < v.var )
				|| ( ( var == v.var ) && ( col_ndx < v.col_ndx ) )
				|| ( ( var == v.var ) && ( col_ndx == v.col_ndx ) && ( GetTabPtr() < v.GetTabPtr() ) )
				|| ( ( var == v.var ) && ( col_ndx == v.col_ndx ) && ( GetTabPtr() == v.GetTabPtr() ) && ( dim < v.dim ) );
		}
		JustATableWeakPtr		tab;		//table corresponding to the alias
	};
	// 4 variants are recognized
	// a) VC_PHYSICAL :  VirtualColumn encapsulates a single physical column only (no expression)
	// b) VC_EXPR : expr != NULL & ! vars.empty() & deterministic			- VirtualColumn encapsulates a non-trivial expression
	// c) VC_EXPR_NONDET : expr != NULL & ! vars.empty() & !deterministic - VirtualColumn encapsulates a non-trivial expression giving not deterministic results (eg. rand())
	// d) VC_CONST : expr != NULL & vars.empty() & deterministic	- VirtualColumn encapsulates a constant
	enum VColStatus {VC_CONST, VC_PHYSICAL, VC_EXPR, VC_EXPR_NONDET};

	/*! \brief Create a virtual column for a given column type.
	 *
	 * \param col_type the column to be "wrapped" by VirtualColumn
	 * \param mind the multiindex to which the VirtualColumn is attached.
	 */
	VirtualColumn(ColumnType const& col_type, MultiIndex* mind);
	VirtualColumn(VirtualColumn const& vc);

	virtual ~VirtualColumn();

	/////////////// Data access //////////////////////

	/*! \brief Get a numeric value from a column
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * Use it for numeric columns only.
	 * Old name: \b GetTable64
	 *
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 * \return numeric value from the column in level-1 encoding.
	 * - Integers values can be used directly
	 * - Decimals must be externally properly shifted
	 * - Doubles must be cast e.g. *(double*)&
	 */
	inline _int64 GetValueInt64(const MIIterator &mit) {
		return DoGetValueInt64(mit);
	}

	/*! \brief Get a numeric value from a column
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * Use it for numeric columns only.
	 * Old name: \b GetTable32
	 *
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 * \return numeric value from the column.
	 * - Integers values can be used directly
	 * - Floats must be cast e.g. *(float*)&
	 */
	inline int GetValueInt32(const MIIterator &mit) {
		return DoGetValueInt32(mit);
	}

	/*! \brief Is the column value NULL ?
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 * \return \b true if column value is NULL, \b false otherwise
	 */
	inline bool IsNull(const MIIterator &mit) {
		return DoIsNull(mit);
	}

	/*! \brief Get a non null-terminated String from a column
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * Can be used to retrieve binary data
	 *
	 * \param s a string object to be filled with a retrieved value
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 * \param bin_as_hex in case of a binary column, present the value as hexadecimal number
	 */
	inline void GetValueString(RCBString& s, const MIIterator &mit, bool bin_as_hex = false) {
		DoGetValueString(s, mit, bin_as_hex);
	}

	/*! \brief Get a numeric value from a column
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 * \return RCNum object - a representation of a numeric value
	 */
	//	 RCNum& GetValueNum(const MIIterator& mit);

	/*! \brief Get a date/time value from a column
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 * \return RCDateTime object - a representation of a date/time value
	 */
	//	 RCDateTime& GetValueDateTime(const MIIterator& mit);

	/*! \brief Get a double value from a column, possibly converting the original value
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 */
	inline double GetValueDouble(const MIIterator& mit) {
		return (DoGetValueDouble(mit));
	}

	/*! \brief Get a binary value from a column
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 * \param size the size of buffer, set to the number of copied bytes on return
	 * \param buf buffer into which the bytes are copied
	 */
	inline void GetValueBin(const MIIterator& mit, int& size, char* buf) {
		return (DoGetValueBin(mit, size, buf));
	}

	/*! \brief Get a value from a colunm, whatever the column type is
	 *
	 * \pre necessary datapacks (containing rows pointed by \e mit) are loaded and locked
	 *
	 * \param mit points to a row in an Multiindex, requested column value comes from this row
	 * \param lookup_to_num legacy meaning (?)
	 */
	inline RCValueObject GetValue(const MIIterator& mit, bool lookup_to_num = false) {
		return (DoGetValue(mit, lookup_to_num));
	}

	/////////////// Statistics //////////////////////

	/*! \brief Get rough minimum for packrow in level-1 encoding
	 *
	 * Old name: GetMin64
	 * \param mit iterator pointing to a packrow, for which min value is requested
	 * \return numeric value from the column in level-1 encoding. or NULL_VALUE_64 if not min available
	 * - Integers values can be used directly
	 * - Decimals must be externally properly shifted
	 * - Doubles must be cast e.g. *(double*)&
	 */
	inline _int64 GetMinInt64(const MIIterator& mit) {
		return DoGetMinInt64(mit);
	}

	/*! \brief Get minimum for packrow in level-1 encoding
	 * 	\return min or NULL_VALUE_64 if min not known exactly
	 */
	inline _int64 GetMinInt64Exact(const MIIterator& mit) {
		return DoGetMinInt64Exact(mit);
	}

	/*! \brief Get rough maximum for packrow in level-1 encoding
	 *
	 * Old name: GetMax64
	 * \param mit iterator pointing to a packrow, for which max value is requested
	 * \return numeric value from the column in level-1 encoding. or NULL_VALUE_64 if not max available
	 * - Integers values can be used directly
	 * - Decimals must be externally properly shifted
	 * - Doubles must be cast e.g. *(double*)&
	 */
	inline _int64 GetMaxInt64(const MIIterator& mit) {
		return DoGetMaxInt64(mit);
	}

	/*! \brief Get maximum for packrow in level-1 encoding
	 * 	\return max or NULL_VALUE_64 if max not known exactly
	 */
	inline _int64 GetMaxInt64Exact(const MIIterator& mit) {
		return DoGetMaxInt64Exact(mit);
	}


	/*! \brief Get (global) rough minimum for numerical column in level-1 encoding
	 *  For floating point types it is a double min encoded on _int64
	 */
	inline _int64 RoughMin() {
		return DoRoughMin();
	}
	/*! \brief Get (global) rough maximum for numerical column in level-1 encoding
	 *  For floating point types it is a double max encoded on _int64
	 */
	inline _int64 RoughMax() {
		return DoRoughMax();
	}
	/*! \brief True, if we can prove that the given data pack contains only nulls.
	*/
	virtual bool RoughNullsOnly(const MIIterator &mit) = 0;

	/*! \brief A coefficient of rough check selectivity for simple equality conditions.
	 *  \return Approximation of k/n, where k - no. of suspected data packs, n - no. of all data packs.
	 *  E.g. 1 means KNs are useless, 1/n means only one pack is to be checked (like for "a = 100" for autoincrement).
	 */
	virtual double RoughSelectivity()
	{ return 1; }

	//	inline RCBString GetMaxString(const MIIterator& mit) {
	//		return DoGetMaxString(mit);
	//	}
	//	inline RCBString GetMinString(const MIIterator& mit) {
	//		return DoGetMinString(mit);
	//	}
	//	 RCDateTime GetMaxDateTime(const MIIterator &mit);
	//	 RCDateTime GetMinDateTime(const MIIterator &mit);
	//	 RCNum GetMaxNum(const MIIterator &mit);
	//	 RCNum GetMinNum(const MIIterator &mit);

	/*!
	 * \brief Get number of NULLs in the packrow pointed by \e mit
	 * \pre \e mit is located at the beginning of current packrow
	 * \return number of nulls in the pack or NULL_VALUE_64 if not available
	 * Note that int may be not enough, as MIIterator packrow may be arbitrarily large.
	 */
	inline _int64 GetNoNulls(const MIIterator& mit) {
		return DoGetNoNulls(mit);
	}

	/*!
	 * \brief Return true if it is possible that the column will return null value,
	 * including nulls generated by multiindex null objects
	 */
	inline bool NullsPossible() {
		return DoNullsPossible();
	}

	/*! \brief Get the sum of values in the packrow pointed by \e mit
	 * \pre \e mit is located at the beginning of current packrow
	 * \return sum or NULL_VALUE_64 if sum not available
	 */
	inline _int64 GetSum(const MIIterator &mit) {
		return DoGetSum(mit);
	}

	/*! \brief Return true if the virtual column contain only non-null distinct values
	 *  Depends on the current multiindex state, e.g. if there is no row repetitions.
	 */
	inline bool IsDistinct() {
		return DoIsDistinct();
	}

	/*! \brief Return the best possible upper approximation of distinct values, incl. nulls.
	 *  Depends on the current multiindex state, e.g. KNs, if there are nulls etc.
	 */
	inline _int64 GetApproxDistVals(bool incl_nulls) {
		return DoGetApproxDistVals(incl_nulls);
	}

	/*! \brief Return the exact number of distinct values of the whole column, without nulls.
	 *  If the exact number is unsure, return NULL_VALUE_64.
	 */
	virtual _int64 GetExactDistVals() {
		return NULL_VALUE_64;
	}

	/*! \brief Return the upper approximation of text size the column may return.
	 *  May depend on the current multiindex state, e.g. KNs.
	 *  Used e.g. to prepare buffers for text values.
	 *  maximal byte string length in column
	 */
	inline ushort MaxStringSize() {
		return DoMaxStringSize();
	}

	inline const char* GetValueStringPtr(MIIterator const& mit) {
		return DoGetValueStringPtr(mit);
	}
	inline BHReturnCode GetValueRCNum(MIIterator const& mit, RCNum& rc) {
		return DoGetValueRCNum(mit, rc);
	}
	inline BHReturnCode GetValueDateTime(MIIterator const& mit, RCDateTime& dt, AttributeType t = RC_UNKNOWN) {
		return DoGetValueDateTime(mit, dt, t);
	}

	virtual RCBString	DecodeValue_S(_int64 code)	{ assert(0); return RCBString(); }	// lookup (physical) only
	virtual vector<_int64> GetListOfDistinctValues(MIIterator const& mit)
	{ vector<_int64> empty; return empty; }

	////////////////////////////////////////////////////////////

	/*! Get list of variables which must be supplied as arguments the obtain a value from this VirtualColumn
	 * It is a list of variables from underlying MysqlExpression, or a the PhysicalColumn if VirtualColumn wraps a PhysicalColumn
	 *
	 */
	virtual std::vector<VarMap>& GetVarMap() {return var_map;}

	/*! Get list of parameters which values must be set before requesting any value from this VirtualColumn
	 * It is a list of variables from underlying MysqlExpression, which do not belong to tables from teh current query
	 *
	 */
	virtual const  MysqlExpression::SetOfVars& GetParams() const {return params;}

	//! \brief parameters were set, so reevaluation is necessary
	virtual void RequestEval() {first_eval = true;} //TODO: checking if new parameter values are different form the previous ones

	//! \return true, if \e d2 is no longer needed
	virtual bool TryToMerge(Descriptor &d1,Descriptor &d2) { return false; }	// default behaviour, should be overriden

	//! \brief Assign a value to a parameter
	//! It is up the to column user to set values of all the parameters before requesting any value from the column
	virtual void SetParamTypes(MysqlExpression::TypOfVars* types) {};

	/*! Mark as true each dimension used as source for variables in this VirtualColumn
	 *
	 * \param dims_usage caller creates this vector with the proper size (number of dimensions in the multiindex)
	 */
	virtual void MarkUsedDims(DimensionVector& dims_usage);

	/*! Check the condition \e desc on a packrow from \e mind pointed by \e mit
	 *
	 * Does not perform a rough check - does row by row comparison
	 *
	 * \pre Assumes that triviality of desc is checked beforehand.
	 * Requires a new MultiIndex functionality:
	 * Tell, that changes to the current packrow are finished
	 * void MultiIndex::CommitPendingPackUpdate();
	 * Apply all pending changes to the MutliIndex, so they become visible
	 * void MultiIndex::CommitPendingUpdate();
	 */
	void EvaluatePack(MIUpdatingIterator& mit, Descriptor& desc) {
		DoEvaluatePack(mit, desc);
	}

	//! check whether any value from the pack may meet the condition
	inline RSValue RoughCheck(const MIIterator& it, Descriptor& d) {
		return DoRoughCheck(it, d);
	}

	//! check whether any value from the pack may meet the condition "... BETWEEN min AND max"
	virtual RSValue RoughCheckBetween(const MIIterator& it, _int64 min, _int64 max); 	// default: use RoughMin / RoughMax


	//! is a datapack pointed by \e mit NULLS_ONLY, UNIFORM, UNIFORM_AND_NULLS or NORMAL
	PackOntologicalStatus GetPackOntologicalStatus(const MIIterator& mit) {
		return DoGetPackOntologicalStatus(mit);
	}

	/*! \brief Lock datapacks providing parameters for contained expression; must be done before any data access.
	 *  \param mit Multiindex iterator pointing to a packrow to be locked
	 *  Lock strategy is controlled by an internal PackGuardian.
	 *  Default strategy: unlocking the pack when the next one is locked.
	 * \code // This is an example of use LockSourcePacks to list all values of VirtualColumn:
	 *       MIIterator it(current_multiindex);
	 *       while(it.IsValid()) {
	 *           if(it.PackrowStarted()) {
	 *               virt_column.LockSourcePack(it);
	 *           }
	 *           cout << virt_column.GetValueInt64(it) << endl;
	 *           ++it;
	 *       }
	 *       // may be unlocked manually, or just left until destructor does it
	 */
	virtual void LockSourcePacks(const MIIterator& mit);

	//! \brief Manually unlock all locked datapacks providing parameters for contained expression.
	//! Done also in destructor.
	virtual void UnlockSourcePacks();
	virtual void DisplayAttrStats()		{}

	///*! \brief Returns status of VirtualColumn
	// * \return VirtualColumn::VColStatus
	// */
	//virtual VColStatus GetStatus() const { return status; }

	/*! \brief Returns true if access to column values does not depend on MultiIndexIterator, false otherwise
	 * \return bool
	 */
	virtual bool IsConst() const = 0;

	/*! \brief Returns true if column values are deterministically set, false otherwise
	 * \return bool
	 * note - not properly implemented in MultivalueColumns.
	 */
	virtual bool IsDeterministic() { return true;}

	/*! \brief Returns true if column has parameters that have to be set before accessing its values, false otherwise
	 * \return bool
	 */
	virtual bool IsParameterized() const { return params.size() != 0; }

	/*! \brief Returns all dimensions (tables) involved by virtual column
	 *
	 */
	virtual dimensions_t GetDimensions();

	/*! \brief Returns the only dimension (table) involved by virtual column, or -1
	 *
	 */
	virtual int GetDim() {return dim;}

	ConnectionInfo& ConnInfo()		{ return conn_info; }
	MultiIndex* 	GetMultiIndex()	{ return mind; }

	/*! change the MultiIndex for which the column has been created, used in subquery execution
	 * \param m - new MultiIndex, should be a copy of the previously assigned one
	 * \param t - this copy of original (compilation-produced) TempTable must be used by the column to take values from
	 */
	void 			SetMultiIndex(MultiIndex* m, JustATablePtr t = JustATablePtr());

	virtual _int64 NoTuples(void);

	//! Is it a CONST column without parameters? (i.e. fully defined by a value)
	bool IsFullConst() { return IsConst() && !IsParameterized(); }

	static bool IsConstExpression(MysqlExpression* expr, int temp_table_alias, const vector<int>* aliases);

	virtual const MysqlExpression::bhfields_cache_t& GetBHItems() const = 0;

	virtual char *ToString(char p_buf[], size_t buf_ct) const	{ return p_buf; }

	typedef std::vector<VarMap> var_maps_t;

protected:
	virtual _int64 DoGetValueInt64(const MIIterator&) = 0;
	virtual int DoGetValueInt32(const MIIterator&) = 0;
	virtual bool DoIsNull(const MIIterator&) = 0;
	virtual void DoGetValueString(RCBString&, const MIIterator&, bool = false) = 0;
	virtual double DoGetValueDouble(const MIIterator&) = 0;
	virtual void DoGetValueBin(const MIIterator&, int&, char*) = 0;
	virtual RCValueObject DoGetValue(const MIIterator&, bool) = 0;
	virtual _int64 DoGetMinInt64(const MIIterator &) = 0;
	virtual _int64 DoGetMinInt64Exact(const MIIterator &) {return NULL_VALUE_64;}
	virtual _int64 DoGetMaxInt64Exact(const MIIterator &) {return NULL_VALUE_64;}
	virtual _int64 DoGetMaxInt64(const MIIterator &) = 0;
	virtual _int64 DoRoughMin() = 0;
	virtual _int64 DoRoughMax() = 0;
	virtual _int64 DoGetNoNulls(const MIIterator &) = 0;
	virtual bool DoNullsPossible() = 0;
	virtual _int64 DoGetSum(const MIIterator &) = 0;
	virtual bool DoIsDistinct() = 0;
	virtual _int64 DoGetApproxDistVals(bool incl_nulls) = 0;
	virtual ushort DoMaxStringSize() = 0;		// maximal byte string length in column
	virtual PackOntologicalStatus DoGetPackOntologicalStatus(const MIIterator&) = 0;
	virtual RSValue DoRoughCheck(const MIIterator&, Descriptor&) = 0;
	virtual void DoEvaluatePack(MIUpdatingIterator& mit, Descriptor&) = 0;
	virtual const char* DoGetValueStringPtr(MIIterator const&) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"not implemented for this type of column");
		return NULL;
	}
	virtual BHReturnCode DoGetValueRCNum(MIIterator const&, RCNum&) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"not implemented for this type of column");
		return BHRC_FAILD;
	}
	virtual BHReturnCode DoGetValueDateTime(MIIterator const& mit, RCDateTime& dt, AttributeType t = RC_UNKNOWN) {
		BHASSERT_WITH_NO_PERFORMANCE_IMPACT(!"not implemented for this type of column");
		return BHRC_FAILD;
	}

protected:
	MultiIndex* mind;
	ConnectionInfo& conn_info;
	//VColStatus status;
	var_maps_t var_map;
	MysqlExpression::SetOfVars params; //parameters - columns from tables not present in the local multiindex
	//MysqlExpression::TypOfVars param_types;  //currently unused
//	mutable MysqlExpression::var_buf_t param_buf; //places to store values assigned to parameters
	mutable ValueOrNull last_val;
	//! for caching, the first time evaluation is obligatory e.g. because a parameter has been changed

	mutable bool first_eval;

	VCPackGuardian pguard;
	int dim;						//valid only for SingleColumn and for ExpressionColumn if based on a single table only
									//then this an easily accessible copy of var_map[0]._dim. Otherwise it should be -1
};

#endif /* not VIRTUALCOLUMN_H_INCLUDED */

